// bsp_use.cpp by SamPlay
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef EXTDLL_H
#pragma warning(disable : 4244)
#pragma warning(disable : 4305)
typedef float vec_t;
#include <cmath>
#include "vector.h"
#endif

#include "vector_ext.h"
#include "u_RzbVector_ctpl.h"

#include "bspfile2.h"
#include "mathlib_tpl.h"
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
const K_Epsilonf= 1E-05;
int       areCoplanar( dplane_t * pPlan1_i,dplane_t * pPlan2_i);
int       areCoplanar( int planenum1_i, int planenum2_i);
int       haveSameNormal( dplane_t * pPlan1_i, dplane_t * pPlan2_i);
int       haveSameNormal( int planenum1_i, int planenum2_i);
int       getLeafFacePlanes( HLBspFile_ct * pHLBspFile_i, const dleaf_t & Leaf_a, int * & PlaneIds_o);
//======================================================
//--------------------------------------------------------
int areCoplanar( const HLBspFile_ct & Bsp_i, int planenum1_i, int planenum2_i)
	{
	return areCoplanar( & Bsp_i. dplanes[planenum1_i], & Bsp_i. dplanes[planenum2_i]);
	}
//--------------------------------------------------------
int areCoplanar( dplane_t * pPlan1_i, dplane_t * pPlan2_i)
	{
	vec3_t CP_a;
	CrossProduct<vec_t>( pPlan1_i-> normal, pPlan2_i-> normal, CP_a);
	return VectorLength<vec_t>(CP_a)< 0.001;
	}
//----------------------------------------------------
int haveSameNormal(const HLBspFile_ct & Bsp_i, int planenum1_i, int planenum2_i)	
	{
	return haveSameNormal( & Bsp_i. dplanes[planenum1_i], & Bsp_i. dplanes[planenum2_i]);
	}
//----------------------------------------------------
int haveSameNormal( dplane_t * pPlan1_i, dplane_t * pPlan2_i)
	{
	return DotProduct<vec_t>( pPlan1_i-> normal, pPlan2_i-> normal)> .999;//MAGIC#
	}
//---------------------------------------------
//-----------------------------------------
int getSafeNormal( HLBspFile_ct* pHLBspFile_i, dface_t * pFace_i, Vector & Normal_o)
	{
	if( pFace_i->numedges== 0)
		return 1;//empty
	if( pFace_i->numedges== 1)
		return 2;//1 point
//	dedge_t* pEdge_a;

	int VId0_a= pHLBspFile_i-> getFirstVtxId( pFace_i, 0);//, pEdge_a);
	Vector V0_a=  pHLBspFile_i-> dvertexes[VId0_a].point;//= pHLBspFile_i-> dvertexes[VId0_a];

	int i= 1;
	float Length_a= 0.0f;

	Vector V1_a;
	while( Length_a< .01/*K_Epsilonf*/  && i< pFace_i->numedges)
		{
		int VId1_a= pHLBspFile_i-> getFirstVtxId( pFace_i, i);//, pEdge_a);
		V1_a= pHLBspFile_i-> dvertexes[VId1_a].point;
		V1_a= V1_a- V0_a;
		Length_a= normalizeL(V1_a);
		++i;
		}
	if( Length_a< .01/*K_Epsilonf*/)
		return 3;//all points same
	Vector V2_a;
	Length_a= 0.0f;
	float LengthNorm_a= 0.0f;
	while( LengthNorm_a< .01/*K_Epsilonf*/ && i< pFace_i->numedges)
		{
		int VId2_a= pHLBspFile_i-> getFirstVtxId( pFace_i, i);//, pEdge_a);
		V2_a= pHLBspFile_i-> dvertexes[VId2_a]. point;
		V2_a= V2_a- V0_a;
		Length_a= normalizeL( V2_a);
		if( Length_a> .01/*K_Epsilonf*/)
			{
			CrossProduct( V1_a, V2_a, Normal_o);
			LengthNorm_a= normalizeL(Normal_o);
			}
		++i;
		}
	if( Length_a< K_Epsilonf)
		return 3;//only 2 diff points
	if( LengthNorm_a< K_Epsilonf)
		return 4;//all aligned
	return 0;
	}
//----------------------------------------------------
char * getTextureName( HLBspFile_ct * pHLBspFile_i, dface_t * pFace_i)
	{
	texinfo_t *ti = pHLBspFile_i-> texinfo + pFace_i-> texinfo;
	dmiptexlump_t *pMipTexLump = (dmiptexlump_t *)pHLBspFile_i-> dtexdata;
   if (pMipTexLump->dataofs[ti->miptex] != -1)  // not unused mip lump
		{
      miptex_t *pMipTex = (miptex_t*)(pHLBspFile_i-> dtexdata + pMipTexLump->dataofs[ti->miptex]);
      return pMipTex-> name;
      }
	return NULL;
	}
//----------------------------------------------------
//----------------------------------------------------
// find the contents of a coordinate in 3D space...
int HLBspFile_ct:: getPointContents( const vec3_t Coord_i)
	{
   // return contents (CONTENTS_EMPTY, CONTENTS_SOLID, CONTENTS_WATER, etc.)
	int LeafNodeNum_a= this-> findLeaf( Coord_i);
	return this-> dleafs[LeafNodeNum_a]. contents;
	}
//----------------------------------------------------
int HLBspFile_ct:: findLeaf( const vec3_t Point_i)
	{
	int      NodeNum_a;
	vec_t    SDist_a;
	dnode_t  *pNode_a;
	dplane_t *pPlane_a;
	NodeNum_a= 0;  //=dmodels[ 0]. headnode[ 0]; start at the root of the BSP tree
	while( NodeNum_a>= 0)// not a leaf
		{
		pNode_a= &this-> dnodes[NodeNum_a];
		pPlane_a= &this-> dplanes[pNode_a-> planenum];
		SDist_a= DotProduct(Point_i, pPlane_a-> normal)- pPlane_a-> dist;
		if( SDist_a< 0)
			NodeNum_a= pNode_a-> children[ 1];
		else
			NodeNum_a= pNode_a-> children[ 0];
		}
	return -1- NodeNum_a;//leaf index in dleafs array
	}
//---------------------------------------------------------------------
unsigned int HLBspFile_ct:: getFaceFlags ( int face_index_i)
{
dface_t * pFace_a = &this-> dfaces[this-> dmodels[ 0].firstface + face_index_i];
dplane_t * pPlane_a = &this-> dplanes[pFace_a->planenum];

unsigned int FaceFlags_a= 0;

if( pFace_a-> side)
	{
	if( pPlane_a-> normal[ 2]<=- 0.70)
		FaceFlags_a|= KB_FD_Walkable;
	}
else	if( pPlane_a-> normal[ 2]>= 0.70)
	FaceFlags_a|= KB_FD_Walkable;
return FaceFlags_a;
}
//----------------------------------------------------
int getLeafFacePlanes( HLBspFile_ct * pHLBspFile_i, const dleaf_t & Leaf_a, int * & PlaneIds_o)
{
	//_ only planes with a face on it
	//	PlaneIds_o must be deleted by calling fct
RszVector_cT<int, 6> PlaneIds_a;
PlaneIds_a. reserve(20);
int j;
int jMax= -1;
unsigned short* p = pHLBspFile_i-> dmarksurfaces + Leaf_a.firstmarksurface;
int bFound_a= 0;
int x= 0;
while( x< Leaf_a. nummarksurfaces)
	{
   bFound_a= 0;
	int face_idx = *p++;
	dface_t * pFace_a= & pHLBspFile_i-> dfaces[face_idx];
   int PlaneNum_a=  pFace_a-> planenum;
	j= 0;
	while( j< PlaneIds_a.size())
		{
		if( PlaneNum_a== PlaneIds_a[ j])
			{
			bFound_a= 1;
			break;
			}
		j++;
		}	
	if( !bFound_a)
		PlaneIds_a.push_back(PlaneNum_a);
	++x;
	}
int Size_a= PlaneIds_a. transfer( & PlaneIds_o, 1/*adjustsize*/);
return Size_a;
}
