//
// BSP_tool - botman's Half-Life BSP utilities
//
// (http://planethalflife.com/botman/)
//
// config.cpp
//
// Copyright (C) 2001 - Jeffrey "botman" Broome
//
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifdef WIN32
#include <windows.h>
#include <direct.h>
#else
#include <stdio.h>
#endif
#include "appgene.h"
#include "scriplib.h"

#include "BspViewConfig_ct.h"
#include "WorldRender_ct.h"
#include "UserInput_ct.h"

extern WorldRender_ct WorldRndr;
extern char           MapName[];

extern    int ScreenWidth;  // display width
extern    int ScreenHeight;  // display height
extern    int Bitsppix;
extern    int refresh_rate;
extern    bool bFullScreen;
extern    int x_pos;
extern    int y_pos;  // X & Y position of upper left corner

extern    bool_t  bEnable_noclip;
extern    float brightness;
extern    float gamma_adjust;

extern    int BaseMoveSpeed;
extern    int bBoundCamRange;
extern    float CollRadius;

extern    int bSurfOnly;
extern    int special_texture_transparency;

extern int   GrSphereNumOfSectors;
extern int   GrSphereNumOfSlices;
extern int   GrCylNumOfSectors;
extern float GrHullLineWidth;
extern vec3_t    AABoxHalfSize;
extern vec3_t    AABoxSize;

extern   char spawnpoint[];
//======================================================
BspViewConfig_ct config;
char BspViewConfig_ct:: ConfigFE[256]="BSP_view.cfg";
//====================================================
//-------------------------------------------------------
void BspViewConfig_ct:: reset()
{
   // set up configuration defaults...
ScreenWidth = 640;
ScreenHeight = 480;
Bitsppix = 16;
refresh_rate = 60;
bFullScreen = FALSE;

x_pos= -1;  // default is to center the display
y_pos= -1;

brightness = 0.75;
gamma_adjust = 1.0;

BaseMoveSpeed = 1;
UserInput_ct:: mouse_sensitivity = 0.15f;

UserInput_ct:: bInverted_mouse = FALSE;
bEnable_noclip = FALSE;

special_texture_transparency = 80;  // about 30% translucent

spawnpoint[0] = 0;
}
//-------------------------------------------------------
int BspViewConfig_ct:: init( char *config_file)
{
reset();
char BspViewConfigPFE_a[ MAX_PATH]; // path where the running application resides
_getcwd( BspViewConfigPFE_a, MAX_PATH);
strcat( BspViewConfigPFE_a, M_PATHSEP);
strcat( BspViewConfigPFE_a, config_file);
::LoadScriptFile( BspViewConfigPFE_a);
this-> ParseScriptFile();
return 0;
}
//---------------------------------------------------------
int BspViewConfig_ct:: ParseScriptFile(void)
{
while (1)
   {
   do
      {
      GetToken (true);  // look for a line starting with a '$'
      if (endofscript)
         return 0;  // indicate success
      if (token[0] == '$')
         break;
      while (TokenAvailable())
         GetToken (false);
      } while (1);

   if (strcmp(token, "$width") == 0)
      {
      GetToken(false);
      sscanf(token, "%d", &ScreenWidth);
      }
   else if (strcmp(token, "$height") == 0)
      {
      GetToken(false);
      sscanf(token, "%d", &ScreenHeight);
      }
   else if (strcmp(token, "$bpp") == 0)
      {
      GetToken(false);
      sscanf(token, "%d", &Bitsppix);
      }
   else if (strcmp(token, "$hertz") == 0)
      {
      GetToken(false);
      sscanf(token, "%d", &refresh_rate);
      }
   else if (strcmp(token, "$enable_fullscreen") == 0)
      {
      GetToken(false);
      sscanf(token, "%d", &bFullScreen);
      }
   else if (strcmp(token, "$x_pos") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &x_pos);
      }
   else if (strcmp(token, "$y_pos") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &y_pos);
      }
   else if (strcmp(token, "$enable_lighting") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &WorldRndr.bEnable_lighting);
      }
   else if (strcmp(token, "$brightness") == 0)
      {
         GetToken(false);
         sscanf(token, "%f", &brightness);
         if ((brightness < 0.0) || (brightness > 1.0))
            brightness = 0.75;
      }
   else if (strcmp(token, "$gamma_adjust") == 0)
      {
         GetToken(false);
         sscanf(token, "%f", &gamma_adjust);
         if ((gamma_adjust < 0.5) || (gamma_adjust > 2.0))
            gamma_adjust = 1.0;
      }

   else if (strcmp(token, "$movement_speed") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &BaseMoveSpeed);
         if (BaseMoveSpeed < 1)
            BaseMoveSpeed = 1;
         if (BaseMoveSpeed > 5)
            BaseMoveSpeed = 5;
      }
   else if (strcmp(token, "$mouse_sensitivity") == 0)
      {
         GetToken(false);
         sscanf(token, "%f", &UserInput_ct:: mouse_sensitivity);
         if (UserInput_ct:: mouse_sensitivity < 0.0)
            UserInput_ct:: mouse_sensitivity = 0.15f;

         if (UserInput_ct:: mouse_sensitivity > 1.0)
            UserInput_ct:: mouse_sensitivity = 1.0;
      }
   else if (strcmp(token, "$inverted_mouse") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &UserInput_ct:: bInverted_mouse);
      }
   else if (strcmp(token, "$camcollisionradius") == 0)
      {
         GetToken(false);
         sscanf(token, "%f", &CollRadius);
      }

   else if (strcmp(token, "$special_texture_transparency") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &special_texture_transparency);
         if ((special_texture_transparency < 0) || (special_texture_transparency > 255))
            special_texture_transparency= 80;  // about 30% translucent
      }
/*
   else if (strcmp(token, "$bspfile") == 0)
      {
      if (MapName[0])
	      Error("Duplicate $bspfile entry found in .cfg file!\n");
      GetToken(false);
      strcpy(MapName, token);
      StripExtension(MapName);
		}
*/
   else if (strcmp(token, "$spawnpoint") == 0)
      {
      if (spawnpoint[0])
         Error("multiple $spawnpoint found in .cfg file!\n");
      GetToken(false);
      sscanf(token, "%s", spawnpoint);
      }
 
   else if (strcmp(token, "$boundcamrange") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &bBoundCamRange);
      }

   else if (strcmp(token, "$spherenumofsectors") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", & GrSphereNumOfSectors);
      }
   else if (strcmp(token, "$spherenumofslices") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", & GrSphereNumOfSlices);
      }
   else if (strcmp(token, "$cylnumofsectors") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", & GrCylNumOfSectors);
      }
   else if (strcmp(token, "$hulllinewidth") == 0)
      {
         GetToken(false);
         sscanf(token, "%f", & GrHullLineWidth);
      }
/*
$AABoxHalfSizeXY 18
$AABoxHalfSizeZ 36
*/
    else if (strcmp(token, "$AABoxHalfSizeXY") == 0)
      {
         GetToken(false);
         sscanf(token, "%f", & AABoxHalfSize[0]);
      AABoxHalfSize[1]= AABoxHalfSize[0];
		AABoxSize[1]=AABoxSize[0]=2.0f*AABoxHalfSize[0];
		}
 
    else if (strcmp(token, "$AABoxHalfSizeZ") == 0)
      {
         GetToken(false);
         sscanf(token, "%f", & AABoxHalfSize[2]);
		AABoxSize[2]=2.0f*AABoxHalfSize[2];
		}

  else if (strcmp(token, "$SurfOnly") == 0)
      {
		GetToken(false);
      sscanf(token, "%d", & bSurfOnly);
      }

/*
   else if (strcmp(token, "$show_edges") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", & WorldRndr. show_edges);
         if ((WorldRndr. show_edges < 0) || (WorldRndr. show_edges > 2))
            WorldRndr. show_edges = 0;
      }
   else if (strcmp(token, "$noclip") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &bEnable_noclip);
      }

   else if (strcmp(token, "$drawsky") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &WorldRndr. bDrawSky);
      }
   else if (strcmp(token, "$drawentities") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &WorldRndr. bDrawEntities);
      }


   else if (strcmp(token, "$renderspectextures") == 0)
      {
         GetToken(false);
         sscanf(token, "%d", &WorldRndr. bRenderSpecialTextures);
      }
*/
   else
      Error("Unknown BspViewConfig file option: %s\n", token);
   }
}
