#ifdef WIN32
#pragma once
#endif
#ifndef CHUNKEDVECTOR_TPL_H_
#define CHUNKEDVECTOR_TPL_H_
//#include <vector>
//vector of T, chunked in sub-vectors "elems" of KT_elem_count coordinates 
//+++++++++++++++++++++++++++++++++++++++++++++++++++
template<typename CoordType_T, const int KT_elem_count, const int KT_InitVElmSize>
class ChunkedVector_cT 
{
public:
typedef CoordType_T Vec_t[KT_elem_count];

    ChunkedVector_cT( int NumVElm_i= 0): coordsizemax(NumVElm_i* KT_elem_count), elemsize(NumVElm_i), Data(0)
		{	Data= (NumVElm_i? new CoordType_T[coordsizemax]: 0);
		}
    ~ChunkedVector_cT() 
		{   clear();
		}
	ChunkedVector_cT( const ChunkedVector_cT & A_i)
		{
		coordsizemax= A_i. coordsizemax;
		elmsize= A_i. elmsize;
		Data= new CoordType_T[ coordsizemax];
		memcpy( Data, A_i. Data, coordsizemax*sizeof(CoordType_T));
		}
	ChunkedVector_cT & operator=( const ChunkedVector_cT & A_i)
		{
		if( this!= & A_i)
			{
			coordsizemax= A_i. elmsize* KT_elem_count;
			elmsize= A_i. elmsize;
			if( Data)
				delete[] Data;
			Data= new CoordType_T[ coordsizemax];
			memcpy( Data, A_i. Data, coordsizemax*sizeof(CoordType_T));
			}
		return *this;
		}

	void clear()
		{   delete []Data;
		}
	void reset()
		{
      clear();
		init();
		}

	void resize( int CoordSize_i)
		{
	   if( coordsizemax<= CoordSize_i) 
			resize0( CoordSize_i);
		if( elemsize< CoordSize_i/KT_elem_count) 
			elemsize= CoordSize_i/KT_elem_count;
		}

	 CoordType_T & add( CoordType_T * Source_i)//add one elem only 
		 {
		 if( coordsizemax< (1+ elemsize)*KT_elem_count) 
         resize0( coordsizemax+ KT_InitVElmSize* KT_elem_count);
		CoordType_T * Dest_a= Data+ elemsize*KT_elem_count;
		memcpy( Dest_a, Source_i, KT_elem_count* sizeof(CoordType_T));//  *Dest_a++ = *Source_i++;
		++elemsize;
		return *Dest_a;
		}

    operator CoordType_T*(void) const { return Data; }
    const CoordType_T* operator[](int i) const { return Data + i*KT_elem_count; }
    CoordType_T* operator[](int i) { return Data + i*KT_elem_count; }

	 int size() const { return elemsize; }//same as STL
	 int max_size() const { return coordsizemax/KT_elem_count;	}//same as STL
	 bool empty() const { return 0==elemsize; }//same as STL

private:
int coordsizemax;
int elemsize;
CoordType_T * Data;
    void resize0( int coordsizemax_i)
		{
	   CoordType_T* NewData_n= new CoordType_T[coordsizemax_i];
  		memcpy( NewData_n, Data, coordsizemax*sizeof(CoordType_T));
		delete[] Data;
      Data= NewData_n;
		coordsizemax= coordsizemax_i;
		}
	void init()
		{
		Data= 0;
		coordsizemax= 0;
		elemsize= 0;
		}
};

#endif //eof