// laefbounds.cpp 60
// Copyright (C) 2005 - Samplay
// 
// portion copyright from...
/***
*
*	Copyright (c) 1996-2002, Valve LLC. All rights reserved.
*	
*	This product contains software technology licensed from Id 
*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*	All Rights Reserved.
*
****/

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#pragma warning (disable: 4244)

#include "u_RzbVector_ctpl.h"

//#define IMPL_CHEBYCHEFF

#include "bsp_ext.h"
#include "World_ct.h"
#ifdef IMPL_CHEBYCHEFF
#include "linprog.h"
#endif

extern WorldExtd_ct World;
extern char MapName[];

void Error( char *,...);
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//Bounds actually include all bounding windings of a leaf ( not only non-solid bounds ie portals)
#define	SIDESPACE	10

#ifdef M_DEBUGC
#ifndef DBGASSERT
#include <cassert>
#define DBGASSERT(a) assert(a)
#endif
#else
#define DBGASSERT(a)
#endif
//#define IMPL_SOLIDBoundS
LeafBound_ct::Ctnr_ct LeafBound_ct::Ctnr;

void removeBound (LeafBound_ct *pBound_i, int Node_i, LeafBound_ct * & pFirstBound_i);
void outTxtLeafs2( HLBspFile_ct * pHLBspFile_i, int PrintLevel_i);
int getLeafFacePlanes( HLBspFile_ct * pHLBspFile_i, const dleaf_t & Leaf_a
	, int * & PlaneIds_o);
//iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii
static void recursbuildBounds( HLBspFile_ct * pHLBspFile_i, BoundLists_ct &, int NodeId_i, LeafBound_ct * pBoundList_i);
static LeafBound_ct * buildNodeSplitPlaneBound ( HLBspFile_ct * pHLBspFile_i, int NodeId_i, LeafBound_ct * pBoundList_i);
static void divideNodeBounds (HLBspFile_ct * pHLBspFile_i, int NodeId_i, LeafBound_ct * & pBoundList_i
	, LeafBound_ct * & pFrontBounds_i, LeafBound_ct * & pBackBounds_i);
static LeafBound_ct	* buildTopNodeBounds (HLBspFile_ct * pHLBspFile_i, BoundLists_ct & BoundLists_i, int headnodenum_i
	, vec3_t mins, vec3_t maxs/*, vecd3_t mins, vecd3_t maxs*/);
//=============================================================

//---------------------------------------------------------------------
int HLBspFileExtsn_ct:: calcBspFileFaceFlags()
	{
	if( FaceFlags)
		delete[] FaceFlags;
	FaceFlags= new unsigned int[pHLBspFile-> dmodels[0]. numfaces];
	for( int face_index_a= 0; face_index_a< pHLBspFile-> dmodels[0]. numfaces; face_index_a++)
		FaceFlags[ face_index_a]= pHLBspFile-> getFaceFlags( face_index_a);
	return 0;
	}
//--------------------------------------------------------------------
void HLBspFileExtsn_ct:: buildBspFileLeafBounds()
	{
	LeafBound_ct::Ctnr. reserve( 3* pHLBspFile->numleafs);

	LeafBoundLists. reserve( pHLBspFile->numleafs);
	LeafBoundLists. Size= pHLBspFile->numleafs;
	for( int i=0; i< pHLBspFile->numleafs;i++)	LeafBoundLists[i]= NULL;

	LeafBound_ct * pBoundList_a= buildTopNodeBounds (pHLBspFile, LeafBoundLists, 0, World.Mins, World.Maxs);
#ifdef IMPL_SOLIDBoundS
	LeafBoundLists[ 0]= pBoundList_a;
#endif
	::recursbuildBounds( pHLBspFile, LeafBoundLists, 0, pBoundList_a);
	}
//--------------------------------------------
void recursbuildBounds( HLBspFile_ct * pHLBspFile_i, BoundLists_ct & BoundLists_i, int NodeId_i, LeafBound_ct * pBoundList_i)
	{
	if( !pBoundList_i)
		 return;
	if( NodeId_i< 0)
		{
		int LeafNum_a= -1-NodeId_i;
		if( LeafNum_a==0)
			{
#ifdef IMPL_SOLIDBoundS
	NOT TESTED
//			LeafNum_a==0 is a particular case as it is the generic solid leaf, 
//       indexed ( or pointed to) by DIFFERENT nodes as one of their (leaf) children;
//       so, as opposed to all the other leafs, 
//       it is referred to more than once in the function call args 

			LeafBound_ct * pBound_a= BoundLists_i;
			LeafBound_ct * pLastBound_a;
			int NodeSide_a;
			while( pBound_a)
				{
				NodeSide_a= pBound_a-> pNodeIds[0]= -1? 0:1;
				pLastBound_a =pBound_a;
				pBound_a= pBound_a-> pNext[ NodeSide_a];
				}
			pLastBound_a-> pNexts[ NodeSide_a]= BoundLists[ 0];
			BoundLists[ 0]= pLastBound_a;
#endif
			}
		else
			{
			DBGASSERT( BoundLists_i[ LeafNum_a]==0);
			BoundLists_i[ LeafNum_a]= pBoundList_i;
			}
		return;
		}

	dnode_t * pNode_i= & pHLBspFile_i-> dnodes[NodeId_i];

	LeafBound_ct * pBoundSplitPlane_a= buildNodeSplitPlaneBound( pHLBspFile_i, NodeId_i, pBoundList_i);
	LeafBound_ct * pFrontBounds_a= 0;
	LeafBound_ct * pBackBounds_a= 0;
	divideNodeBounds( pHLBspFile_i, NodeId_i, pBoundList_i, pFrontBounds_a, pBackBounds_a);

	pBoundSplitPlane_a-> NodeIds[0]= pNode_i-> children[0];
	pBoundSplitPlane_a-> pNexts[0]= pFrontBounds_a; pFrontBounds_a= pBoundSplitPlane_a;
	pBoundSplitPlane_a-> NodeIds[1]= pNode_i-> children[1];
	pBoundSplitPlane_a-> pNexts[1]= pBackBounds_a; pBackBounds_a= pBoundSplitPlane_a;
	
	recursbuildBounds( pHLBspFile_i, BoundLists_i, pNode_i-> children[0], pFrontBounds_a);
	recursbuildBounds( pHLBspFile_i, BoundLists_i, pNode_i-> children[1], pBackBounds_a);
	}
//--------------------------------------------------------------------
LeafBound_ct	* buildTopNodeBounds (HLBspFile_ct * pHLBspFile_i, BoundLists_ct & BoundLists_i, int headnodenum_i
	, vec3_t mins, vec3_t maxs/*, vecd3_t mins, vecd3_t maxs*/)
	{
	int			i, j, n;

	vecd3_t		bounds[2];
	LeafBound_ct	*pBound_n, *Bounds[6];
	dplane_t	bplanes[6], *pl;
	
	for (i=0 ; i<3 ; i++)
		{
		bounds[0][i] = mins[i]- SIDESPACE;
		bounds[1][i] = maxs[i]+ SIDESPACE;
		}
	for (i=0 ; i<3 ; i++)
		{
		for (j=0 ; j<2 ; j++)
			{
			n = j*3 + i;

			pBound_n= LeafBound_ct::create();
			Bounds[n]= pBound_n;
			
			pl= &bplanes[n];
			memset(pl, 0x00, sizeof(*pl));
			if( j)
				{
				pl-> normal[ i] = -1;
				pl-> dist = -(float)bounds[j][i];
				}
			else
				{
				pl-> normal[i]= 1;
				pl-> dist = (float)bounds[j][i];
				}
			pBound_n-> plane= *pl;
			pBound_n-> winding= Winding_ct::BaseWindingForPlane (pl);
			}
		}
	// clip the basewindings by all the other planes
	for (i=0 ; i<6 ; i++)
		{
		for (j=0 ; j<6 ; j++)
			{
			if (j == i)
				continue;
			Bounds[i]-> winding-> clip( &bplanes[j], true);
			}
		}

	LeafBound_ct * pBoundList_a;
	pBoundList_a= 0;
	for (n=0 ; n< 6 ; n++)
		{
		Bounds[ n]-> NodeIds[ 0]= headnodenum_i;
		Bounds[ n]-> NodeIds[ 1]= -1;//dleafs[-1-(-1)]= dleafs[0]"universal" solid leaf with no face
		Bounds[ n]-> pNexts[ 0]= pBoundList_a;
		Bounds[ n]-> pNexts[ 1]= pBoundList_a;
		pBoundList_a= Bounds[n];
		}
	return pBoundList_a;//
	}
//--------------------------------------------------
LeafBound_ct * buildNodeSplitPlaneBound ( HLBspFile_ct * pHLBspFile_i, int NodeId_i, LeafBound_ct * pBoundList_i)
	{
	LeafBound_ct	* new_Bound;
	LeafBound_ct	* p;
	dplane_t	   * pSplittingPlane_a;
	dplane_t	     Clipplane_a;
	Winding_ct	* w;
	int			  nodeside;

	dnode_t * pNode_a= & pHLBspFile_i-> dnodes[ NodeId_i];
	pSplittingPlane_a= & pHLBspFile_i-> dplanes[ pNode_a-> planenum];//splitting plane
	// create a huge winding based on splitting plane
	w= Winding_ct:: BaseWindingForPlane( pSplittingPlane_a);

	new_Bound= new LeafBound_ct;
	new_Bound-> plane= *pSplittingPlane_a;
	new_Bound-> GenNode= pNode_a;

	nodeside= 0;
	// clip winding by pre-existing Bounds
	p= pBoundList_i;
	while( p)	
		{
		Clipplane_a= p-> plane;
		if( p-> NodeIds[ 0]== NodeId_i)
			nodeside= 0;
		else if( p-> NodeIds[ 1]== NodeId_i)
			{
			Clipplane_a.dist= -Clipplane_a.dist;
			for (int i=0 ; i<3 ; i++)	Clipplane_a.normal[i]=- Clipplane_a.normal[i];
			nodeside= 1;
			}
		else
			Error ("buildNodeSplitPlaneBound: mislinked Bound");

		// clip splitting plane winding to this Bound plane
		w-> clip( &Clipplane_a, true);
		if(!w->numpoints)
			{
			delete new_Bound;
			return NULL;
			}
		p= p-> pNexts[ nodeside];
		}
	LeafBound_ct::Ctnr. push_back(new_Bound);
	new_Bound-> winding = w;	
	return new_Bound;
	}
//-----------------------------------------------------------------
void divideNodeBounds (HLBspFile_ct * pHLBspFile_i, int NodeId_i, LeafBound_ct * & pBoundList_i
	, LeafBound_ct * & pFrontBounds_i, LeafBound_ct * & pBackBounds_i)
	{
	pFrontBounds_i=0;
	pBackBounds_i=0;

	LeafBound_ct	*pBound_a;
	int		fId, bId;
	int		nodeside;
	dplane_t	*pSplittingPlane_a;
	Winding_ct	*frontwinding, *backwinding;

	dnode_t * pNode_i= & pHLBspFile_i-> dnodes[NodeId_i];
	pSplittingPlane_a= &pHLBspFile_i-> dplanes[pNode_i->planenum];
	fId= pNode_i-> children[0];
	bId= pNode_i-> children[1];

	while( pBound_a= pBoundList_i) 	
		{
		if( pBound_a-> NodeIds[ 0]== NodeId_i)
			nodeside= 0;
		else if( pBound_a-> NodeIds[ 1]== NodeId_i)
			nodeside= 1;
		else
			Error ("divideNodeBounds: mislinked Bound");
		pBoundList_i= pBound_a-> pNexts[ nodeside];//detach from node's list
		
		// cut the Bound into two Bounds, one on each side of the cut plane
		Winding_ct::DivideWinding( pBound_a-> winding, pSplittingPlane_a, &frontwinding, &backwinding);
		
		if(!frontwinding)
			{
			pBound_a-> NodeIds[ nodeside]= bId;
			pBound_a-> pNexts[ nodeside]= pBackBounds_i;
			pBackBounds_i= pBound_a;
			continue;
			}
		if (!backwinding)
			{
			pBound_a-> NodeIds[ nodeside]= fId;
			pBound_a-> pNexts[ nodeside]= pFrontBounds_i;
			pFrontBounds_i= pBound_a;	
			continue;
			}
			
		// the winding is split
		delete pBound_a-> winding;
		
		LeafBound_ct	* backBound;
		backBound= LeafBound_ct::create();
		backBound-> copyAsDividedFrom( *pBound_a);
		backBound-> winding= backwinding;
		pBound_a-> winding= frontwinding;

		pBound_a-> NodeIds[ nodeside]= fId;
		pBound_a-> pNexts[ nodeside]= pFrontBounds_i;
		pFrontBounds_i= pBound_a;

		backBound-> NodeIds[ nodeside]= bId;
		backBound-> pNexts[ nodeside]= pBackBounds_i;
		pBackBounds_i= backBound;

		backBound-> NodeIds[ nodeside^1]= pBound_a-> NodeIds[ nodeside^1];
		backBound-> pNexts[ nodeside^1]= pBound_a-> pNexts[ nodeside^1];
		pBound_a-> pNexts[ nodeside^1]= backBound;
		}
	}
//-----------------------------------------------------
void removeBound (LeafBound_ct *pBound_i, int NodeId_i, LeafBound_ct * & pFirstBound_i)
	{ 
	LeafBound_ct * * ppBoundPN_a= & pFirstBound_i;
	LeafBound_ct * pBound_a;
	int Side_a;
	while( (pBound_a= *ppBoundPN_a) && pBound_a!= pBound_i)
		{
		Side_a= ( pBound_a-> NodeIds[ 0]== NodeId_i? 0: 1);
		ppBoundPN_a= & pBound_a-> pNexts[ Side_a];// ct address of the next pointer of pBound_a
		}
#if MC_CCOMPIL & MCK_DEBUG
	if(! pBound_a)
		{
		MF_DBGOS2( KB_fErr, <<"trying to remove a Bound not in a Bound list");		
		}
#endif
	DBGASSERT( pBound_a);
	Side_a= ( pBound_a-> NodeIds[ 0]== NodeId_i? 0: 1);
	*ppBoundPN_a= pBound_a-> pNexts[ Side_a];	
	pBound_a-> NodeIds[ Side_a]= -1;
	}
//--------------------------------------
void outTxtLeafs2( HLBspFile_ct * pHLBspFile_i, BoundLists_ct & BoundLists_i, int PrintLevel_i)
	{
	char FE_a[256];
	sprintf( FE_a, "_%s_Leafs2", MapName);

//	if( TypeFlag== K_BS_Std)
//		{	
		strcat(FE_a , ".txt");
//		}
//	else
//		{	strcat(FE_a , "M.txt");
//		}

	remove(FE_a);
	FILE * pFile_a= fopen(FE_a,"a");

	pHLBspFile_i-> outText_stats( pFile_a);

	int i;
//	int j;
	for( i=0; i< pHLBspFile_i-> numleafs; i++)
		{
		dleaf_t* pLeaf_a= & pHLBspFile_i-> dleafs[i];
		fprintf( pFile_a, "Leaf#%5i contents=%d\n", i , pLeaf_a-> contents);
/*
		// print faces
		unsigned short *p = this-> dmarksurfaces + pLeaf_a-> firstmarksurface;
		for ( j= 0; j< pLeaf_a-> nummarksurfaces; j++)
			{
			int face_idx = *p++;
			dface_t * pFace_a= & this-> dfaces[ face_idx];
			dplane_t * pPlane_a= & this-> dplanes[ pFace_a->planenum];
			fprintf(pFile_a, "  face#%i plane#%i side=%i\n", face_idx, pFace_a-> planenum, pFace_a-> side);
			}
*/
/*
		if( PrintLevel_i>= 2)
			{
			for( int i= 0; i< LBPlanes_a. LeafVtces.size(); i++)
				{
				vec_t * Vtx_a;
				Vtx_a= LBPlanes_a. LeafVtces[i];
				fprintf( pFile_a, "v#=%d  %6.2f %6.2f %6.2f\n", i, Vtx_a[0], Vtx_a[1], Vtx_a[2]);
				}
			}	
*/
		LeafBound_ct * pBound_a= BoundLists_i[i];
		while( pBound_a)
			{
			dplane_t * pPlane_a= & pBound_a-> plane;
			int NodeSide_a= (pBound_a-> NodeIds[0]== -1-i? 0:1);
			int NeighbourLeafId_a= -1-pBound_a-> NodeIds[NodeSide_a^1];
			fprintf(pFile_a, "  bdgp"
				" Nrm= %6.2f %6.2f %6.2f d=%5.0f nb#%4i nbcontents%2i\n"
				, pPlane_a->normal[0], pPlane_a->normal[1], pPlane_a->normal[2], pPlane_a->dist
				, NeighbourLeafId_a, pHLBspFile_i-> dleafs[NeighbourLeafId_a]. contents );
			pBound_a= pBound_a-> pNexts[NodeSide_a];
			}		
		fprintf(pFile_a,"\n");
		}
	fflush(pFile_a);
	fclose(pFile_a);
	}
//----------------------------------------------------
int HLBspFileExtsn_ct:: getLeafBoundingPlanes( const dleaf_t & Leaf_a, PlaneOrient_st * & PlaneIds_o)
{
int LeafNum_a= (& Leaf_a- pHLBspFile-> dleafs)/sizeof( dleaf_t);
return getLeafBoundingPlanes( LeafNum_a, PlaneIds_o);
}
//----------------------------------------------------
int HLBspFileExtsn_ct:: getLeafBoundingPlanes( int LeafIndex_i, PlaneOrient_st * & PlaneIds_o)
{
	//_ all planes, including planes with no face
	// PlaneIds_o must be deleted by calling fct
RszVector_cT<PlaneOrient_st, 4> PlaneIds_a;
PlaneIds_a.reserve(20);
LeafBound_ct * pBound_a= LeafBoundLists[ LeafIndex_i];
int j;
int bFound_a;
while( pBound_a)
	{
	int NodeSide_a= (pBound_a-> NodeIds[0]== -1- LeafIndex_i? 0:1);
   bFound_a= 0;
	dplane_t * pPlane_a= & pBound_a-> plane;
   int PlaneNum_a=  pBound_a-> PlaneId_();
	DBGASSERT( PlaneNum_a!=-1);
	j= 0;
	while( j< PlaneIds_a.size())
		{
		if( PlaneNum_a== PlaneIds_a[ j]. PlaneNum)
			{
			bFound_a= 1;
			break;
			}
		j++;
		}	
	if( !bFound_a)
		{
		PlaneIds_a.push_back(PlaneOrient_st( PlaneNum_a,NodeSide_a));
		}
	pBound_a= pBound_a-> pNexts[NodeSide_a];
	}		

int Size_a= PlaneIds_a. transfer( & PlaneIds_o, 1/*adjustsize*/);
return Size_a;
}
//
#ifdef IMPL_CHEBYCHEFF
//------------------------------------------------
int getChebyCenter( HLBspFileExtsn_ct * pBSE_i, int LeafIndex_i, Vector & Center_i, float & Radius_i)
{

FILE * pFile_a= fopen("_lpin.txt","w");
int j;

PlaneOrient_st * LeafBoundingPlanes;//must be delete after getLeafBoundingPlanes call
int planeCount_a= pBSE_i-> getLeafBoundingPlanes( LeafIndex_i, LeafBoundingPlanes);

fprintf( pFile_a, "4 %d\n", 1+ planeCount_a);
fprintf( pFile_a, "0. 0. 0. 1. 0.\nprojective\n1 0. 0. 0. 0. 1.\n");

for ( j=0; j< planeCount_a; j++)
	{
	PlaneOrient_st * PlaneOrient_a= & LeafBoundingPlanes[j];
	dplane_t * pPlane_a= & pHLBspFile-> dplanes[PlaneOrient_a->PlaneNum];
	if( !PlaneOrient_a->Side)
		fprintf( pFile_a, " 1 %f %f %f 1.0 %f\n"
			, pPlane_a->normal[0], pPlane_a->normal[1], pPlane_a->normal[2], -pPlane_a->dist);
	else
		fprintf( pFile_a, " 1 %f %f %f 1.0 %f\n"
			, -pPlane_a->normal[0], -pPlane_a->normal[1], -pPlane_a->normal[2], pPlane_a->dist);
	}
fclose(pFile_a);

double Result_a[ 5];
int LpStatus_a;
LinProg_ct LinProg_a;
LpStatus_a=LinProg_a. initFromFile("_lpin.txt", Result_a);
unsigned seed_a= 0;//LinProg_ct:: K_FNOPERM;

int NumTrials= 10;
int CurrTrialNum= 0;
int bSuccess= 0;
//char Msg_a[256];

while( ++CurrTrialNum< NumTrials && !bSuccess)
	{	
	bSuccess= 1;
	LpStatus_a= LinProg_a. run( seed_a);
	if( (LpStatus_a!= LinProg_ct:: K_MINIMUM) && (LpStatus_a!= LinProg_ct:: K_AMBIGUOUS))
		{
//		sprintf( Msg_a, "PROBLEM getChebyCenter(): Leaf#%d  linprog Status=%d\n", i, LpStatus_a);
		bSuccess= 0;
		}
	else
		{
		for(int k= 0; k<3; k++) Center_i[k]=(float)(Result_a[k]);
		Radius_i= -(float)(Result_a[3]);	
		//checks
/*
		for ( j=0; j< ;j++)
			{
			PlaneOrient_st * pPlaneOrient_a= LeafBoundingPlanes[j];
			dplane_t * pPlane_a= & dplanes[pPlaneOrient_a-> PlaneNum];
			int Side_a= pPlaneOrient_a-> Side;
			float InPointDist_a= DotProduct( Center_i, pPlane_a-> normal)- pPlane_a->dist;
			if( (!Side_a )&& InPointDist_a<= 0)
				{	
				bSuccess= 0;
//				sprintf( Msg_a,"PROBLEM getChebyCenter(): Leaf#%d P#%d Side=%d Red=%d IPDist=%f<=0!!!\n", i, pNode_a-> planenum, Side_a, (( pBPNode_a->Flags & KB_Redund)!=0), InPointDist_a);
				}
			else if( Side_a && InPointDist_a>= 0)
				{	
				bSuccess= 0;
//				sprintf( Msg_a,"PROBLEM getChebyCenter(): Leaf#%d P#%d Side=%d Red=%d IPDist=%f>=0!!!\n", i, pNode_a-> planenum, Side_a, (( pBPNode_a->Flags & KB_Redund)!=0), InPointDist_a);
				}
			}
*/
		int LN_a= pHLBspFile-> findLeaf( Center_i);
		if( LN_a!= LeafIndex_i)
			{
			bSuccess= 0;
//			sprintf( Msg_a,"PROBLEM getChebyCenter(): Leaf#%d  different from findleaf(center)=%d\n", i, LN_a);
			}
		++seed_a;
		}
	}

LinProg_a. clear();
delete [] LeafBoundingPlanes;
return bSuccess;
}
#endif //IMPL_CHEBYCHEFF