/* 
 * HL rendering engine
 * Copyright (c) 2000,2001 Bart Sekura
 *
 * Permission to use, copy, modify and distribute this software
 * is hereby granted, provided that both the copyright notice and 
 * this permission notice appear in all copies of the software, 
 * derivative works or modified versions.
 *
 * THE AUTHOR ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION AND DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES 
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 */

//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


#include "mathlib_tpl.h"
//#include "mathlib.h"

#ifndef __plane_h__
#define __plane_h__

///////////////////////////////////////////////////////////////
#define EPSILON (1.0e-05f)
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class RenderPlane_ct 
{
public:
vec3_t   normal;
float    Dist;

    RenderPlane_ct(): Dist(0.0f) {}
    RenderPlane_ct(vec3_t a, vec3_t b, vec3_t c) 
		{
      //_ compute normal
		vec3_t p;
      VectorSubtract(b, a, p);
      vec3_t q;
      VectorSubtract(c, a, q);
      CrossProduct(p, q, normal);
      VectorNormalize(normal);
		// M dist to plane=  DotProduct( M- PointOnPlane, Plane normal)
		// =  DotProduct( M, Plane normal) - Dist with
      // Dist= DotProduct( PointOnPlane, Plane normal)= dist of plane from origin(0,0,0))

		Dist= (float)DotProduct(normal, a);
		}
    RenderPlane_ct(const RenderPlane_ct& p) 
		{
      VectorCopy( p.normal, normal);
      Dist= p. Dist;
		}
    RenderPlane_ct(vec3_t Normal_i, vec3_t pt) 
		{
      VectorCopy( Normal_i, normal);
      Dist= (float)DotProduct(normal, pt);
		}
    RenderPlane_ct(vec3_t Normal_i, float Dist_i) 
		{
      VectorCopy( Normal_i, normal);
      Dist= Dist_i;
		}

    void changeOrient()
		{
		normal[0]= -normal[0];
		normal[1]= -normal[1];
		normal[2]= -normal[2];
		Dist= - Dist;
		}
/*
	 void perpendicularTo(const vec3_t a, const vec3_t b, vec3_t normal) 
		{
      vec3_t temp;
      VectorAdd(b, normal, temp);
      *this= RenderPlane_ct(a, b, temp);
		}
*/
    float distOf(const vec3_t pt) const 
		{  return (float)DotProduct(normal, pt)- Dist;
		}
    int checkIntersect(const vec3_t start, const vec3_t dir, vec3_t ret_o) const;

    // accessors
    const vec3_t & normal_() const { return normal; }
    float Dist_() const{ return Dist; }

};

#endif //eof

