// update.cpp by SamPlay
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <GL/glut.h>
#include  <time.h> // for srand init
#include  <string.h>
#include  <stdlib.h>

#include "appgene.h"
#include "Camera.h"
#include "World_ct.h"

extern float            frames_per_second;
extern Camera_ct        Camera;
extern vec3_t           spawn_point;
extern float            spawn_point_yaw;
extern char             MapName[];
extern int              CollisionCutoff;

float                   calcSpeed();
void                    outBSPToFile();
void                    outLeafPlanes();
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
WorldExtd_ct            World;
Vector                  g_vecZero(0,0,0);

int                     bGetMoveVol= 0;

bool_t                  bEnable_noclip;

   // move data
int                     BaseMoveSpeed;;
vec3_t                  MoveDeltaDist;
vec3_t                  ViewAngDRotation;
float                   MoveSpeed;
vec3_t                  MoveAng;
vec3_t                  CheckedPos;

void                    check_collisions(const vec3_t Src_i, const vec3_t Dir_i, vec3_t out);
//===================================================
//-------------------------------------------------
void parseCmdLine( int argc, char * argv[])
{
int i= 1;

while( i< argc)
	{
	if( strcmp( argv[i], "+map")==0 )
		{
		++i;
		strcpy( MapName, argv[i]);//overwrites possible input from bspview.cfg
		}
	else if( strcmp( argv[i], "+getmovvol")==0 )
		{
		bGetMoveVol= 1;
		}
	++i;
	}
}
//--------------------------------------------------
void Simul_init( int argc, char * argv[]) 
{
parseCmdLine( argc, argv);

    // initialize the random seed (for random spawn point)
srand((unsigned)time(NULL));
	
::World. loadBSPPlus( World_ct::K_LW_ForBspView);


MoveDeltaDist[0]=0; MoveDeltaDist[1]=0; MoveDeltaDist[2]=0;
ViewAngDRotation[0]=0; ViewAngDRotation[1]=0; ViewAngDRotation[2]=0;
MoveSpeed= 0;
MoveAng[0]= 0; MoveAng[1]= 0;MoveAng[2]= 0;
}
//-------------------------------------
void Cam_init()
{
float x, y, z;
x= spawn_point[0];
y= spawn_point[1];
z= spawn_point[2];
    // setup camera
Camera.move(x, y, z);
Camera.rotate(90- spawn_point_yaw, -90.0, 0.0);
}
//---------------------------------------
void Simul_update()
{
vec3_t PreviousPos_a;
VectorCopy( Camera. eye(), PreviousPos_a);
CollisionCutoff= 0;
check_collisions(PreviousPos_a, MoveDeltaDist, CheckedPos);
}
//---------------------------------------
void Cam_update()
{
Camera. move( CheckedPos);
Camera. rotate_delta( ViewAngDRotation);
Camera. transform();
}
/*
//--------------------------------------------------
void update0()
{
static bool was_inactive = FALSE;

SysTime_Update();
if (was_inactive)  // are we just waking up?
    {
    was_inactive= FALSE;
    // do we need to reset the mouse or do anything after waking up here?
    }

::MoveData. Speed= calcSpeed();
}
//--------------------------------------------------
void Cam_update0()
{
vec3_t eye_a;
VectorCopy( Camera. eye(), eye_a);
vec3_t dst_a;

::MoveData. Dir[ 0]= MoveData. Dx;  
::MoveData. Dir[ 1]= MoveData. Dy;  
::MoveData. Dir[ 2]= MoveData. Dz;

::MoveData. Rotation[ 1]= MoveData. pitch;
::MoveData. Rotation[ 2]= 0;//MoveData. roll;
::MoveData. Rotation[ 0]= MoveData. yaw;
cutoff=0;
check_collisions(eye_a, MoveData. Dir, dst_a);

Camera. move( dst_a);
Camera. rotate_delta( MoveData. Rotation);
::MoveData. Dx= 0;  
::MoveData. Dy= 0;  
::MoveData. Dz= 0;
::MoveData. pitch= 0;
::MoveData. roll= 0;
::MoveData. yaw= 0;

Camera. transform();
}
*/
//---------------------------------------------------
float calcSpeed()
{
float fps= frames_per_second;
if ((fps > 999.9) || (fps < 0.001))
       fps = 30;  // prevent overflow/underflow for bad fps values
if (fps > 30.0)
       fps = 30.0;  // clamp to 30 frames per second
    // movement speed is based on frame rate
float speed= 120.0f/ fps;
if(speed< 4.0f)
    speed= 4.0f;
speed= speed * BaseMoveSpeed;
return speed;
}
