// userinput.cpp by SamPlay
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <GL/glut.h>
#include "appgene.h"
#include "camera.h"
#include "WorldRender_ct.h"

#include "UserInput_ct.h"
#include "console_ct.h"
#include "parser.h"
#include "World_ct.h"
#include "bsptrace.h"

extern vec3_t MoveDeltaDist;
extern vec3_t ViewAngDRotation;
extern vec3_t MoveAng;
extern float  MoveSpeed;

extern Camera_ct Camera;
extern WorldRender_ct WorldRndr;
//extern BspMoveRender_ct BspMoveRndr;

extern UserInput_ct UserInputs;

extern int   ScreenCenterX;
extern int   ScreenCenterY;

extern float gametime;
extern float frames_per_second;

extern bool_t bEnable_noclip;
extern bool_t bSwitchableTextureOff;
extern bool_t bDrawTraceline;
extern char MapName[];

extern TraceData_st DispTraceData;
extern int TraceType;
extern vec3_t    AABoxHalfSize;
extern vec3_t    AABoxSize;

#include "DisplayOption_ct.h"
DisplayOption_ct DispOpt;

void resetMouseWin32();
float calcSpeed();
void saveEntites( char * );
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
int bSurfOnly=0;

#ifdef _WIN32
#include "uw_Win_VkCode.h"// win32 specific
#else
/*
#define VK_UP GLUT_KEY_UP
#define VK_DOWN GLUT_KEY_DOWN
#define VK_LEFT GLUT_KEY_RIGHT
#define VK_RIGHT GLUT_KEY_LEFT
*/
#endif


static const float piover180= 0.0174532925f;
static const float piover2= 1.570796327f;

Csl_ct Console;
char   Csl_ct:: ConsoleTxt[Csl_ct:: K_CslBufLen];
char   Csl_ct:: ConsoleMsg[Csl_ct:: K_CslBufLen];
char * Csl_ct:: pChar= NULL;
int    Csl_ct:: bConsoleMode= 0;

bool       UserInput_ct::keys[MAX_KEYS];
float      UserInput_ct:: keybounce_time= 0.0f;
float      UserInput_ct:: keybounce_delay= 0.25f;
float      UserInput_ct:: mouse_sensitivity;
bool       UserInput_ct:: bInverted_mouse;
int        UserInput_ct:: bLeft_button;
int        UserInput_ct:: bRight_button;
int        UserInput_ct:: MousePosX;
int        UserInput_ct:: MousePosY;
//================================================
//++++++++++++++++++++++++++++++++++++++++++++
class CslInterpret_ct: public Parser
{
public:
static char * pToInterpret;
static 	TokenList toklist;
static 	TokenFile tokfile;
static   void clear() { toklist. clear(); }
	virtual int interpret(const char *);
	virtual void init();
};
TokenList CslInterpret_ct:: toklist;
TokenFile CslInterpret_ct:: tokfile;
char * CslInterpret_ct:: pToInterpret='\0';
static CslInterpret_ct CslInterpret;
//-------------------------------------------------------
void Csl_ct:: init()
	{
	// Create a parser with the default operators (the default list is similar
	// to C++)
	CslInterpret.init();
	}
//-------------------------------------------
void CslInterpret_ct:: init()
	{
	Create();
	ReserveKeyword( "gotoleaf");
	ReserveKeyword( "saveents");
	ReserveKeyword( "sethullsize");
	}
//-------------------------------------------
int CslInterpret_ct:: interpret(const char * Txt_i)	
	{	
	if(!ProcessSourceTxt(Txt_i, &toklist))//$$
		{
//		MF_DBGOS2( KB_cout|KB_ods,<<"parser failed"<<'\n');
		return 1;
		}

	TokenList::iterator itr;
	itr = toklist.begin();

		Token & Token_a= *itr;// cout << toklist.GetDescriptiveString(itr).c_str() << endl;
		if( !Token_a. IsKeyword())
			{
//			MF_DBGOS2( KB_cout|KB_ods,<<"not a keyword"<<'\n');
			return 2;
			}
		if( strcmp( Token_a. GetKeyword(), "gotoleaf")==0)
			{
//			strcpy( Csl_ct::ConsoleMsg, "test succeeded");
			Token & Token2_a= *(++itr);
			if( !Token2_a.IsInteger())
				{
				strcpy( Csl_ct::ConsoleMsg, "not a leaf index type\n");
//				MF_DBGOS2( KB_cout|KB_ods,<<"not a face index type\n");
				return 3;
				}
			int i= Token2_a.GetInteger();
			if( i< 0)
				{
				strcpy( Csl_ct::ConsoleMsg, "no negative leaf index\n");
//				MF_DBGOS2( KB_cout|KB_ods,<<"not a face index "<< i<<'\n');
				return 1;
				}
			if( i>= WorldRndr. pHLBspFile-> numleafs)
				{
				strcpy( Csl_ct::ConsoleMsg, " leaf index out of range\n");
//				MF_DBGOS2( KB_cout|KB_ods,<<"face # out of range"<< i<<" max= "<< pFaces.size()-1<<'\n');
				return 4;
				}
			WorldRndr.FocusLeafId= i;
//			LeafBoundingPlane_st & BP_a= WorldRndr. pHLBspFile-> LeafsBoundingPlanes[ WorldRndr. FocusLeafId];
//			Camera. move( BP_a.Center);
			}
		else if(	strcmp( Token_a. GetKeyword(), "saveents")==0)
			{
			char FE_a[256];
			sprintf( FE_a,"%s_entities%s", MapName,".txt");
			saveEntites( FE_a);	
			}
		else if(	strcmp( Token_a. GetKeyword(), "sethullsize")==0)
			{
			float HalfSize_a[ 2];
			for( int i=0; i<2; i++)
				{
				Token & Token2_a= *(++itr);
				if( !Token2_a.IsReal() &&  !Token2_a.IsInteger())
					{
					strcpy( Csl_ct::ConsoleMsg, "not hull size\n");
					return 3;
					}
				if( Token2_a.IsInteger())
					HalfSize_a[ i]= Token2_a.GetInteger();
				else if( Token2_a.IsReal())	
					HalfSize_a[ i]= Token2_a.GetReal();
				}
			AABoxHalfSize[1]= AABoxHalfSize[0]=  HalfSize_a[ 0];
			AABoxHalfSize[2]= HalfSize_a[ 1];
			AABoxSize[1]= AABoxSize[0]=  2* AABoxHalfSize[0];
			AABoxSize[2]= 2* AABoxHalfSize[2];
			}
	
	return 0;
	}
//-------------------------------------------------------
void UserInput_ct:: init()
{
bLeft_button= 0;
bRight_button= 0;
MousePosX= 0;
MousePosY= 0;

for (int index = 0; index < MAX_KEYS; index++)
	keys[index] = 0;

Console.init();
}
//-------------------------------------------
void UserInput_ct:: update()
{
if( Csl_ct:: bConsoleMode)
	{
	char Key_a= 32;
	if( keys['\r'])
		{
		Console. get( '\r');
		CslInterpret. interpret( Csl_ct::ConsoleTxt);
		CslInterpret_ct::clear();
		keys['\r']= 0;
		return;
		}
	while( !keys[Key_a] && Key_a<= 126){ ++Key_a;}
	if( keys[Key_a])
		{
		if( Console. get( Key_a))
			{
//			cout<< "READ: "<< Csl_ct::ConsoleTxt<<'\n';
			CslInterpret. interpret( Csl_ct::ConsoleTxt);
			}
		keys[Key_a]= 0;
		}
	return;
	}

if( keys['0'] && (keybounce_time < gametime))
	{
	keybounce_time= gametime + keybounce_delay;
	Csl_ct:: bConsoleMode= 1;
	Console. in();
	keys['0']=0;
	return;
	}

HLBspFile_ct * pOld_a= DispOpt. pHLBspFileForInfo;

float Dx_a= 0.0f;
float Dy_a= 0.0f;
float Dz_a= 0.0f;

float DPitch_a= 0.0f;
float DRoll_a= 0.0f;
float DYaw_a= ( -MousePosX+ ScreenCenterX)* mouse_sensitivity;
if (bInverted_mouse)
   DPitch_a= ( MousePosY- ScreenCenterY)* mouse_sensitivity;
else
   DPitch_a= ( ScreenCenterY- MousePosY)* mouse_sensitivity;
#ifdef WIN32
resetMouseWin32();
#endif

float ViewAngPitchRad_a= (-Camera.pitch()-90)+ DPitch_a;
float ViewAngYawRad_a= 90- Camera.yaw()+ DYaw_a;
ViewAngPitchRad_a*= piover180;
ViewAngYawRad_a*= piover180;


float MoveAngPitchRad_a;
float MoveAngYawRad_a;
float Speed_a= 0;

if( keys[VK_UP])
    {
	 Speed_a= calcSpeed();
	 MoveAngYawRad_a= ViewAngYawRad_a;
	 MoveAngPitchRad_a= ViewAngPitchRad_a;
    }
if( keys[VK_DOWN])
    {
	 Speed_a= calcSpeed();
	 MoveAngPitchRad_a= -ViewAngPitchRad_a;
	 MoveAngYawRad_a= M_PI+ ViewAngYawRad_a;
    }

if( keys[VK_LEFT])
    {
	 Speed_a= calcSpeed();
	 MoveAngYawRad_a= ViewAngYawRad_a+ piover2;
	 MoveAngPitchRad_a= 0;
    }

if( keys[VK_RIGHT])
    {
	 Speed_a= calcSpeed();
	 MoveAngYawRad_a= ViewAngYawRad_a- piover2;
	 MoveAngPitchRad_a= 0;
    }

if( keys['1'] && (keybounce_time < gametime))
	{
	keybounce_time= gametime + keybounce_delay;
	WorldRndr.DrawEntities=(++WorldRndr.DrawEntities)%3;
	}
if( keys['2'] && (keybounce_time < gametime))
	{
	keybounce_time= gametime + keybounce_delay;
	WorldRndr.DrawEntities=2;
	WorldRndr.getNextEntity(WorldRndr.FocusEntIdx);
	}
if( keys['3'] && (keybounce_time < gametime))
	{
	keybounce_time= gametime + keybounce_delay;
	WorldRndr.bDrawSky= !WorldRndr.bDrawSky;
	}
if( keys['4'] && (keybounce_time < gametime))
	{
	keybounce_time= gametime + keybounce_delay;
	WorldRndr.bRenderSpecialTextures= !WorldRndr.bRenderSpecialTextures;
	}

/*
if( keys['B']&& (keybounce_time < gametime))
    {
 	 keybounce_time= gametime + keybounce_delay;
	 }
*/
if( keys['C'] && (keybounce_time < gametime))
	{
	keybounce_time= gametime + keybounce_delay;
	bEnable_noclip= !bEnable_noclip;
	}
#ifdef MI_HLBSPBUILD
if( keys['D'] && (keybounce_time < gametime)) 
	{
    keybounce_time = gametime + keybounce_delay;
	++DispOpt. DrawSurfModelIdx;
	}
#endif
if( keys['E'] && (keybounce_time < gametime))
    {
    keybounce_time = gametime + keybounce_delay;
    WorldRndr. show_edges++;
    if (WorldRndr. show_edges > 2)
       WorldRndr. show_edges = 0;
    }

#ifdef MI_HLBSPBUILD
if (keys['F'] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
 	 DispOpt. onChangeFace();
	 }
#endif

if( keys['L'] && (keybounce_time < gametime))
    {
    keybounce_time = gametime + keybounce_delay;
    WorldRndr. bEnable_lighting= !WorldRndr. bEnable_lighting;
    }

#ifdef M_DEBUGC
if( keys['M'] && (keybounce_time < gametime) && bDrawTraceline) 
	{
	keybounce_time= gametime + keybounce_delay;
	affDbgMsg(0,0);
	keys['M']= 0;
	}
#endif

if( keys['N'] && (keybounce_time < gametime) && bDrawTraceline) 
	{
	keybounce_time= gametime + keybounce_delay;
	bDrawTraceline = !bDrawTraceline;
	}

if( keys['O'] && (keybounce_time < gametime))
    {
    keybounce_time = gametime + keybounce_delay;
    bSwitchableTextureOff = !bSwitchableTextureOff;
    }

if (keys['S']) 
	{ 
	Speed_a= calcSpeed();
	MoveAngYawRad_a= 0;
	MoveAngPitchRad_a= piover2;
	}  // move straight up




if ((keys['T']) && (keybounce_time < gametime))
    {
	 keybounce_time= gametime + keybounce_delay;
	 WorldRndr. bContinuousTrace= 0;
	 WorldRndr. bTraceAvail= 0;
	 bDrawTraceline= 0;	
	 keys['T']= 0;
	 }

if ((keys['I']) && (keybounce_time < gametime))
    {
	 keybounce_time= gametime + keybounce_delay;
	 TraceType=(++TraceType)%Trace_ct::K_Traces;
	 keys['I']= 0;
	 }
if ((keys['Y']) && (keybounce_time < gametime))
    {
	 keybounce_time= gametime + keybounce_delay;
	 WorldRndr. bContinuousTrace= 0;
	 WorldRndr. getTextureInfo();
	 WorldRndr. bTraceAvail= TRUE;
	 keys['Y']= 0;
	 }
if ((keys['U']) && (keybounce_time < gametime))
    {
	 keybounce_time= gametime + keybounce_delay;
	 WorldRndr. bContinuousTrace= !WorldRndr. bContinuousTrace;
	 if( WorldRndr. bContinuousTrace)
		 WorldRndr. bTraceAvail= TRUE;
	 keys['U']= 0;
	 }

//if( WorldRndr. bContinuousTrace)
//	WorldRndr. getTextureInfo();


if (keys['W']) 
	{	
	Speed_a= calcSpeed();
	MoveAngYawRad_a= 0;
	MoveAngPitchRad_a= -piover2;
	}  //  move straight down

if (keys['X'] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    WorldRndr. bLockCurrLeaf^= 1;
	 }

#ifdef MI_HLBSPBUILD
if (keys['Z'] && (keybounce_time < gametime))
    {
 	 keybounce_time= gametime + keybounce_delay;
	 DispOpt. onChangeSurf();
	 }
#endif

if (keys[VK_F1] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    WorldRndr. bDraw^= 1;
	 }
/*
if (keys[VK_F2] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    BspMoveRndr. bDraw^= 1;
	 }
*/
#ifdef MI_HLBSPBUILD
if (keys[VK_F3] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    DispOpt. bDrawAllModelSurfs^= 1;
	 }
if (keys[VK_F4] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    DispOpt. bDrawSurf^= 1;
	 if( DispOpt. bDrawSurf== 1)
		onChangeSurf()
	 }
if (keys[VK_F5] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    DispOpt. bDrawFace^= 1;
	 }
#endif

if (keys[VK_F6] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    if(WorldRndr. FaceDisplayFlags & KB_FD_Walkable)
		WorldRndr. FaceDisplayFlags&= ~KB_FD_Walkable;
	 else
		WorldRndr. FaceDisplayFlags|= KB_FD_Walkable;
	 }

if (keys[VK_F7] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    WorldRndr. bDrawFocusLeafBounds^= 1;
	 }
if (keys[VK_F8] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    WorldRndr. bDrawFocusLeafBoundEdges^= 1;
	 }

if (keys[VK_F9] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    WorldRndr. bDrawFocusLeafEdges^= 1;
	 }

if (keys[VK_F11] && (keybounce_time < gametime))
    {
    keybounce_time= gametime + keybounce_delay;
    WorldRndr. bDrawFocusLeafFaces^= 1;
	 }

if( Speed_a!= 0.0f)
	{
   float SpeedCos_a= Speed_a* fcos(MoveAngPitchRad_a);
	Dx_a= fcos(MoveAngYawRad_a)* SpeedCos_a;
   Dy_a= fsin(MoveAngYawRad_a)* SpeedCos_a;
   Dz_a= fsin(MoveAngPitchRad_a)* Speed_a;
		// if speed== 0=> previous values unchanged
	MoveAng[ 0]= MoveAngYawRad_a;
	MoveAng[ 1]= MoveAngPitchRad_a;
	MoveAng[ 2]= 0;
	}

MoveSpeed= Speed_a;
MoveDeltaDist[0]= Dx_a;
MoveDeltaDist[1]= Dy_a;
MoveDeltaDist[2]= Dz_a;
ViewAngDRotation[ 0]= DYaw_a;
ViewAngDRotation[ 1]= DPitch_a;
ViewAngDRotation[ 2]= DRoll_a;

/*
if( BspMoveRndr. bDraw && BspMoveRndr.pHLBspFile)
	DispOpt. pHLBspFileForInfo= BspMoveRndr.pHLBspFile;
else
*/
	DispOpt. pHLBspFileForInfo= WorldRndr.pHLBspFile;

if( DispOpt. pHLBspFileForInfo!= pOld_a)
	{	WorldRndr.FocusLeafId= -1;
	}
}
