#ifdef _WIN32
#pragma once
#endif
#ifndef VectorEXTENDED_H_
#define VectorEXTENDED_H_
//==============================================
#ifndef real_t
#define MD_REAL
typedef float real_t;
#endif

template <typename Vec3_T1, typename Vec3_T2> 
int compareBox( const Vec3_T1 & V1_i, const Vec3_T2 & V2_i, float Epsilon_i)
{
return ( (fabs( V1_i[0]- V2_i[0])< Epsilon_i )
			&& (fabs(V1_i[1]- V2_i[1])< Epsilon_i)
			&& (fabs(V1_i[2]- V2_i[2])< Epsilon_i));
}

template <typename Vec3_T> 
inline real_t dotProduct(const Vec3_T& a, const Vec3_T& b) 
	{ return( a.x* b.x+ a.y* b.y+ a.z* b.z); 
	}

template <typename Vec_T> 
inline real_t dotProductXY(const Vec_T& a, const Vec_T& b) 
	{ return( a.x* b.x+ a.y* b.y); 
	}

template <typename Vec3_T> inline Vec3_T memberProd ( const Vec3_T& a, const Vec3_T& b)
	{	return Vec3_T( a.x* b.x, a.y* b.y, a.z* b.z);
	} 

template <typename Vec3_T> inline void memberProd ( const Vec3_T& a, const Vec3_T& b,  Vec3_T & Result_a)
	{
	Result_a. x= a.x* b.x;
	Result_a. y= a.y* b.y;
	Result_a. z= a.z* b.z;
	} 
/*
template <typename T> 
inline Vec3_T crossProduct(const Vec3_T& a, const Vec3_T& b) 
	{ return Vec3_T( a.y* b.z- a.z* b.y, a.z* b.x- a.x* b.z, a.x* b.y- a.y* b.x ); 
	}
*/
template <typename Vec3_T> 
inline real_t crossProductZ(const Vec3_T& a, const Vec3_T& b) 
	{ return a.x* b.y- a.y* b.x; 
	}

template <typename Vec3_T> 
inline int normalize(Vec3_T & V_i)
	{
	real_t Len_a = V_i. Length();
	if (Len_a== 0) 
		return 1;
	Len_a= 1/ Len_a;
	V_i.x*= Len_a; 
	V_i.y*= Len_a; 
	V_i.z*= Len_a;
	return 0;//ie zero pb
	}
template <typename Vec3_T> inline real_t normalizeL( Vec3_T & V_i)
	{
	real_t Len_a = V_i. Length();
	if (Len_a== 0) 
		return 0;
	real_t InvLen_a= 1/ Len_a;
	V_i.x*= InvLen_a; 
	V_i.y*= InvLen_a; 
	V_i.z*= InvLen_a;
	return Len_a;
	}

template <typename Vec_T> inline int normalizeXY ( Vec_T & V_i )
	{
	real_t Len_a = V_i. Length();
	if ( Len_a == 0 )
		return 1;
	Len_a = 1/ Len_a;
	V_i.x*= Len_a; 
	V_i.y*= Len_a;
	return 0;//ie zero pb
	}

template <typename Vec_T> 
inline real_t lengthXYPow2( const Vec_T & V_i)
	{	return (V_i.x*V_i.x+V_i.y*V_i.y);
	}
template <typename Vec_T> 
inline real_t lengthXY( const Vec_T & V_i)
	{	return (real_t)sqrt(V_i.x*V_i.x+V_i.y*V_i.y);
	}

template <typename Vec3_T> 
inline real_t lengthPow2( const Vec3_T & V_i)
	{	return (V_i.x*V_i.x+V_i.y*V_i.y+V_i.z*V_i.z);
	}
template <typename Vec3_T> 
inline real_t length( const Vec3_T & V_i)
	{	return (real_t)sqrt(V_i.x*V_i.x+V_i.y*V_i.y+V_i.z*V_i.z);
	}

template <typename Vec3_T> 
inline real_t distPow2( const Vec3_T & Point0_i, const Vec3_T & Point1_i)
	{	return (Point1_i. x- Point0_i. x)*(Point1_i. x- Point0_i. x)
		+ (Point1_i. y- Point0_i. y)*(Point1_i. y- Point0_i. y) 
		+ (Point1_i. z- Point0_i. z)*(Point1_i. z- Point0_i. z);
	}
template <typename Vec3_T> 
inline real_t dist( const Vec3_T & Point0_i, const Vec3_T & Point1_i)
	{	return (real_t)sqrt(distPow2( Point0_i, Point1_i));
	}

template <typename Vec3_T> 
inline Vec3_T & changeSign (Vec3_T & V_i) 
	{ 
	V_i.x= -V_i.x; 
	V_i.y= -V_i.y; 
	V_i.z= -V_i.z; 
	return V_i; 
	}


/*
	//same as vec3t
template <typename Vec3T> int VectorCompare(const Vec3T & v1, const Vec3T & v2)
{
	int		i;
	for (i=0 ; i<3 ; i++)
		if (fabs(v1[i]-v2[i]) > EQUAL_EPSILON)
			return false;
	return true;
}
*/

//mathlib lookalike ( do not use x,y,z)
#ifdef MD_VECTORTPL
template <typename T> 
inline void VectorSubtract (const Vector3d_cT<T> & va , const  Vector3d_cT<T> & vb, Vector3d_cT<T> & out)
	{
	out[0]= va[0]- vb[0];
	out[1]= va[1]- vb[1];
	out[2]= va[2]- vb[2];
	}
template <typename T> 
inline void VectorAdd (const Vector3d_cT<T> & va , const Vector3d_cT<T> & vb, Vector3d_cT<T> & out)
	{
	out[0]= va[0]+ vb[0];
	out[1]= va[1]+ vb[1];
	out[2]= va[2]+ vb[2];
	}

template <typename T>  
inline void VectorChangeSign (Vector3d_cT<T> & v)
{
	v[0]= -v[0];
	v[1]= -v[1];
	v[2]= -v[2];
}

template <typename T> 
inline void VectorCopy (const Vector3d_cT<T> & in, Vector3d_cT<T> & out)
	{
	out[0]= in[0];
	out[1]= in[1];
	out[2]= in[2];
	}
template <typename T1,typename T2> 
inline void VectorScale (const Vector3d_cT<T1> & v, real_t scale, Vector3d_cT<T2> & out)
	{
	out[0]= (T2) (v[0] * scale);
	out[1]= (T2) (v[1] * scale);
	out[2]= (T2) (v[2] * scale);
	}

template <typename T>  
void CrossProduct (const Vector3d_cT<T> & v1, const Vector3d_cT<T> & v2, Vector3d_cT<T> & cross)
	{
	cross[0] = v1[1]*v2[2] - v1[2]*v2[1];
	cross[1] = v1[2]*v2[0] - v1[0]*v2[2];
	cross[2] = v1[0]*v2[1] - v1[1]*v2[0];
	}

/*
template <typename T1, typename T2> 
inline double dotProductd ( const Vector3d_cT<T1> & v1, const Vector3d_cT<T2> & v2)
	{	return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
	}
*/
#else

void CrossProduct (const Vector & v1, const Vector & v2, Vector & cross)
	{
	cross[0] = v1[1]*v2[2] - v1[2]*v2[1];
	cross[1] = v1[2]*v2[0] - v1[0]*v2[2];
	cross[2] = v1[0]*v2[1] - v1[1]*v2[0];
	}
#endif //MD_VECTORTPL


//#define VectorAvg(a) ( ( (a)[0] + (a)[1] + (a)[2] ) / 3 )

/*
#define VectorSubtract _VectorSubtract
#define VectorAdd _VectorAdd
#define VectorCopy  M_VectorCopy
#define VectorScale _VectorScale
#define DotProduct _DotProduct
#define VectorFill M_VectorFill
#define VectorInverse VectorChangeSign

#define M_VectorFill(a,b) { (a)[0]=(b); (a)[1]=(b); (a)[2]=(b);}
#define M_VectorCopy(a,b) { b[0]= a[0]; b[1]= a[1]; b[2]= a[2]; }

template <typename Vec3T1, typename Vec3T2> inline double _DotProduct( const Vec3T1 & v1, const Vec3T2 & v2)
	{	return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
	}

template <typename Vec3T> 
inline void _VectorSubtract (const Vec3T & va, const Vec3T & vb, Vec3T & out)
	{
	out[0] = va[0]-vb[0];
	out[1] = va[1]-vb[1];
	out[2] = va[2]-vb[2];
	}

template <typename Vec3T> inline void 
_VectorAdd (const Vec3T & va, const Vec3T & vb, Vec3T & out)
	{
	out[0] = va[0]+vb[0];
	out[1] = va[1]+vb[1];
	out[2] = va[2]+vb[2];
	}

template <typename Vec3T> 
inline void _VectorCopy (const Vec3T & in, Vec3T & out)
	{
	out[0] = in[0];
	out[1] = in[1];
	out[2] = in[2];
	}

template <typename Vec3T1,typename Vec3T2> 
inline void _VectorScale (const Vec3T1 & v, double scale, Vec3T2 & out)
	{
	out[0] = (v[0] * scale);
	out[1] = (v[1] * scale);
	out[2] = (v[2] * scale);
	}

template <typename Vec3T>  
double Vector2DLength(const Vec3T & v)
{
   int   i;
   double length;

   length = 0;
   for (i=0 ; i< 2 ; i++)  // ignore the Z axis
      length += v[i]*v[i];
   length = (float)sqrt(length); // FIXME

   return length;
}

template <typename Vec3T>  
double VectorLength(const Vec3T & v)
{
	int		i;
	double	length;
	
	length = 0;
	for (i=0 ; i< 3 ; i++)
		length += v[i]*v[i];
	length = sqrt (length);		// FIXME

	return length;
}

template <typename Vec3T>  
inline void VectorMA (const Vec3T & va, double scale, const Vec3T & vb, Vec3T & vc)
{
	vc[0] = va[0] + scale*vb[0];
	vc[1] = va[1] + scale*vb[1];
	vc[2] = va[2] + scale*vb[2];
}

template <typename Vec3T>  
inline void VectorInterp (const Vec3T & vstart, const Vec3T & vend, double scale, Vec3T & vc)
{
	vc[0] = vstart[0] + scale*(vend[0]-vstart[0]);
	vc[1] = vstart[1] + scale*(vend[1]-vstart[1]);
	vc[2] = vstart[2] + scale*(vend[2]-vstart[2]);
}

template <typename Vec3T>  
void CrossProduct (const Vec3T & v1, const Vec3T & v2, Vec3T & cross)
{
	cross[0] = v1[1]*v2[2] - v1[2]*v2[1];
	cross[1] = v1[2]*v2[0] - v1[0]*v2[2];
	cross[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

template <typename Vec3T>  
Vec3T VectorNormalize (Vec3T & v)
{
	int		i;
	double	length;

	if ( fabs(v[1] - 0.000215956) < 0.0001)
		i=1;

	length = 0;
	for (i=0 ; i< 3 ; i++)
		length += v[i]*v[i];
	length = sqrt (length);
	if (length == 0)
		return 0;
		
	for (i=0 ; i< 3 ; i++)
		v[i]= (Vec3T)(v[i]/length);	

	return length;
}

template <typename Vec3T>  
inline void VectorChangeSign (Vec3T & v)
{
	v[0] = -v[0];
	v[1] = -v[1];
	v[2] = -v[2];
}
*/
#endif //eof