/***
*
*	Copyright (c) 1996-2002, Valve LLC. All rights reserved.
*	
*	This product contains software technology licensed from Id 
*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*	All Rights Reserved.
*
****/
//qbsp.c modified by SamPlay

#include "appgene.h"
#include "bsp_ext.h"
#include "u_RzbVector_ctpl.h"
extern float WorldSizeMax;
//===========================================================
//-----------------------------------------------------------
Winding_ct * Winding_ct:: BaseWindingForPlane (dplane_t *p)
	{
	Winding_ct::Vec3_t	org, vright, vup, temp;
	Winding_ct	*w;

	VectorScale (p->normal, (float)p->dist, org);

	VectorFill(vup,0);
	VectorFill(vright,0);
	if (!p->normal[1] && !p->normal[2])
		{
		vup[2] = WorldSizeMax;
		vright[1] = WorldSizeMax*p->normal[0];
		}
	else if (!p->normal[0] && !p->normal[2])
		{
		vup[2] = WorldSizeMax;
		vright[0] = -WorldSizeMax*p->normal[1];
		}
	else if (!p->normal[0] && !p->normal[1])
		{
		vup[1] = WorldSizeMax;
		vright[0] = WorldSizeMax*p->normal[2];
		}
	else
		{
		vup[0] = -2*p->normal[1]*p->normal[2];
		vup[1] = p->normal[0]*p->normal[2];
		vup[2] = p->normal[0]*p->normal[1];

		VectorNormalize (vup);

		VectorCopy (p->normal, temp);
		CrossProduct (vup, temp, vright);

		VectorScale (vup, WorldSizeMax, vup);
		VectorScale (vright, WorldSizeMax, vright);
		}

// project a really big	axis aligned box onto the plane
	w = new Winding_ct(4);
	
	VectorSubtract (org, vright, w->points[0]);
	VectorAdd (w->points[0], vup, w->points[0]);
	
	VectorAdd (org, vright, w->points[1]);
	VectorAdd (w->points[1], vup, w->points[1]);
	
	VectorAdd (org, vright, w->points[2]);
	VectorSubtract (w->points[2], vup, w->points[2]);
	
	VectorSubtract (org, vright, w->points[3]);
	VectorSubtract (w->points[3], vup, w->points[3]);
	
	return w;	
	}
//-----------------------------------------------
Winding_ct	*Winding_ct:: CopyWinding (Winding_ct *w)
	{
	Winding_ct	*c= new Winding_ct( w->numpoints);
	memcpy (c->points, w->points, w->numpoints* sizeof(Vec3_t));
	return c;
	}
//--------------------------------------------------
/*
Clips the winding to the plane, returning the new winding on the positive side
Frees the input winding.
If keepon is true, an exactly on-plane winding will be saved, otherwise
it will be clipped away.
*/
void Winding_ct::clip ( dplane_t *split, int keepon)
	{
	RszVector_cT<double,1>	dists( 1+numpoints);
	RszVector_cT<int,1>	sides( 1+numpoints);
	int		counts[3];
	double	dot;
	int		i, j;
	double	*p1, *p2;
	Vec3_t	mid;
	int		maxpts;
	
	counts[0] = counts[1] = counts[2] = 0;
	// determine sides for each point
	// do this exactly, with no epsilon so tiny portals still work
	for (i=0 ; i<numpoints ; i++)
		{
		dot= DotProduct (points[i], split->normal)- split->dist;
		dists[i] = dot;
		if (dot > 0)
			sides[i] = SIDE_FRONT;
		else if (dot < 0)
			sides[i] = SIDE_BACK;
		else
			sides[i] = SIDE_ON;
		counts[sides[i]]++;
		}
	sides[i]= sides[0];
	dists[i]= dists[0];
	
	if (keepon && !counts[0] && !counts[1])
		{
		return;
		}
	if (!counts[0])
		{
		clear();
		return;
		}
	if (!counts[1])
		{
		return;
		}
	maxpts= numpoints+4;	// can't use counts[0]+2 because of fp grouping errors
	Vec3_t * NewPoints_n= new Vec3_t[maxpts];
	int NewNumPoints= 0;
	for (i=0 ; i<numpoints ; i++)
		{
		p1= points[i];
		
		if (sides[i] == SIDE_ON)
			{
			VectorCopy (p1, NewPoints_n[NewNumPoints]);
			NewNumPoints++;
			continue;
			}
	
		if (sides[i] == SIDE_FRONT)
			{
			VectorCopy (p1, NewPoints_n[NewNumPoints]);
			NewNumPoints++;
			}
		
		if (sides[i+1] == SIDE_ON || sides[i+1] == sides[i])
			continue;
			
	// generate a split point
		p2 = points[(i+1)%numpoints];
		dot = dists[i] / (dists[i]-dists[i+1]);
		for (j=0 ; j<3 ; j++)
			{	// avoid round off error when possible
			if (split->normal[j] == 1)
				mid[j] = split->dist;
			else if (split->normal[j] == -1)
				mid[j] = -split->dist;
			else
				mid[j] = p1[j] + dot*(p2[j]-p1[j]);
			}
		VectorCopy (mid, NewPoints_n[NewNumPoints]);
		NewNumPoints++;
		}
	
	clear();
	points= NewPoints_n;
	numpoints= NewNumPoints;
	return;
	}
/*
==================
Divides a winding by a plane, producing one or two windings.  The
original winding is not damaged or freed.  If only on one side, the
returned winding will be the input winding.  If on both sides, two
new windings will be created.
==================
*/
//#define	DIVIDE_EPSILON	0.5
#define DIVIDE_EPSILON	ON_EPSILON

void	Winding_ct::DivideWinding (Winding_ct *in, dplane_t *split, Winding_ct **front, Winding_ct **back)
	{
	RszVector_cT<double,1>	dists( 1+in-> numpoints);
	RszVector_cT<int,1>	sides( 1+in-> numpoints);

	int		counts[3];
	double	dot;
	int		i, j;
	Coord_t *p1, *p2;
	Vec3_t	mid;
	int		maxpts;
	
	counts[0] = counts[1] = counts[2] = 0;

// determine sides for each point
	for (i=0 ; i<in->numpoints ; i++)
		{
		dot = DotProduct (in->points[i], split->normal)- split->dist;
		dists[i]= dot;
		if (dot> DIVIDE_EPSILON)
			sides[i] = SIDE_FRONT;
		else if (dot< -DIVIDE_EPSILON)
			sides[i]= SIDE_BACK;
		else
			sides[i]= SIDE_ON;
		counts[sides[i]]++;
		}
	sides[i]= sides[0];
	dists[i]= dists[0];
	
	*front = *back = NULL;

	if (!counts[0])
		{
		*back = in;
		return;
		}
	if (!counts[1])
		{
		*front = in;
		return;
		}

	maxpts= in->numpoints+4;// can't use counts[0]+2 because of fp grouping errors
	Vec3_t * FPoints= new Vec3_t[maxpts];
	Vec3_t * BPoints= new Vec3_t[maxpts];
	int fnumpoints=0;
	int bnumpoints=0;
		
	for (i=0 ; i<in->numpoints ; i++)
		{
		p1= in->points[i];
		if (sides[i]== SIDE_ON)
			{
			VectorCopy (p1, FPoints[fnumpoints]); fnumpoints++;
			VectorCopy (p1, BPoints[bnumpoints]); bnumpoints++;
			continue;
			}
	
		if (sides[i] == SIDE_FRONT)
			{
			VectorCopy (p1, FPoints[fnumpoints]); 
			fnumpoints++;
			}
		if (sides[i] == SIDE_BACK)
			{
			VectorCopy (p1, BPoints[bnumpoints]);
			bnumpoints++;
			}

		if(sides[i+1]== SIDE_ON || sides[i+1] == sides[i])
			continue;
			
	// generate a split point
		p2 = in->points[(i+1)%in->numpoints];
		
		dot = dists[i] / (dists[i]-dists[i+1]);
		for (j=0 ; j<3 ; j++)
			{	// avoid round off error when possible
			if (split->normal[j] == 1)
				mid[j] = split->dist;
			else if (split->normal[j] == -1)
				mid[j] = -split->dist;
			else
				mid[j] = p1[j] + dot*(p2[j]-p1[j]);
			}
		VectorCopy (mid, FPoints[fnumpoints]); 
		fnumpoints++;
		VectorCopy (mid, BPoints[bnumpoints]); 
		bnumpoints++;
		}

	*front= new Winding_ct(fnumpoints);
	memcpy( (*front)-> points, FPoints, fnumpoints*sizeof(Vec3_t));
	delete [] FPoints;
	*back= new Winding_ct(bnumpoints);
	memcpy( (*back)-> points, BPoints, bnumpoints*sizeof(Vec3_t));
	delete [] BPoints;
	}
//---------------------------------------------------
void	Winding_ct::calcCenter (Winding_ct::Vec3_t center)
	{
	int		i;
	float	scale;

	VectorFill(center,0);
	for (i=0 ; i<numpoints ; i++)
		VectorAdd (points[i], center, center);

	scale = 1.0f/numpoints;
	VectorScale (center, scale, center);
	}
//---------------------------------------------------
float Winding_ct:: calcArea ()
{
	int		i;
	float	total;
	vecd3_t	v1, v2, cross;
	
	total = 0;
	for (i=1 ; i<numpoints ; i++)
		{
		VectorSubtract (points[i], points[0], v1);
		VectorSubtract (points[i+1], points[0], v2);
		CrossProduct (v1, v2, cross);
		total+= (float)VectorLength (cross);
		}
	return total*0.5f;
}
//------------------------------------------
void Winding_ct:: calcPlane( dplane_t *plane)
{
	vecd3_t		v1, v2;
	vecd3_t		plane_normal;

// calc plane
	VectorSubtract (points[2], points[1], v1);
	VectorSubtract (points[0], points[1], v2);

	CrossProduct (v2, v1, plane_normal);
	VectorNormalize (plane_normal);
	VectorCopy( plane_normal, plane->normal);
	plane->dist= DotProduct (points[0], plane->normal);
}

