//
// BSP_tool - botman's Half-Life BSP utilities
//
// (http://planethalflife.com/botman/)
//
// render.cpp
//
// Copyright (C) 2001 - Jeffrey "botman" Broome
//
/* 
 * HL rendering engine
 * Copyright (c) 2000,2001 Bart Sekura
 *
 * Permission to use, copy, modify and distribute this software
 * is hereby granted, provided that both the copyright notice and 
 * this permission notice appear in all copies of the software, 
 * derivative works or modified versions.
 *
 * THE AUTHOR ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION AND DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES 
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * OpenGL window
 */
//
// Parts of this code is from:
//
//                  POLY ENGINE version 0.79
//                Written by Alexey Goloshubin
//
// Parts of this code is from:
//
// WinQuake, GLQuake, QuakeWorld, and GLQuakeWorld
// Copyright (C) 1996-1997 Id Software, Inc.
//
// modified  by SamPlay
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <GL/glut.h>
#include "appGene.h"
#include "WorldRender_ct.h"
#include "matrix.h"
//#include "bspfile2.h"
#include "gr_util3DObjects_ct.h"
#include "bsp_ext.h"
#include "camera.h"
#include "sys_time.h"
#include "entity.h"
#include "world_ct.h"
#include "bsptrace.h"

#include "fastmath.h"
#include "MoveData_st.h"

extern Camera_ct    Camera;
extern WorldExtd_ct World;

extern int num_textures;
extern Texture_ct *textures[];
extern Texture_ct *sky_textures[];

extern HLBspFileExtsn_ct BspExtsn;

extern float gametime;
extern float frames_per_second;
extern  MoveData_st MoveData;
extern bool_t bSwitchableTextureOff;
extern float WorldHalfSizeMax;

#include "bsp_ext.h"
#include "HalfLifeMods_ct.h"

extern HLBspFileExtsn_ct BspExtsn;

void drawSurfs();
int frustum_checkCull(const short* mins, const short* maxs);
void drawTraceline(vec3_t traceline_pt[3]);
char * getTextureName( HLBspFile_ct * pHLBspFile_i, dface_t * pFace_i);
void ExtractFileBase (char *path, char *dest);//cmdlib.h
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
int num_models;
char model_classname[ MAX_MODELS][ 64];
char model_filename[ MAX_MODELS][ 64];
StudioModel * pStudioModels[ MAX_MODELS];

bool_t  bDrawTraceline = FALSE;
TraceData_st DispTraceData;
Trace_ct		 DispTrace;
int TraceType= Trace_ct::K_Point;

float    brightness;
// studio model stuff..
vec3_t g_vright;  // needs to be set to viewer's right in order for chrome to work
float  g_lambert = 1.5;

// statistics
int   rendered;


void      calcSurfaceAABox (const HLBspFile_ct & Bsp_i,  dface_t *s, float *mins, float *maxs);
void      ComputeLightmap(const HLBspFile_ct & Bsp_i,  dface_t *dface, RenderFace_ct *face, float *mins, float *maxs, int *lh, int *lw);
void      drawCompl();
//iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii
#define SKY_DEPTH  4000.0f

#define SKY_FRONT  (1<<0)
#define SKY_BACK   (1<<1)
#define SKY_LEFT   (1<<2)
#define SKY_RIGHT  (1<<3)
#define SKY_UP     (1<<4)
#define SKY_DOWN   (1<<5)

static St_array_ct st_array;
static Lst_array_ct lst_array;
static int current_style = 0;

//void drawTranspPane( vec3_t Coords_i[4], float R, float G, float B, int bRev_i);

static  int bDrawHull= 0;

int   GrSphereNumOfSectors= 8;
int   GrSphereNumOfSlices= 6;
int   GrCylNumOfSectors= 8;
float GrHullLineWidth= 3.0f;
vec3_t    AABoxHalfSize={ 18, 18, 36};
vec3_t    AABoxSize={ 36, 36, 72};

static CylinderModel_ct CylinderModel_s( GrCylNumOfSectors, 5.0f/*radius*/, 1.0f/*length*/, 1/*drawcirclebase (useless if wireframe)*/);
//=====================================================
//-----------------------------------------------------------
RenderFace_ct:: RenderFace_ct() : texture(-1), lightmap(0), next(0) 
	{
   for(int i= 0; i < MAXLIGHTMAPS;i++) 
		{
      styles[i]= -1;
      lightmaps[i]= NULL;
		}
	}
//-----------------------------------------------------------
RenderFace_ct:: ~RenderFace_ct() 
	{ 
   for(int i= 0; i < MAXLIGHTMAPS; i++) 
		{
      if(lightmaps[i]) 
			{
         delete lightmaps[i];
         lightmaps[i]= NULL;
         styles[i]= -1;
         }
		}
	}
//-----------------------------------------------------------
WorldRender_ct::WorldRender_ct()
{
pHLBspFile= NULL;

bDraw= 1;
DrawEntities= 1;//all
FocusEntIdx= 0;

bDrawSky= FALSE;
bDrawSpecTextures= TRUE;

bDrawFocusLeafEdges= FALSE;
bDrawFocusLeafBounds= FALSE;
bDrawFocusLeafBoundEdges= FALSE;
DrawBoundFlags= /*KB_DP_Solid|*/ KB_DP_NotSolid;

bEnable_lighting = TRUE;
show_edges= 0;
bRenderSpecialTextures=TRUE;
bLockCurrLeaf= FALSE;

FaceDisplayFlags= KB_FD_Visible;

bDrawTraceline = FALSE;
bTraceAvail= FALSE;
bContinuousTrace= FALSE;

FocusLeafId=-1;

faces= NULL;  // pointer to array of RenderFace_ct pointers
skyfaces= NULL;

// visible faces by texture index
VisibleSameTextureFaceLists= NULL;  // pointer to array of RenderFace_ct pointers
max_visible_faces=0;

MarkedFacesBitsets= NULL;

for (int index = 0; index < num_models; index++)
   {
   if (::pStudioModels[index])
     {
	  delete ::pStudioModels[index];
		::pStudioModels[index]=0;
		}
	}
}
//------------------------------------------------------
WorldRender_ct::~WorldRender_ct()
{
clear();
}
//--------------------------------------------------
void WorldRender_ct:: clear()
{
Visibility. clear();
delete[] faces;
delete[] MarkedFacesBitsets;
delete[] VisibleSameTextureFaceLists;
freeTextures();
}
//--------------------------------------------------
void WorldRender_ct:: reset()
{
clear();
faces = NULL;
MarkedFacesBitsets = NULL;
VisibleSameTextureFaceLists= NULL;
}
//-----------------------------------------------------
void WorldRender_ct:: init()
{
int index;
int ent_index= -1;

if ((ent_index= findEntityByClassname(pHLBspFile, ent_index, "worldspawn"))!= -1)
    {
    char *key_value;
    char value[MAX_VALUE];
    char *pos, *term;
    char wadfilename[64];
    key_value = ValueForKey(&pHLBspFile-> entities[ent_index], "wad");
    if (key_value[0])
        {
        strcpy(value, key_value);
        pos= value;
        while ((term = strstr(pos, ";")) != NULL)
            {
            *term= 0;  // change ; to null terminator
            ExtractFileBase(pos, wadfilename);
            strcat(wadfilename, ".wad");
            LoadWadFile(wadfilename);
            pos= term + 1;
            }
        ExtractFileBase(pos, wadfilename);
        strcat(wadfilename, ".wad");
        LoadWadFile(wadfilename);
        }
    else
        Error("Can't find WAD key for worldspawn entity!\n");
    }
else
    Error("Can't find worldspawn entity!\n");


for(index = 0; index < MAX_MODELS; index++)
   pStudioModels[index] = NULL;

initSpawnPoint(pHLBspFile);

    // load the textures...
loadTextures(pHLBspFile);

max_visible_faces = num_textures;

    // load the models...
for(index = 0; index < num_entvars; index++)
    {
    // check if this classname needs a studio model...
    for (int model_index = 0; model_index < num_models; model_index++)
        {
         if (stricmp(entvars[index].classname, model_classname[model_index]) == 0)
            {
            // has the studio model not be allocated for this classname yet?
            if (pStudioModels[model_index] == NULL)
                {
                // load the studio model for this entity...
                StudioModel *pStudioModel_n = new StudioModel;
                if (pStudioModel_n-> Init(model_filename[model_index]))
                    {
                    pStudioModel_n-> SetController(0, 0.0f);
                    pStudioModel_n-> SetController(1, 0.0f);
                    pStudioModel_n-> SetController(2, 0.0f);
                    pStudioModel_n-> SetController(3, 0.0f);

                    pStudioModel_n-> SetMouth(0);
                    pStudioModel_n-> SetBodygroup(0, 0);
                    pStudioModel_n-> SetSkin(0);

                    pStudioModel_n-> SetSequence(0);
                    ::pStudioModels[model_index] = pStudioModel_n;
                    }
                else
                    {
                    delete pStudioModel_n;  //delete the unused studio model
                    }
                }
            // all entities of the same classname share the same studio model...
            entvars[index].studio_model= ::pStudioModels[model_index];
            }
        }
    }

Visibility. pHLBspFile= pHLBspFile;
Visibility. Decompress_VIS();

	//render specific faces
faces= new RenderFace_ct* [pHLBspFile-> numfaces];
if(!faces)
    Error("Error allocating memory for faces!\n");
    // setup textures, faces and lightmaps...
for (int i= 0; i< pHLBspFile-> numfaces; i++)
    {
    RenderFace_ct *face = new RenderFace_ct;
    if( !face)
         Error("Error allocating memory for a face!\n");

    // get plane info
    memcpy(&face->plane, &pHLBspFile-> dplanes[pHLBspFile-> dfaces[i].planenum], sizeof(dplane_t));
    face-> side= pHLBspFile-> dfaces[i]. side;

    // determine best primitive type
    switch(pHLBspFile-> dfaces[i]. numedges) 
		  {
        case 3: face->type= GL_TRIANGLES; break;
        case 4: face->type= GL_QUADS; break;
        default: face->type= GL_POLYGON;
        }

        // get texture index
    texinfo_t *ti= pHLBspFile-> texinfo+ pHLBspFile-> dfaces[i].texinfo;
    face-> texture= ti->miptex; // texture index
    face-> flags= ti->flags;
        //
        // the following computations are based on:
        // PolyEngine (c) Alexey Goloshubin and Quake I source by id Software
        //
        
    int EdgeIdx_a;
	 float min[2], max[2];
    int lw,lh;

        // compute s and t extents
    ::calcSurfaceAABox( *pHLBspFile, &pHLBspFile-> dfaces[i], min, max);

    face->lightmap = NULL;
    for (int ndx = 0; ndx < MAXLIGHTMAPS; ndx++)
        face->lightmaps[ndx] = NULL;

    if (!(face->flags & TEX_SPECIAL))
       ::ComputeLightmap(*pHLBspFile, &pHLBspFile-> dfaces[i], face, min, max, &lh, &lw);
    //////////////////////////////////////////////////////

    if( textures[ti->miptex] == NULL)
        {
        faces[i]= NULL;
        continue;
        }

   float is= 1.0f/(float)textures[ti->miptex]->width;
   float it= 1.0f/(float)textures[ti->miptex]->height;

   face-> first = vt_array.size();
   face-> count = pHLBspFile-> dfaces[i].numedges;
   for( EdgeIdx_a = 0; EdgeIdx_a < pHLBspFile-> dfaces[i].numedges; EdgeIdx_a++)
        {
        float v[7];
//        dedge_t* pEdge_a;
		  int VtxId_a= pHLBspFile-> getFirstVtxId( & pHLBspFile-> dfaces[i], EdgeIdx_a);//, pEdge_a);
		  VectorCopy(pHLBspFile-> dvertexes[VtxId_a].point, v);
/*	
        int SurfEdge_a = pHLBspFile-> dsurfedges[ pHLBspFile-> dfaces[i].firstedge + EdgeIdx_a];
		  if(SurfEdge_a < 0)
            {
            pEdge_a = &pHLBspFile-> dedges[-SurfEdge_a];
            VectorCopy(pHLBspFile-> dvertexes[pEdge_a-> v[ 1]].point, v);
            }
        else
            {
            pEdge_a = &pHLBspFile-> dedges[SurfEdge_a];
				VectorCopy(pHLBspFile-> dvertexes[pEdge_a-> v[ 0]].point, v);
            }
*/
        float s = DotProduct(v, ti->vecs[0]) + ti->vecs[0][3];
        float t = DotProduct(v, ti->vecs[1]) + ti->vecs[1][3];
        v[3]= s*is;
        v[4]= t*it;
        if (!(face->flags & TEX_SPECIAL))
            {
            // compute lightmap coords
            float mid_poly_s = (min[0] + max[0])/2.0f;
            float mid_poly_t = (min[1] + max[1])/2.0f;
            float mid_tex_s = (float)lw / 2.0f;
            float mid_tex_t = (float)lh / 2.0f;
            float ls = mid_tex_s + (s - mid_poly_s) / 16.0f;
            float lt = mid_tex_t + (t - mid_poly_t) / 16.0f;
            ls /= (float) lw;
            lt /= (float) lh;

            v[5] = ls;
            v[6] = lt;
            }

            ////////////////////////
        vt_array.add(v);
        st_array.add(v+3);          
        lst_array.add(v+5);
        }
        ///////////////////////////////
    vec3_t n;
    VectorCopy(face-> plane. normal, n);
    float d= face-> plane. dist;
    if( face-> side)
        {
        VectorChangeSign(n);
        face->Plane= RenderPlane_ct(n, -d);
        }
    else
        face->Plane= RenderPlane_ct(n, +d);
    // add face to array of faces
    faces[i]= face;
    }
    // setup vertex arrays
glEnableClientState(GL_VERTEX_ARRAY);
glEnableClientState(GL_TEXTURE_COORD_ARRAY);
glVertexPointer(3, GL_FLOAT, 0, vt_array);


	//frame temps
VisibleSameTextureFaceLists= new RenderFace_ct*[max_visible_faces];
if (VisibleSameTextureFaceLists== NULL)
   Error("Error allocating memory for visible faces!\n");

MarkedFacesBitsets= new unsigned char[(pHLBspFile-> numfaces+7)>>3];  // allocate enough bits
if( !MarkedFacesBitsets)
    Error("Error allocating memory for marked faces!\n");

FocusEntIdx= getNextEntity(0);
}
//------------------------------------------------
void WorldRender_ct:: update()
{
animateTextures();
vec3_t Eye_a;
VectorCopy(Camera.eye(), Eye_a);
calcFacesToRenderFrom( Eye_a);
if( bContinuousTrace)
	getTextureInfo();
}
//-------------------------------------------------------------------
int WorldRender_ct:: calcFacesToRenderFrom( vec3_t Eye_i)
{
    // find leaf we're in
CamLeafIdx= pHLBspFile-> findLeaf(Eye_i);

memset( MarkedFacesBitsets, 0x00, (pHLBspFile-> numfaces+7)>> 3);
memset( VisibleSameTextureFaceLists, 0x00, max_visible_faces * sizeof(RenderFace_ct *));
skyfaces= NULL;
rendered= 0;

vis_leaf_t & vis_leaf= Visibility. VisLeaf( CamLeafIdx);
    // go thru leaf visibility list
for( int i = 0; i < vis_leaf.num_leafs; i++)
	{
   if( bDrawFocusLeafFaces && FocusLeafId>=0 && vis_leaf.LeafIds[i]!= FocusLeafId)
		continue;
	
	const dleaf_t & Leaf_a= pHLBspFile-> dleafs[ vis_leaf.LeafIds[i]];
    // discard leafs outside frustum
	if( frustum_checkCull( Leaf_a.mins, Leaf_a.maxs))
		continue;
	else	
		{
      unsigned short* p= pHLBspFile-> dmarksurfaces+ Leaf_a.firstmarksurface;
      for( int x= 0; x< Leaf_a. nummarksurfaces; x++)
			{
         // don't render those already rendered
         int face_idx= *p++;
         if(!(MarkedFacesBitsets[face_idx>>3] & (1<<(face_idx & 7))))
				{
            // back face culling
            RenderFace_ct* f= faces[face_idx];
            if( f== NULL)
					continue;

            float d;
				unsigned int FaceFlags_a= 0;
				d= DotProduct(Eye_i, f->plane.normal)- f-> plane.dist;
            if( f-> side)
					{
               if( d<= 0)
						FaceFlags_a|= KB_FD_Visible;
					}
				else
               {
               if( d>= 0)
						FaceFlags_a|= KB_FD_Visible;
					}
				FaceFlags_a|= BspExtsn. FaceFlags[ face_idx];
				if( (FaceFlags_a & FaceDisplayFlags)!= FaceDisplayFlags)
					continue;
	 
			   // mark face as visible
            MarkedFacesBitsets[ face_idx>> 3]|= (1<<(face_idx & 7));

            int idx= f-> texture;
            if( textures[ idx] && (textures[ idx]-> flags & TEXTURE_SKY))
					{
               f-> next= skyfaces;
               skyfaces= f;
               }
            else
               {
               f-> next= VisibleSameTextureFaceLists[ idx];
					VisibleSameTextureFaceLists[ idx]= f;
               }
            ++rendered;

                    // chose lightmap style
            f-> lightmap= f-> lightmaps[ 0];
            if( current_style != 0) 
					{
               for(int c = 0; c < MAXLIGHTMAPS; c++) 
						{
						if(f->styles[c]== current_style) 
							{
                     f->lightmap = f->lightmaps[c];
                     break;
							}
						}
					}
				}
			}
		}        
	}
return CamLeafIdx;
}
//---------------------------------------------------
void getColorsFromVect( vecf3_t Vect_i, float & R_a, float & G_a, float & B_a)
	{
	R_a=  0.0f;
	G_a=  0.5*(1+ fabs( Vect_i[0])- fabs( Vect_i[1]));
	B_a=  0.5*(1+ Vect_i[2]);
	}
//---------------------------------------------------
void getColorsFromVect( vecd3_t Vect_i, float & R_a, float & G_a, float & B_a)
	{
	R_a=  0.0f;
	G_a=  0.5*(1+ fabs( Vect_i[0])- fabs( Vect_i[1]));
	B_a=  0.5*(1+ Vect_i[2]);
	}
//----------------------------------------------------
void renderBoundEdges(int LeafId_i)
	{
	glDisable(GL_BLEND);
	glEnable(GL_LINE_SMOOTH);  // more visible, but a lot slower
	glLineWidth(3.0f/*1.0*/);
	glColor3f (0.0f, 0.0f, 1.0f);

	LeafBound_ct * pBound_a= BspExtsn.LeafBoundLists[LeafId_i];
	if(!pBound_a)
		return;
	int NodeSide_a;
	while( pBound_a)
		{
		NodeSide_a= (pBound_a-> NodeIds[0]== -1-LeafId_i? 0:1);

		Winding_ct * w= pBound_a-> winding;
		glBegin (GL_LINES);
		for( int i=1; i< w->numpoints;i++)
			{
			glVertex3f( (float)(w->points[i-1][0]), (float)(w->points[i-1][1]), (float)(w->points[i-1][2]));
			glVertex3f( (float)(w->points[i][0]), (float)(w->points[i][1]), (float)(w->points[i][2]));
			}
		--i;
		glVertex3f( (float)(w->points[i][0]), (float)(w->points[i][1]), (float)(w->points[i][2]));
		glVertex3f( (float)(w->points[0][0]), (float)(w->points[0][1]), (float)(w->points[0][2]));
		glEnd ();
		
		pBound_a= pBound_a-> pNexts[NodeSide_a];
		}
	glDisable(GL_LINE_SMOOTH);
	}
//---------------------------------------------------
void renderLeafBounds( int LeafId_i)
	{
	LeafBound_ct * pBound_a= BspExtsn.LeafBoundLists[LeafId_i];
	if(!pBound_a)
		return;
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	dplane_t * pPlane_a;
	int NodeSide_a;
	while( pBound_a)
		{
		NodeSide_a= (pBound_a-> NodeIds[0]== -1-LeafId_i? 0:1);
		int NeighbourNodeId_a= pBound_a-> NodeIds[NodeSide_a^1];
		if( NeighbourNodeId_a==-1)
			{
			pBound_a= pBound_a-> pNexts[NodeSide_a];
			continue;
			}

		pPlane_a= & pBound_a-> plane;
		float R_a, G_a, B_a;
		getColorsFromVect(pPlane_a-> normal, R_a, G_a, B_a);	
	
		glColor4f( R_a, G_a, B_a, 0.5f);
	
		Winding_ct * w= pBound_a-> winding;
		glBegin( GL_POLYGON );
		for( int i=0; i< w->numpoints;i++)
			glVertex3f( (float)(w->points[i][0]), (float)(w->points[i][1]), (float)(w->points[i][2]));
		glEnd();

		R_a=  1.0f- R_a;
		glColor4f( R_a, G_a, B_a, 0.5f);

		glBegin( GL_POLYGON );
		for( i=w->numpoints- 1; i>=0 ;i--)
			glVertex3f( (float)(w->points[i][0]), (float)(w->points[i][1]), (float)(w->points[i][2]));
		glEnd();

//		glFlush();
		pBound_a= pBound_a-> pNexts[NodeSide_a];
		}

	glDisable(GL_BLEND);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	}

//---------------------------------------------------
void WorldRender_ct:: renderLeafEdges( int LeafId_a, vec3_t Pos2_a)
{
glDisable(GL_BLEND);
glEnable(GL_LINE_SMOOTH);  // more visible, but a lot slower
glLineWidth(3.0f);
glColor3f (1.0f, 0.0f, 0.0f);

const dleaf_t & leaf = pHLBspFile-> dleafs[ LeafId_a];
unsigned short* p = pHLBspFile-> dmarksurfaces + leaf.firstmarksurface;
for( int x = 0; x < leaf. nummarksurfaces; x++)
	{
   int face_idx = *p++;
   RenderFace_ct* pFace_a = faces[face_idx];
	float d;
	unsigned int FaceFlags_a= 0;
	d= DotProduct(Pos2_a, pFace_a->plane.normal)- pFace_a-> plane.dist;
   if( pFace_a-> side)
		{
      if( d<= 0)
			FaceFlags_a|= KB_FD_Visible;
		if( pFace_a->plane. normal[ 2]<=- 0.71)//&& pFace_a->plane.normal[ 2]!= 1)
			FaceFlags_a|= KB_FD_Walkable;
		}
	else
      {
      if( d>= 0)
			FaceFlags_a|= KB_FD_Visible;
		if( pFace_a->plane. normal[ 2]>= 0.71)
			FaceFlags_a|= KB_FD_Walkable;
		}
	if( (FaceFlags_a & FaceDisplayFlags)!= FaceDisplayFlags)
		continue;

	render_edges( pFace_a);						
	}
glDisable(GL_LINE_SMOOTH);
}
//-----------------------------------------------------------
void WorldRender_ct::render_edges(const RenderFace_ct *face)
{
int i, v1, v2;
glBegin (GL_LINES);
for (i = 0; i < face->count; i++)
   {
   v1 = face->first + i;
   v2 = face->first + ((i + 1) % face->count);
   glVertex3f (vt_array[v1][0], vt_array[v1][1], vt_array[v1][2]);
   glVertex3f (vt_array[v2][0], vt_array[v2][1], vt_array[v2][2]);
   }
glEnd ();
}
//-------------------------------------------------
int WorldRender_ct::render_visible_edges(void)
{
if ( show_edges== 0)
   return 0;

if ( show_edges== 2)
   {
   glEnable(GL_LINE_SMOOTH);  // more visible, but a lot slower
   glLineWidth(2.0f/*1.0*/);
   }
else
   glLineWidth(2.0f);  // faster, but not as visible

glColor3f (1.0f, 0.0f, 0.0f);

for (int i = 0; i < max_visible_faces; i++)
   {
   RenderFace_ct* f = VisibleSameTextureFaceLists[i];
   while (f)
      {
      render_edges(f);
      f= f-> next;
      }
   }
glColor3f (1.0f, 1.0f, 1.0f);
if ( show_edges == 2)
   glDisable(GL_LINE_SMOOTH);
return 1;
}
//--------------------------------------------------
void WorldRender_ct::render_face(const RenderFace_ct* face)
{
if( !textures[ face-> texture])
    return;
textures[ face-> texture]-> bind();
glTexCoordPointer( 2, GL_FLOAT, 0, st_array);
glDrawArrays(face-> type, face-> first, face-> count);
}
//------------------------------------------------
void WorldRender_ct::render_visible_faces()
{
for (int i = 0; i < max_visible_faces; i++)
    {
    RenderFace_ct* f = this-> VisibleSameTextureFaceLists[ i];
    while( f)
        {
        // is this NOT a special texture?
        if(!(f->flags & TEX_SPECIAL))
            render_face(f);
        f= f->next;
        }
    }
}
//------------------------------------------------------
void WorldRender_ct::render_lightmap(const RenderFace_ct* face)
{
if ((face->lightmap) && (bEnable_lighting))
	{
    glBlendFunc(GL_ZERO, GL_SRC_COLOR);
    glEnable(GL_BLEND);
    glDepthMask(GL_FALSE);
    glDepthFunc(GL_EQUAL);

    face->lightmap->bind();

    glTexCoordPointer(2, GL_FLOAT, 0, lst_array);
    glDrawArrays(face->type, face->first, face->count);
        
    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
    glDepthFunc(GL_LESS);
	}
}
//----------------------------------------------------
void WorldRender_ct::render_visible_lightmaps()
{
for (int i = 0; i < max_visible_faces; i++)
	{
    RenderFace_ct* f = VisibleSameTextureFaceLists[i];
    while (f)
		{  // is this NOT a special texture?
      if (!(f->flags & TEX_SPECIAL))
			render_lightmap(f);
      f= f->next;
		}
	}
}
//-------------------------------------------------------
void WorldRender_ct::render_special_textures()
{
	//_ render special textures ("clip", "origin" and "aaatrigger")


int ent_index;
int model_index;

if(!bRenderSpecialTextures)
   return;

glEnable(GL_BLEND);  // Turn Blending On
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
glColor4f(1.0f, 1.0f, 1.0f, 0.4f);

for (int i = 0; i < max_visible_faces; i++)
   {
   RenderFace_ct* f = VisibleSameTextureFaceLists[i];
   while( f)
      {
         // is this a special texture?
      if( f->flags & TEX_SPECIAL)
            render_face(f);
      f= f->next;
      }
   }

for (ent_index = 0; ent_index < num_entvars; ent_index++)
   {
      // is this a brush model?
   if (entvars[ent_index].brush_model_index> 0)
      {
      model_index = entvars[ent_index].brush_model_index;
      glPushMatrix();
      dmodel_t* pModel_a = &pHLBspFile-> dmodels[model_index];
      glTranslatef( pModel_a->origin[0], pModel_a->origin[1], pModel_a->origin[2]);
      for (int c = 0; c < pModel_a-> numfaces; c++)
         {
         RenderFace_ct *pFace_a= faces[pModel_a-> firstface + c];
         if( !pFace_a)
            continue;

        // is this a special texture?
         if (pFace_a->flags & TEX_SPECIAL)
            render_face(pFace_a);
         }
      glPopMatrix();
      }
   }
glDisable( GL_BLEND);  // Turn Blending Off
glColor4f( 1.0f, 1.0f, 1.0f, 1.0f);  // back to white, full alpha
}
//-------------------------------------------------
int WorldRender_ct:: getNextEntity(int)
{
int Old_a= FocusEntIdx;
do
	{
	FocusEntIdx= (++FocusEntIdx)%num_entvars;
	if( (entvars[FocusEntIdx].brush_model_index > 0)|| (entvars[FocusEntIdx].studio_model))
		break;
	}while( FocusEntIdx!= Old_a);
return FocusEntIdx;
}
//------------------------------------------------------------
void WorldRender_ct::render_entvars()
{
   int ent_index;
   int model_index;
   int pass;
   vec3_t v1, v2;
   float dot;

   // set up right vector for studio model chrome rendering...
   Camera.right_vector(g_vright);

   // two passes, render solid objects first, then non-solid/masked objects
for(pass= 0; pass< 2; pass++)
   {
   for (ent_index= 0; ent_index < num_entvars; ent_index++)
      {
	   // is this a brush model?
      if (entvars[ent_index].brush_model_index > 0)
         {
          model_index = entvars[ent_index].brush_model_index;
          if (((pass == 0) && (entvars[ent_index].renderamt == 255)) 
				|| ((pass == 1) && (entvars[ent_index].renderamt != 255)))
            {
            glPushMatrix();
            dmodel_t* m = &pHLBspFile-> dmodels[model_index];
            glTranslatef(m->origin[0],m->origin[1],m->origin[2]);
            glEnable(GL_BLEND);  // Turn Blending On
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

//               glColor4f(entvars[ent_index].rendercolor[0],
//                         entvars[ent_index].rendercolor[1],
//                         entvars[ent_index].rendercolor[2],
//                         entvars[ent_index].renderamt);
            glColor4f(1.0, 1.0, 1.0, entvars[ent_index].renderamt);
            for (int c = 0; c < m->numfaces; c++)
               {
               RenderFace_ct *face = faces[m->firstface + c];
               if (face == NULL)
                  continue;
                  // is this NOT a special texture?
               if (!(face->flags & TEX_SPECIAL))
                  render_face(face);
               }
            glDisable(GL_BLEND);  // Turn Blending Off
            glPopMatrix();
            }
         }
     else if ((pass == 0) && (entvars[ent_index].studio_model))
         {
         Camera.fwd_vec(v1);
         VectorSubtract(entvars[ent_index].origin, Camera.eye(), v2);
         dot= DotProduct(v1, v2);

            // is this entity outside the view frustum?
         if(dot > 0.7660)  // greater than sin(50)?
            continue;

         glPushMatrix();

         glTranslatef(entvars[ent_index].origin[0],
                      entvars[ent_index].origin[1],
                      entvars[ent_index].origin[2]);

            // rotate about the Z axis (by yaw degrees)...
         glRotatef(entvars[ent_index].angles[1], 0, 0, 1);

         entvars[ent_index].studio_model->SetBlending(0, 0.0f);
         entvars[ent_index].studio_model->SetBlending(1, 0.0f);

         entvars[ent_index].studio_model-> DrawModel();
         glPopMatrix();
         }
      }
   }
glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
}
//------------------------------------------------------------
void WorldRender_ct::render_entvar( int ent_index)
	{
   if( ent_index>= num_entvars)
		return;
   int model_index;
   int pass;
   vec3_t v1, v2;
   float dot;

   // set up right vector for studio model chrome rendering...
   Camera.right_vector(g_vright);

   // two passes, render solid objects first, then non-solid/masked objects
	for(pass= 0; pass< 2; pass++)
		{
	   // is this a brush model?
      if (entvars[ent_index].brush_model_index > 0)
         {
          model_index = entvars[ent_index].brush_model_index;
          if (((pass == 0) && (entvars[ent_index].renderamt == 255)) 
				|| ((pass == 1) && (entvars[ent_index].renderamt != 255)))
            {
            glPushMatrix();
            dmodel_t* m = &pHLBspFile-> dmodels[model_index];
            glTranslatef(m->origin[0],m->origin[1],m->origin[2]);
            glEnable(GL_BLEND);  // Turn Blending On
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

//               glColor4f(entvars[ent_index].rendercolor[0],
//                         entvars[ent_index].rendercolor[1],
//                         entvars[ent_index].rendercolor[2],
//                         entvars[ent_index].renderamt);
            glColor4f(1.0, 1.0, 1.0, entvars[ent_index].renderamt);
            for (int c = 0; c < m->numfaces; c++)
               {
               RenderFace_ct *face = faces[m->firstface + c];
               if (face == NULL)
                  continue;
                  // is this NOT a special texture?
               if (!(face->flags & TEX_SPECIAL))
                  render_face(face);
               }
            glDisable(GL_BLEND);  // Turn Blending Off
            glPopMatrix();
            }
         }
      else if ((pass == 0) && (entvars[ent_index].studio_model))
         {
         Camera.fwd_vec(v1);
         VectorSubtract(entvars[ent_index].origin, Camera.eye(), v2);
         dot= DotProduct(v1, v2);

            // is this entity outside the view frustum?
         if(dot > 0.7660)  // greater than sin(50)?
            continue;

         glPushMatrix();

         glTranslatef(entvars[ent_index].origin[0],
                      entvars[ent_index].origin[1],
                      entvars[ent_index].origin[2]);

            // rotate about the Z axis (by yaw degrees)...
         glRotatef(entvars[ent_index].angles[1], 0, 0, 1);

         entvars[ent_index].studio_model->SetBlending(0, 0.0f);
         entvars[ent_index].studio_model->SetBlending(1, 0.0f);

         entvars[ent_index].studio_model-> DrawModel();
         glPopMatrix();
         }
      }
   
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	}
//-------------------------------------------------
void WorldRender_ct::render_skyfaces()
{
   int sky_flag;  // which skybox faces should be rendered
   vec3_t cam;

   if (sky_textures[0] == NULL)  // no sky textures loaded?
      return;

   VectorCopy(Camera.eye(), cam);

   // I should REALLY be using the skyfaces linked list to walk through
   // the faces and only render the ones that are visible, but I seem
   // to be having some problems with that.  For now, just render ALL
   // of the skyfaces (even the ones that aren't possibly visible)...

   sky_flag = SKY_FRONT | SKY_BACK | SKY_LEFT | SKY_RIGHT | SKY_UP | SKY_DOWN;

if (sky_flag & SKY_FRONT)
   {
     sky_textures[0]->bind();  // FRONT
      glBegin(GL_QUADS);
      glTexCoord2f(0,0);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(0,1);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]+SKY_DEPTH);
      glTexCoord2f(1,1);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]+SKY_DEPTH);
      glTexCoord2f(1,0);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]-SKY_DEPTH);
      glEnd();
   }

if (sky_flag & SKY_BACK)
   {
      sky_textures[1]->bind();  // BACK

      glBegin(GL_QUADS);

      glTexCoord2f(1,0);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(0,0);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(0,1);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]+SKY_DEPTH);
      glTexCoord2f(1,1);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]+SKY_DEPTH);

      glEnd();
   }

if (sky_flag & SKY_LEFT)
   {
      sky_textures[2]->bind();  // LEFT

      glBegin(GL_QUADS);

      glTexCoord2f(1,0);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(0,0);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(0,1);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]+SKY_DEPTH);
      glTexCoord2f(1,1);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]+SKY_DEPTH);

      glEnd();
   }

if (sky_flag & SKY_RIGHT)
   {
      sky_textures[3]->bind();  // RIGHT

      glBegin(GL_QUADS);

      glTexCoord2f(1,0);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(0,0);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(0,1);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]+SKY_DEPTH);
      glTexCoord2f(1,1);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]+SKY_DEPTH);

      glEnd();
   }

if (sky_flag & SKY_UP)
   {
      sky_textures[4]->bind();  // UP

      glBegin(GL_QUADS);

      glTexCoord2f(0,1);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]+SKY_DEPTH);
      glTexCoord2f(1,1);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]+SKY_DEPTH);
      glTexCoord2f(1,0);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]+SKY_DEPTH);
      glTexCoord2f(0,0);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]+SKY_DEPTH);

      glEnd();
   }

if (sky_flag & SKY_DOWN)
   {
      sky_textures[5]->bind();  // DOWN

      glBegin(GL_QUADS);

      glTexCoord2f(0,1);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(1,1);
      glVertex3f(cam[0]+SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(1,0);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]-SKY_DEPTH, cam[2]-SKY_DEPTH);
      glTexCoord2f(0,0);
      glVertex3f(cam[0]-SKY_DEPTH, cam[1]+SKY_DEPTH, cam[2]-SKY_DEPTH);

      glEnd();
   }
}
//--------------------------------------------------
void WorldRender_ct:: traceLine()
	{
	vec3_t vec_start;
	vec3_t vec_end;
	VectorCopy( Camera.eye(), vec_start);
	VectorCopy( vec_start, traceline_pt[0]);  // save the starting point
	Camera.fwd_vec( vec_end);
	VectorNormalize( vec_end);
	VectorScale( vec_end, (vec_t)-WorldHalfSizeMax, vec_end);
	VectorAdd( vec_start, vec_end, vec_end);

	DispTrace. pHLBspFile= pHLBspFile;
	DispTrace. Flags= 0;

	DispTrace. Type= Trace_ct::K_Point;
	DispTrace. Flags= Trace_ct::KB_Face| Trace_ct:: KB_Texture;

	VectorCopy( DispTrace. Start, traceline_pt[0]);  // save the starting point
	VectorCopy( DispTrace. End, traceline_pt[1]);  // save the ending point
	VectorCopy( DispTraceData.endpos, traceline_pt[2]);  // save the hit point
	}
//-------------------------------------
void WorldRender_ct:: getTextureInfo()
	{
	vec3_t vec_start;
	vec3_t vec_end;

	VectorCopy(Camera.eye(), vec_start);
	Camera.fwd_vec(vec_end);
	VectorNormalize(vec_end);
	VectorScale(vec_end, (vec_t)-WorldHalfSizeMax, vec_end);
	VectorAdd(vec_start, vec_end, vec_end);
	
	DispTrace. pHLBspFile= pHLBspFile;
	DispTrace. Type= TraceType;
	DispTrace. Flags= Trace_ct:: KB_Face| Trace_ct:: KB_Texture;

	if( DispTrace. Type== Trace_ct::K_AABox)
		DispTrace. setAABoxCenterSize( vec_start, AABoxHalfSize); 
	else if( DispTrace. Type== Trace_ct::K_AACylinder)
		DispTrace. setAACylinder( vec_start, AABoxHalfSize[0], AABoxHalfSize[2], 2);
	else if( DispTrace. Type== Trace_ct::K_Sphere)
		DispTrace. setSphere( vec_start, AABoxHalfSize[0]);

	DispTrace. traceMove( vec_start, vec_end, DispTraceData);

	VectorCopy( DispTrace.Start, traceline_pt[0]);  // save the starting point
	VectorCopy( DispTraceData.endpos, traceline_pt[2]);  // save the ending point
	VectorCopy( DispTrace.End, traceline_pt[1]);  // save the ending point
/*
	Trace_ct Trace_a(pHLBspFile);
	Trace_a. Type= TraceType;
	Trace_a. Flags= Trace_ct:: KB_Face| Trace_ct:: KB_Texture;
	if( TraceType== Trace_ct::K_AABox)
		Trace_a. setAABoxCenterSize( vec_start, AABoxHalfSize); 
	else if( TraceType== Trace_ct::K_AACylinder)
		Trace_a. setAACylinder( vec_start, AABoxHalfSize[0], AABoxHalfSize[2], 2);
	else if( TraceType== Trace_ct::K_Sphere)
		Trace_a. setSphere( vec_start, AABoxHalfSize[0]);

	Trace_a. traceMove( vec_start, vec_end, DispTraceData);

	VectorCopy( Trace_a.Start, traceline_pt[0]);  // save the starting point
	VectorCopy( DispTraceData.endpos, traceline_pt[2]);  // save the ending point
	VectorCopy( Trace_a.End, traceline_pt[1]);  // save the ending point
*/

	bDrawTraceline= 1;
	if( DispTrace. Type!= Trace_ct::K_Point)
		bDrawHull= ( DispTraceData.fraction< 1.0f);
	}
//-----------------------------------------------------
void drawZCylinder( float Radius_i, float HalfHeight_i, vec3_t Pos_i)
{
glColor3f( 1.0f, 0.0, 0.0f);
glLineWidth(GrHullLineWidth);
if( CylinderModel_s.n!= GrCylNumOfSectors)
	CylinderModel_s. setNumOfSectors(GrCylNumOfSectors); 
Vector Start_a( Pos_i);
Start_a[2]-= HalfHeight_i;
Vector End_a( Pos_i);
End_a[2]+= HalfHeight_i;
CylinderModel_s. drawWF( Radius_i, Start_a[0], Start_a[1], Start_a[2]
	, End_a[0], End_a[1], End_a[2]);
}
//-----------------------------------------------------
void drawAABox( vec3_t Size_i, vec3_t Pos_i)
{
glColor3f(1.0f,0.0,0.0f);
glLineWidth(3.0f);
glPushMatrix();
glTranslatef( Pos_i[0], Pos_i[1], Pos_i[2]);
glScalef( Size_i[0],Size_i[1],Size_i[2]);
glutWireCube( 1);
glPopMatrix();
}
//-----------------------------------------------------
void drawSphere( float Radius, vec3_t Pos_i)
{
glColor3f(1.0f,0.0,0.0f);
glLineWidth(3.0f);
glPushMatrix();
glTranslatef( Pos_i[0], Pos_i[1], Pos_i[2]);
glutWireSphere( Radius, 8, 6);
glPopMatrix();
}
//--------------------------------------------------
void WorldRender_ct:: draw()
{
static vec3_t cam;

if(!bLockCurrLeaf)
	{
	VectorCopy(Camera.eye(), cam);
	if(bTraceAvail)
		{
		FocusLeafId= DispTraceData.FurthestNonBlockingLeafNum;
		}
	else
		FocusLeafId= pHLBspFile-> findLeaf(cam);
	}

if(!this-> bDraw)
	return;

if( bDrawTraceline && bDrawHull)
	{
	if( TraceType== Trace_ct:: K_AABox)
		drawAABox( AABoxSize, traceline_pt[2]);
	else if( TraceType== Trace_ct:: K_Sphere)
		drawSphere( AABoxHalfSize[0], traceline_pt[2]);
	else if(TraceType== Trace_ct:: K_AACylinder)
		drawZCylinder( AABoxHalfSize[0], AABoxHalfSize[2], traceline_pt[2]);
	}

if( bDrawFocusLeafBoundEdges && FocusLeafId>= 0)
	renderBoundEdges( FocusLeafId);

if( bDrawFocusLeafEdges && FocusLeafId>= 0)
	renderLeafEdges( FocusLeafId, cam);

if( !bDrawFocusLeafEdges)
	render_visible_edges();

//if( bDrawFocusLeafFaces && FocusLeafId>= 0)
render_visible_faces();

render_visible_lightmaps();

if( bDrawSky)
	render_skyfaces();  // render skybox

if( DrawEntities==1)
	render_entvars();  // render entities (brush models and models)
else if( DrawEntities==2)
	render_entvar( FocusEntIdx);

if( bDrawTraceline)
	drawTraceline( traceline_pt);


glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

if( bDrawSpecTextures && DrawEntities< 2)
	render_special_textures();  // aaatrigger, face and origin textures

if( bDrawFocusLeafBounds && FocusLeafId>= 0)
	renderLeafBounds( FocusLeafId);		

}
//----------------------------------------------------------
void ComputeLightmap(const HLBspFile_ct & Bsp_i,  dface_t *dface, RenderFace_ct *face, float *mins, float *maxs, int *lh, int *lw)
	{
    int c, i;
    int width, height;
    unsigned char * data;

    // compute lightmap size
    int size[2];
    for (c = 0; c < 2; c++)
	    {
        float tmin = (float) floor(mins[c]/16.0f);
        float tmax = (float) ceil(maxs[c]/16.0f);

        size[c] = (int) (tmax-tmin);
		 }

    width = size[0]+1;
    height = size[1]+1;

    *lw = width;
    *lh = height;

    int lsz = width * height * 3;  // RGB buffer size

    // generate lightmaps texture
//    for(c = 0; c < MAX_LIGHTMAPS; c++)
    for (c = 0; c < 1; c++)
		{
        if (dface->styles[c] == -1)
            break;

        face->styles[c] = dface->styles[c];
        data = new  unsigned char[lsz];
        memcpy(data, Bsp_i. dlightdata + dface->lightofs + (lsz * c), lsz);

        float f, light_adjust;
        int inf;

        light_adjust = 1.0f - brightness;
        // scale lightmap value...
        for (i = 0; i < lsz; i++)
			{
            f = (float)pow((data[i]+1)/256.0, light_adjust);
            inf = f * 255.0f + 0.5f;
            if (inf < 0)
               inf = 0;
            if (inf > 255)
               inf = 255;
            data[i] = (int)inf;
			 }
        face->lightmaps[c] = new Texture_ct(data, width, height, GL_RGB,
                                         nearest_mipmap_nearest);
        delete[] data;
		}
    face->lightmap = face->lightmaps[0];
	}
//----------------------------------------------------------
void calcSurfaceAABox ( const HLBspFile_ct & Bsp_i, dface_t *pFace_i, float *mins_o, float *maxs_o)
{
float val;
int  i,j;
int e;
dvertex_t * pVertex_a;
texinfo_t * pTexInfo_a;

mins_o[0]= mins_o[1]=  999999;
maxs_o[0]= maxs_o[1]= -999999;

pTexInfo_a= Bsp_i. texinfo+ pFace_i-> texinfo;

for( i=0 ; i< pFace_i->numedges; i++)
   {
   e= Bsp_i. dsurfedges[ pFace_i->firstedge+ i];
   if( e>= 0)
      pVertex_a= &Bsp_i. dvertexes[Bsp_i. dedges[e].v[ 0]];
   else
      pVertex_a= &Bsp_i. dvertexes[Bsp_i. dedges[-e].v[ 1]];
      
   for( j= 0; j< 2 ; j++)
      {
      val=  pVertex_a-> point[ 0] * pTexInfo_a-> vecs[ j][ 0] 
          + pVertex_a-> point[ 1] * pTexInfo_a-> vecs[ j][ 1]
          + pVertex_a-> point[ 2] * pTexInfo_a-> vecs[ j][ 2]
          + pTexInfo_a-> vecs[ j][ 3];
      if( val< mins_o[ j])
          mins_o[ j] = val;
      if( val> maxs_o[ j])
          maxs_o[ j] = val;
      }
   }

//   for (i=0 ; i<2 ; i++)
//   {  
//      bmins[i] = floor(mins[i]/16);
//      bmaxs[i] = ceil(maxs[i]/16);
//      s->texturemins[i] = bmins[i] * 16;
//      s->extents[i] = (bmaxs[i] - bmins[i]) * 16;
//      if ( !(tex->flags & TEX_SPECIAL) && s->extents[i] > 512 /* 256 */ )
//         Sys_Error ("Bad surface extents");
//   }
}


