/* 
 * HL rendering engine
 * Copyright (c) 2000,2001 Bart Sekura
 *
 * Permission to use, copy, modify and distribute this software
 * is hereby granted, provided that both the copyright notice and 
 * this permission notice appear in all copies of the software, 
 * derivative works or modified versions.
 *
 * THE AUTHOR ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION AND DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES 
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * world render class
 */

//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef WORLDRENDER_CT_
#define WORLDRENDER_CT_

//#include <GL/glut.h>

#include "ChunkedVector_ct.h"
#include "bspfile2.h"
#include "bsp_ext.h"
//#include "mathlib.h"
#include "mathlib_tpl.h"
#include "vis.h"
#include "texture.h"
#include "RenderPlane_ct.h"
#include "RenderFace_ct.h"
#include "matrix.h"


//+++++++++++++++++++++++++++++++++++++++++++++++
class WorldRender_ct 
{
public:
static enum{ KB_DP_Solid=1, KB_DP_NotSolid=2};
HLBspFile_ct * pHLBspFile;

int bDraw;
int DrawEntities;
int FocusEntIdx;
int bDrawSky;
int bDrawSpecTextures;
int bEnable_lighting;
int bRenderSpecialTextures;

int FocusLeafId;
int bDrawFocusLeafEdges;
int bDrawFocusLeafFaces;
int bDrawFocusLeafBounds;
int bDrawFocusLeafBoundEdges;

unsigned DrawBoundFlags;

int   show_edges;
int   bLockCurrLeaf;

vec3_t traceline_pt[3];
int bTraceAvail;
int bContinuousTrace;

unsigned int FaceDisplayFlags;

int CamLeafIdx;

    WorldRender_ct();
    virtual ~WorldRender_ct();

    virtual void clear();
    virtual void reset();
    virtual void init();
    virtual void draw();
	 virtual void update();
//    virtual void shutdown();

    int process_visible_faces();
    int calcFacesToRenderFrom( vec3_t Pos_i);

    virtual void render_face(const RenderFace_ct* f);
    void render_visible_faces();
    void render_lightmap(const RenderFace_ct* face);
    void render_visible_lightmaps();
    void render_special_textures();
    void render_skyfaces();
    void render_edges(const RenderFace_ct *face);
    int render_visible_edges();
	void render_entvars();
	void render_entvar( int ent_index);
	void renderLeafEdges( int LeafId_i, vec3_t Pos2_a);
	void getTextureInfo();
	void traceLine();
	int getNextEntity(int);
   Visibility_ct Visibility;
	Vt_array_ct vt_array;

// visible faces by texture index
	RenderFace_ct ** VisibleSameTextureFaceLists;  // pointer to array of RenderFace_ct pointers
	int max_visible_faces;

unsigned char * MarkedFacesBitsets;
RenderFace_ct ** faces;  // pointer to array of RenderFace_ct pointers
RenderFace_ct * skyfaces;
};
/*
//++++++++++++++++++++++++++++++++
class BspMoveRender_ct: public WorldRender_ct
{
public:

    BspMoveRender_ct();
	 void init();
    void draw();
	 void update();
//int calcFacesToRenderFrom( vec3_t  Eye_i);
//    void render_face(const RenderFace_ct* f);
};
*/
//
#endif // eof