/***
*
*  Copyright (c) 1998, Valve LLC. All rights reserved.
*
*  This product contains software technology licensed from Id 
*  Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*  All Rights Reserved.
*
****/
// modified by "Botman"
// modified by "SamPlay"
//
// bspfile.cpp
//
//#define MI_NOBSPENTITIES
//#define MI_NOBSPCHECKSUM
//#define MI_NOBSPWRITE

#include <math.h>

#include "cmdlib.h"
#include "bspfile2.h"
#ifndef MI_NOBSPENTITIES
 #include "scriplib.h"
#endif
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

dheader_t   *header;
FILE        *wadfile;
dheader_t   outheader;

#ifndef MI_NOBSPENTITIES
typedef int condOnEpair_ft( epair_t  *);
int          parseEntity ( HLBspFile_ct * );
void         parseEntities ( HLBspFile_ct * );
void         unparseEntities ( HLBspFile_ct * );

int			 findEntityByClassname( HLBspFile_ct * , int index, const char *classname);
int          findEntityByWildcard( HLBspFile_ct * ,int index, const char *classname, int length);
int          findEntity( HLBspFile_ct * ,int index, condOnEpair_ft condOnEpair_i);
#endif //MI_NOBSPENTITIES

//=============================================================================
#ifndef MI_NOBSPCHECKSUM
#ifdef __linux__
unsigned _rotl ( unsigned val, int shift)
{
    register unsigned hibit;        /* non-zero means hi bit set */
    register unsigned num = val;    /* number to rotate */

    shift &= 0x1f;                  /* modulo 32 -- this will also make
                                       negative shifts work */
    while (shift--) {
        hibit = num & 0x80000000;   /* get high bit */
        num <<= 1;                  /* shift left one bit */
        if (hibit)
            num |= 1;               /* set lo bit if hi bit was set */
    }

    return num;
}
#endif //__linux__
#endif //MI_NOBSPCHECKSUM
//--------------------------------------------------
void HLBspFile_ct:: init()
{
this-> nummodels = 0;
this-> dmodels = NULL;

this-> visdatasize = 0;
this-> dvisdata = NULL;

this-> lightdatasize = 0;
this-> dlightdata = NULL;

this-> texdatasize = 0;
this-> dtexdata = NULL;

this-> entdatasize = 0;
this-> dentdata = NULL;

this-> numleafs = 0;
this-> dleafs = NULL;

this-> numplanes = 0;
this-> dplanes = NULL;

this-> numvertexes = 0;
this-> dvertexes = NULL;

this-> numnodes = 0;
this-> dnodes = NULL;

this-> numtexinfo = 0;
this-> texinfo = NULL;

this-> numfaces = 0;
this-> dfaces = NULL;

this-> numclipnodes = 0;
this-> dclipnodes = NULL;

this-> numedges = 0;
this-> dedges = NULL;

this-> nummarksurfaces = 0;
this-> dmarksurfaces = NULL;

this-> numsurfedges = 0;
this-> dsurfedges = NULL;

this-> num_entities = 0;
for (int i=0; i < MAX_MAP_ENTITIES; i++)
	this-> entities[i].epairs = NULL;
}
//---------------------------------------
void HLBspFile_ct:: clear()
{
   for (int i=0; i < num_entities; i++)
      entities[i]. clear();
	delete [] dmodels;
	delete [] dvisdata;
	delete [] dlightdata;
	delete [] dtexdata;
	delete [] dentdata;
	delete [] dleafs;
	delete [] dplanes;
	delete [] dvertexes;
	delete [] dnodes;
	delete [] texinfo;
   delete [] dfaces;
   delete [] dclipnodes;
	delete [] dedges;
	delete [] dmarksurfaces;
	delete [] dsurfedges;
}
//--------------------------------------------------
#ifndef MI_NOBSPCHECKSUM
int FastChecksum(char *buffer, int bytes)
{
   int   checksum = 0;
   while( bytes-- )
   {
      checksum = _rotl(checksum, 4) ^ *(char *)buffer;
      buffer++;
   }
   return checksum;
}
#endif //MI_NOBSPCHECKSUM
//----------------------------------------
int HLBspFile_ct:: compressVis (byte *vis, byte *dest)
{
   int   j;
   int   rep;
   int   visrow;
   byte  *dest_p;

   dest_p = dest;
   visrow = ( numleafs + 7)>>3;

   for (j=0 ; j<visrow ; j++)
   {
      *dest_p++ = vis[j];
      if (vis[j])
         continue;
      rep = 1;
      for ( j++; j<visrow ; j++)
         if (vis[j] || rep == 255)
            break;
         else
            rep++;
      *dest_p++ = rep;
      j--;
   }

   return dest_p - dest;
}


/*
===================
DecompressVis
===================
*/
void HLBspFile_ct:: decompressVis (byte *in, byte *decompressed)
{
   int   c;
   byte  *out;
   int   row;

   row = (numleafs+7)>>3;
   out = decompressed;

   do
   {
      if (*in)
      {
         *out++ = *in++;
         continue;
      }

      c = in[1];
      in += 2;
      while (c)
      {
         *out++ = 0;
         c--;
      }
   } while (out - decompressed < row);
}

//=============================================================================
/*
=============
SwapBSPFile

Byte swaps all data in a bsp file.
=============
*/
//---------------------------------------
void HLBspFile_ct:: swapBSPFileModels(qboolean todisk)
{
	if( nummodels==0)
		return;
   int   i, j;
   dmodel_t *d;


// models
	for (i=0 ; i<nummodels ; i++)
		{
      d = &dmodels[i];
      for (j=0 ; j<MAX_MAP_HULLS ; j++)
         d-> headnode[j] = LittleLong (d->headnode[j]);

      d-> visleafs = LittleLong (d->visleafs);
      d-> firstface = LittleLong (d->firstface);
      d-> numfaces = LittleLong (d->numfaces);
        
      for (j=0 ; j<3 ; j++)
			{
         d-> mins[j] = LittleFloat(d->mins[j]);
         d-> maxs[j] = LittleFloat(d->maxs[j]);
         d-> origin[j] = LittleFloat(d->origin[j]);
			}	
		}
}
//-------------------------------------
void HLBspFile_ct:: swapBSPFileTexinfo(qboolean todisk)
{
if(numtexinfo==0)
	return; 
int i,j;

for ( i=0 ; i<numtexinfo ; i++)
   {
   for (j=0 ; j<8 ; j++)
      texinfo[i]. vecs[0][j] = LittleFloat (texinfo[i].vecs[0][j]);
      texinfo[i]. miptex = LittleLong (texinfo[i].miptex);
      texinfo[i]. flags = LittleLong (texinfo[i].flags);
   }
}
//----------------------------------------
void HLBspFile_ct:: swapBSPFileMiptex (qboolean todisk)
{
   if (!texdatasize)
		return;

	dmiptexlump_t  *mtl;
	int i,c;
   mtl = (dmiptexlump_t *)dtexdata;
   if (todisk)
		c = mtl->nummiptex;
   else
      c = LittleLong(mtl->nummiptex);
	mtl->nummiptex = LittleLong (mtl->nummiptex);
   for (i=0 ; i<c ; i++)
      mtl->dataofs[i] = LittleLong(mtl->dataofs[i]);
}
//----------------------------------------------------
void HLBspFile_ct:: swapBSPFile (qboolean todisk)
{
	swapBSPFileModels (todisk);

	// vertexes
	int i,j;
   for (i=0 ; i<numvertexes ; i++)
		{
      for (j=0 ; j<3 ; j++)
         dvertexes[i].point[j] = LittleFloat (dvertexes[i].point[j]);
		}
	// planes
	for (i=0 ; i<numplanes ; i++)
		{
		for (j=0 ; j<3 ; j++)
			dplanes[i].normal[j] = LittleFloat (dplanes[i].normal[j]);
		dplanes[i].dist = LittleFloat (dplanes[i].dist);
		dplanes[i].type = LittleLong (dplanes[i].type);
		}

	swapBSPFileTexinfo(todisk);

	// faces
	for (i=0 ; i<numfaces ; i++)
		{
		dfaces[i].texinfo = LittleShort (dfaces[i].texinfo);
		dfaces[i].planenum = LittleShort (dfaces[i].planenum);
		dfaces[i].side = LittleShort (dfaces[i].side);
		dfaces[i].lightofs = LittleLong (dfaces[i].lightofs);
		dfaces[i].firstedge = LittleLong (dfaces[i].firstedge);
		dfaces[i].numedges = LittleShort (dfaces[i].numedges);
		}

	// nodes
	for (i=0 ; i<numnodes ; i++)
		{
		dnodes[i].planenum = LittleLong (dnodes[i].planenum);
		for (j=0 ; j<3 ; j++)
			{
			dnodes[i].mins[j] = LittleShort (dnodes[i].mins[j]);
			dnodes[i].maxs[j] = LittleShort (dnodes[i].maxs[j]);
			}
			dnodes[i].children[0] = LittleShort (dnodes[i].children[0]);
			dnodes[i].children[1] = LittleShort (dnodes[i].children[1]);
			dnodes[i].firstface = LittleShort (dnodes[i].firstface);
			dnodes[i].numfaces = LittleShort (dnodes[i].numfaces);
		}

	// leafs
	for (i=0 ; i<numleafs ; i++)
		{
		dleafs[i].contents = LittleLong (dleafs[i].contents);
		for (j=0 ; j<3 ; j++)
			{
         dleafs[i].mins[j] = LittleShort (dleafs[i].mins[j]);
         dleafs[i].maxs[j] = LittleShort (dleafs[i].maxs[j]);
			}
		dleafs[i].firstmarksurface = LittleShort (dleafs[i].firstmarksurface);
		dleafs[i].nummarksurfaces = LittleShort (dleafs[i].nummarksurfaces);
		dleafs[i].visofs = LittleLong (dleafs[i].visofs);
	   }

	// clipnodes
   for (i=0 ; i<numclipnodes ; i++)
	   {
      dclipnodes[i].planenum = LittleLong (dclipnodes[i].planenum);
      dclipnodes[i].children[0] = LittleShort (dclipnodes[i].children[0]);
      dclipnodes[i].children[1] = LittleShort (dclipnodes[i].children[1]);
		}

	// miptex
	swapBSPFileMiptex( todisk);

	// marksurfaces
   for (i=0 ; i<nummarksurfaces ; i++)
      dmarksurfaces[i] = LittleShort (dmarksurfaces[i]);

	// surfedges
	for (i=0 ; i<numsurfedges ; i++)
      dsurfedges[i] = LittleLong (dsurfedges[i]);

	// edges
   for (i=0 ; i<numedges ; i++)
	   {
      dedges[i].v[0] = LittleShort (dedges[i].v[0]);
      dedges[i].v[1] = LittleShort (dedges[i].v[1]);
		}
}
//------------------------------------------------------------
int CopyLump (int lump, void **dest, int size)
{
   int length = header->lumps[lump].filelen;
   int ofs = header->lumps[lump].fileofs;
   if (length % size)
      Error ("LoadBSPFile: odd lump size");
   int num= length/ size;
   if (*dest == NULL)
	   {
      // allocate memory for this lump...
      *dest= new byte [num * size];
      if (*dest == NULL)
         Error("Error allocating memory for BSP file lump!\n");
		}
   memcpy (*dest, (byte *)header + ofs, length);
   return num;
}

/*
=============
LoadBSPFile
=============
*/
void HLBspFile_ct:: loadFile (char *filename, unsigned LoadFlags_i)
{
   int   i;
	// load the file header
	byte * pHeaderBytes;
   LoadFile (filename, & pHeaderBytes);
	header= (dheader_t*) pHeaderBytes;
	// swap the header
   for (i=0 ; i< sizeof(dheader_t)/4 ; i++)    
		((int *)header)[i] = LittleLong ( ((int *)header)[i]);
   if (header->version != BSPVERSION)
      Error ("%s is version %i, not %i", filename, header->version, BSPVERSION);

	loadFile0 ( LoadFlags_i);

   delete []pHeaderBytes;    // everything has been copied out
}
//-------------------------------------------------
void HLBspFile_ct:: loadFile0 ( unsigned LoadFlags_i)
{
#ifndef MI_NOBSPWRITE
LoadFlags_i=~0;//for safety
#endif
if( LoadFlags_i & KB_LD_BspCore)
	{
	nummodels = CopyLump (LUMP_MODELS, (void **)&dmodels, sizeof(dmodel_t));
   numvertexes = CopyLump (LUMP_VERTEXES, (void **)&dvertexes, sizeof(dvertex_t));
   numplanes = CopyLump (LUMP_PLANES, (void **)&dplanes, sizeof(dplane_t));
   numleafs = CopyLump (LUMP_LEAFS, (void **)&dleafs, sizeof(dleaf_t));
   numnodes = CopyLump (LUMP_NODES, (void **)&dnodes, sizeof(dnode_t));
   numtexinfo = CopyLump (LUMP_TEXINFO, (void **)&texinfo, sizeof(texinfo_t));
   numclipnodes = CopyLump (LUMP_CLIPNODES, (void **)&dclipnodes, sizeof(dclipnode_t));
   numfaces = CopyLump (LUMP_FACES, (void **)&dfaces, sizeof(dface_t));
   nummarksurfaces = CopyLump (LUMP_MARKSURFACES, (void **)&dmarksurfaces, sizeof(dmarksurfaces[0]));
   numsurfedges = CopyLump (LUMP_SURFEDGES, (void **)&dsurfedges, sizeof(dsurfedges[0]));
   numedges = CopyLump (LUMP_EDGES, (void **)&dedges, sizeof(dedge_t));
	}

if( LoadFlags_i & KB_LD_BspTextures)
   texdatasize = CopyLump (LUMP_TEXTURES, (void **)&dtexdata, 1);
if( LoadFlags_i & KB_LD_BspVis)
   visdatasize = CopyLump (LUMP_VISIBILITY, (void **)&dvisdata, 1);
if( LoadFlags_i & KB_LD_BspLighting)
   lightdatasize = CopyLump (LUMP_LIGHTING, (void **)&dlightdata, 1);
if( LoadFlags_i & KB_LD_BspEntities)
	entdatasize = CopyLump (LUMP_ENTITIES, (void **)&dentdata, 1);

   swapBSPFile (false);

//   dmodels_checksum = FastChecksum( (char *)dmodels, nummodels*sizeof(dmodels[0]) );
//   dvertexes_checksum = FastChecksum( (char *)dvertexes, numvertexes*sizeof(dvertexes[0]) );
//   dplanes_checksum = FastChecksum( (char *)dplanes, numplanes*sizeof(dplanes[0]) );
//   dleafs_checksum = FastChecksum( (char *)dleafs, numleafs*sizeof(dleafs[0]) );
//   dnodes_checksum = FastChecksum( (char *)dnodes, numnodes*sizeof(dnodes[0]) );
//   texinfo_checksum = FastChecksum( (char *)texinfo, numtexinfo*sizeof(texinfo[0]) );
//   dclipnodes_checksum = FastChecksum( (char *)dclipnodes, numclipnodes*sizeof(dclipnodes[0]) );
//   dfaces_checksum = FastChecksum( (char *)dfaces, numfaces*sizeof(dfaces[0]) );
//   dmarksurfaces_checksum = FastChecksum( (char *)dmarksurfaces, nummarksurfaces*sizeof(dmarksurfaces[0]) );
//   dsurfedges_checksum = FastChecksum( (char *)dsurfedges, numsurfedges*sizeof(dsurfedges[0]) );
//   dedges_checksum = FastChecksum( (char *)dedges, numedges*sizeof(dedges[0]) );
//   dtexdata_checksum = FastChecksum( (char *)dtexdata, numedges*sizeof(dtexdata[0]) );
//   dvisdata_checksum = FastChecksum( (char *)dvisdata, visdatasize*sizeof(dvisdata[0]) );
//   dlightdata_checksum = FastChecksum( (char *)dlightdata, lightdatasize*sizeof(dlightdata[0]) );
//   dentdata_checksum = FastChecksum( (char *)dentdata, entdatasize*sizeof(dentdata[0]) );

}
//
#ifndef MI_NOBSPWRITE
//============================================================================
void AddLump (int lumpnum, void *data, int len)
{
   lump_t *lump;
   lump = &header->lumps[lumpnum];
   lump->fileofs = LittleLong( ftell(wadfile) );
   lump->filelen = LittleLong(len);
   SafeWrite (wadfile, (byte*)data, (len+3)&~3);
}

/*
=============
WriteBSPFile

Swaps the bsp file in place, so it should not be referenced again
=============
*/
void HLBspFile_ct:: writeFile (char *filename)
{       
   header = &outheader;
   memset (header, 0, sizeof(dheader_t));

   swapBSPFile (true);

   header->version = LittleLong (BSPVERSION);

   wadfile = SafeOpenWrite (filename);
   SafeWrite (wadfile, (byte*)header, sizeof(dheader_t)); // overwritten later

   AddLump (LUMP_PLANES, dplanes, numplanes*sizeof(dplane_t));
   AddLump (LUMP_LEAFS, dleafs, numleafs*sizeof(dleaf_t));
   AddLump (LUMP_VERTEXES, dvertexes, numvertexes*sizeof(dvertex_t));
   AddLump (LUMP_NODES, dnodes, numnodes*sizeof(dnode_t));
   AddLump (LUMP_TEXINFO, texinfo, numtexinfo*sizeof(texinfo_t));
   AddLump (LUMP_FACES, dfaces, numfaces*sizeof(dface_t));
   AddLump (LUMP_CLIPNODES, dclipnodes, numclipnodes*sizeof(dclipnode_t));
   AddLump (LUMP_MARKSURFACES, dmarksurfaces, nummarksurfaces*sizeof(dmarksurfaces[0]));
   AddLump (LUMP_SURFEDGES, dsurfedges, numsurfedges*sizeof(dsurfedges[0]));
   AddLump (LUMP_EDGES, dedges, numedges*sizeof(dedge_t));
   AddLump (LUMP_MODELS, dmodels, nummodels*sizeof(dmodel_t));

   AddLump (LUMP_LIGHTING, dlightdata, lightdatasize);
   AddLump (LUMP_VISIBILITY, dvisdata, visdatasize);
   AddLump (LUMP_ENTITIES, dentdata, entdatasize);
   AddLump (LUMP_TEXTURES, dtexdata, texdatasize);

   fseek (wadfile, 0, SEEK_SET);
   SafeWrite (wadfile, (byte*)header, sizeof(dheader_t));
   fclose (wadfile);
}
#endif// MI_NOBSPWRITE

#ifndef MI_NOBSPENTITIES
/*
=================
ParseEpair
=================
*/
epair_t *ParseEpair (void)
{
   epair_t  *e= new epair_t;
   if (strlen(token) >= MAX_KEY-1)
      Error ("ParseEpar: token too long");
   e->key = new char[strlen(token)+1];
   strcpy(e->key, token);
   GetToken (false);
   if (strlen(token) >= MAX_VALUE-1)
      Error ("ParseEpar: token too long");
   e->value = new char[strlen(token)+1];
   strcpy (e->value, token);

   return e;
}
/*
================
ParseEntity
================
*/
int  parseEntity (HLBspFile_ct * pHLBspFile_i)
{
   epair_t  *e;
   entity_t *mapent;

   if (!GetToken (true))
      return 0;

   if (strcmp (token, "{") )
      Error ("ParseEntity: { not found");

   if ( pHLBspFile_i-> num_entities == MAX_MAP_ENTITIES)
      Error ("pHLBspFile_i-> num_entities == MAX_MAP_ENTITIES");

   mapent = & pHLBspFile_i-> entities[ pHLBspFile_i-> num_entities];
   pHLBspFile_i-> num_entities++;

   do
   {
      if (!GetToken (true))
         Error ("ParseEntity: EOF without closing brace");
      if (!strcmp (token, "}") )
         break;
      e = ParseEpair ();
      e->next = mapent->epairs;
      mapent->epairs = e;
   } while (1);

   return 1;
}

/*
================
ParseEntities

Parses the dentdata string into entities
================
*/
void parseEntities (HLBspFile_ct * pHLBspFile_i)
{
   if( !pHLBspFile_i-> dentdata)
		return;
	pHLBspFile_i-> num_entities = 0;
   ParseFromMemory (pHLBspFile_i-> dentdata, pHLBspFile_i-> entdatasize);

   while (parseEntity (pHLBspFile_i));
}
//
#ifndef MI_NOBSPWRITE
/*
================
UnparseEntities

Generates the dentdata string from all the entities
================
*/
void unparseEntities (HLBspFile_ct * pHLBspFile_i)
{
   char  *buf, *end;
   epair_t  *ep;
   char  line[2048];
   int   i;

   buf = pHLBspFile_i-> dentdata;
   end = buf;
   *end = 0;

   for (i=0 ; i< pHLBspFile_i-> num_entities ; i++)
   {
      ep = pHLBspFile_i-> entities[i].epairs;
      if (!ep)
         continue;   // ent got removed
        
      strcat (end,"{\n");
      end += 2;
        
      for (ep = pHLBspFile_i-> entities[i].epairs ; ep ; ep=ep->next)
      {
         sprintf (line, "\"%s\" \"%s\"\n", ep->key, ep->value);
         strcat (end, line);
         end += strlen(line);
      }
      strcat (end,"}\n");
      end += 2;

      if (end > buf + MAX_MAP_ENTSTRING)
         Error ("Entity text too long");
   }
   pHLBspFile_i-> entdatasize = end - buf + 1;
}
#endif// MI_NOBSPWRITE
//---------------------------------------------------------
void SetKeyValue (entity_t *ent, char *key, char *value)
{
   epair_t  *ep;
   epair_t  *prev_ep = NULL;

   for (ep=ent->epairs ; ep ; ep=ep->next)
   {
      if (!strcmp (ep->key, key) )
      {
         delete []ep->value;
   		ep->value = new char[strlen(value)+1];
			strcpy (ep->value, value);
        return;
      }
   }

   prev_ep = NULL;
   for (ep=ent->epairs ; ep ; prev_ep=ep, ep=ep->next)
      ;  // get to the end of the linked list

   ep= new epair_t;
   if (prev_ep)
      prev_ep->next = ep;  // link it at the end
   else
      ent->epairs = ep;  // link it at the beginning (first and only)

   ep->key = new char[strlen(key)+1];
   strcpy(ep->key, key);
   ep->value = new char[strlen(value)+1];
   strcpy(ep->value, value);
}
//--------------------------------------------------------
void RemoveKey(entity_t *ent, char *key)
	{
   epair_t  *prev_ep = NULL;
   epair_t  *ep=ent-> epairs ;
	while( ep)
	   {
      if (!strcmp (ep->key, key) )
		   {
         if (prev_ep)
	         prev_ep->next= ep->next;
         else
            ent->epairs= ep->next;
			delete ep;
         return;
			}
		prev_ep= ep;
		ep=ep->next;
		}
	}
//-------------------------------------------
int findEntityByClassname( HLBspFile_ct * pHLBspFile_i, int index, const char *classname)
{
epair_t  *ep;
   // index should be -1 to start at first entity...
index++;
while (index < pHLBspFile_i-> num_entities)
   {
   ep = pHLBspFile_i-> entities[index]. epairs;
   while (ep)
      {
      if ((stricmp(ep->key, "classname") == 0) &&
         (stricmp(ep->value, classname) == 0))
         {
         return index;
         }
       ep = ep->next;
      }
   index ++;
   }
return -1;  // entity not found
}
//-------------------------------------------
int findEntityByWildcard(HLBspFile_ct * pHLBspFile_i, int index, const char *classname, int length)
{
epair_t  *ep;

   // index should be -1 to start at first entity...
index++;

while (index < pHLBspFile_i-> num_entities)
   {
   ep = pHLBspFile_i-> entities[index].epairs;
   while (ep)
      {
      if ((stricmp(ep->key, "classname") == 0) &&
         (strncmp(ep->value, classname, length) == 0))
         {
         return index;
         }
      ep = ep->next;
      }
   index ++;
   }
return -1;  // entity not found
}
//-------------------------------------------
int findEntity(HLBspFile_ct * pHLBspFile_i, int index, condOnEpair_ft condOnEpair_i)
{
epair_t  *ep;

   // index should be -1 to start at first entity...
index++;

while (index < pHLBspFile_i-> num_entities)
   {
   ep = pHLBspFile_i-> entities[index].epairs;
   while (ep)
      {
      if (condOnEpair_i(ep))
         return index;
      ep = ep->next;
      }
   index ++;
   }
return -1;  // entity not found
}
//---------------------------------------------------
char *ValueForKey (entity_t *ent, char *key)
{
   epair_t  *ep;
   for (ep=ent->epairs ; ep ; ep=ep->next)
      if (!strcmp (ep->key, key) )
         return ep->value;
   return "";
}
//---------------------------------------------------
vec_t FloatForKey (entity_t *ent, char *key)
{
   char  *k = ValueForKey (ent, key);
   return (float)atof(k);
}
//---------------------------------------------------
void GetVectorForKey (entity_t *ent, char *key, vecd3_t vec)
{
   char  *k;
   double   v1, v2, v3;

   k = ValueForKey (ent, key);
	// scanf into doubles, then assign, so it is vec_t size independent
   v1 = v2 = v3 = 0;
   sscanf (k, "%lf %lf %lf", &v1, &v2, &v3);
   vec[0] = (float)v1;
   vec[1] = (float)v2;
   vec[2] = (float)v3;
}
#endif //IMPL_ENTITY

#ifndef MI_NOBSPFILESIZEPRINT
//============================================================================
//#define ENTRIES(a)   (sizeof(a)/sizeof(*(a)))
//#define ENTRYSIZE(a) (sizeof(*(a)))

int ArrayUsage( char *szItem, int items, int maxitems, int itemsize )
{
   float percentage = maxitems ? items * (float)100.0 / maxitems : (float)0.0;

   printf("%-12s  %7i/%-7i  %7i/%-7i  (%4.1f%%)", 
      szItem, items, maxitems, items * itemsize, maxitems * itemsize, percentage );
   if ( percentage > 80.0 )
      printf( "VERY FULL!\n" );
   else if ( percentage > 95.0 )
      printf( "SIZE DANGER!\n" );
   else if ( percentage > 99.9 )
      printf( "SIZE OVERFLOW!!!\n" );
   else
      printf( "\n" );
   return items * itemsize;
}

GlobUsage( char *szItem, int itemstorage, int maxstorage )
{
   float percentage = maxstorage ? itemstorage * (float)100.0 / maxstorage : (float)0.0;
   printf("%-12s     [variable]    %7i/%-7i  (%4.1f%%)", 
          szItem, itemstorage, maxstorage, percentage );
   if ( percentage > 80.0 )
      printf( "VERY FULL!\n" );
   else if ( percentage > 95.0 )
      printf( "SIZE DANGER!\n" );
   else if ( percentage > 99.9 )
      printf( "SIZE OVERFLOW!!!\n" );
   else
      printf( "\n" );
   return itemstorage;
}
/*
=============
PrintBSPFileSizes

Dumps info about current file
=============
*/
void HLBspFile_ct:: printBSPFileSizes (void)
{
   int numtextures = texdatasize ? ((dmiptexlump_t*)dtexdata)->nummiptex : 0;
   int totalmemory = 0;

   printf("\n");
   printf("Object names  Objects/Maxobjs  Memory / Maxmem  Fullness\n" );
   printf("------------  ---------------  ---------------  --------\n" );

   totalmemory += ArrayUsage( "models", nummodels, MAX_MAP_MODELS, sizeof(dmodel_t) );
   totalmemory += ArrayUsage( "planes", numplanes, MAX_MAP_PLANES, sizeof(dplane_t) );
   totalmemory += ArrayUsage( "vertexes", numvertexes, MAX_MAP_VERTS, sizeof(dvertex_t) );
   totalmemory += ArrayUsage( "nodes", numnodes, MAX_MAP_NODES, sizeof(dnode_t) );
   totalmemory += ArrayUsage( "texinfos", numtexinfo, MAX_MAP_TEXINFO, sizeof(texinfo_t) );
   totalmemory += ArrayUsage( "faces", numfaces, MAX_MAP_FACES, sizeof(dface_t) );
   totalmemory += ArrayUsage( "clipnodes", numclipnodes, MAX_MAP_CLIPNODES, sizeof(dclipnode_t) );
   totalmemory += ArrayUsage( "leaves", numleafs, MAX_MAP_LEAFS, sizeof(dleaf_t) );
   totalmemory += ArrayUsage( "marksurfaces", nummarksurfaces, MAX_MAP_MARKSURFACES, sizeof(unsigned short) );
   totalmemory += ArrayUsage( "surfedges", numsurfedges, MAX_MAP_SURFEDGES, sizeof(int) );
   totalmemory += ArrayUsage( "edges", numedges, MAX_MAP_EDGES, sizeof(dedge_t) );

   totalmemory += GlobUsage( "texdata", texdatasize, MAX_MAP_MIPTEX );
   totalmemory += GlobUsage( "lightdata", lightdatasize, MAX_MAP_LIGHTING );
   totalmemory += GlobUsage( "visdata", visdatasize, MAX_MAP_VISIBILITY );
   totalmemory += GlobUsage( "entdata", entdatasize, MAX_MAP_ENTSTRING );

   printf( "=== Total BSP file data space used: %d bytes ===\n", totalmemory );
}
#endif

