/***
*
*  Copyright (c) 1998, Valve LLC. All rights reserved.
*
*  This product contains software technology licensed from Id 
*  Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*  All Rights Reserved.
*
****/
#include <iostream>
using namespace std;

#include "mathlib_tpl.h"


#ifndef BSPFILE_H
#define BSPFILE_H

// upper design bounds

#ifndef MI_NOBSPFILESIZEPRINT
#define MAX_MAP_HULLS      4

#define MAX_MAP_MODELS     400
#define MAX_MAP_BRUSHES    4096
#define MAX_MAP_ENTITIES   1024
#define MAX_MAP_ENTSTRING  (128*1024)

#define MAX_MAP_PLANES     32767
#define MAX_MAP_NODES      32767 // because negative shorts are contents
#define MAX_MAP_CLIPNODES  32767 //
#define MAX_MAP_LEAFS      8192
#define MAX_MAP_VERTS      65535
#define MAX_MAP_FACES      65535
#define MAX_MAP_MARKSURFACES 65535
#define MAX_MAP_TEXINFO    8192
#define MAX_MAP_EDGES      256000
#define MAX_MAP_SURFEDGES  512000
#define MAX_MAP_TEXTURES   512
#define MAX_MAP_MIPTEX     0x200000
#define MAX_MAP_LIGHTING   0x200000
#define MAX_MAP_VISIBILITY 0x200000

#define MAX_MAP_PORTALS    65536
#endif
// key / value pair sizes

#define MAX_KEY   32
#define MAX_VALUE 4096

#ifndef byte
typedef unsigned char byte;
#endif
//=============================================================================


#define BSPVERSION   30
#define TOOLVERSION  2

//++++++++++++++++++++++++++++++++++++++++++
typedef struct
{
   int   fileofs, filelen;
} lump_t;

#define LUMP_ENTITIES   0
#define LUMP_PLANES     1
#define LUMP_TEXTURES   2
#define LUMP_VERTEXES   3
#define LUMP_VISIBILITY 4
#define LUMP_NODES      5
#define LUMP_TEXINFO    6
#define LUMP_FACES      7
#define LUMP_LIGHTING   8
#define LUMP_CLIPNODES  9
#define LUMP_LEAFS      10
#define LUMP_MARKSURFACES 11
#define LUMP_EDGES      12
#define LUMP_SURFEDGES  13
#define LUMP_MODELS     14

#define HEADER_LUMPS    15
//++++++++++++++++++++++++++++++++++++++++++
typedef struct
{
   float    mins[3], maxs[3];
   float    origin[3];
   int      headnode[MAX_MAP_HULLS];
   int      visleafs;   // not including the solid leaf 0
   int      firstface, numfaces;
} dmodel_t;
//++++++++++++++++++++++++++++++++++++++++++
typedef struct
{
   int      version;
   lump_t   lumps[HEADER_LUMPS];
} dheader_t;
//++++++++++++++++++++++++++++++++++++++++++
typedef struct
{
   int      nummiptex;
   int      dataofs[4]; // [nummiptex]
} dmiptexlump_t;
//++++++++++++++++++++++++++++++++++++++++++
#define MIPLEVELS 4
//++++++++++++++++++++++++++++++++++++++++++
typedef struct miptex_s
{
   char     name[16];
   unsigned width, height;
   unsigned offsets[MIPLEVELS];  // four mip maps stored
} miptex_t;
//++++++++++++++++++++++++++++++++++++++++++
typedef struct
{
   float point[3];
} dvertex_t;
//++++++++++++++++++++++++++++++++++++++++++


// 0-2 are axial planes
#define PLANE_X   0
#define PLANE_Y   1
#define PLANE_Z   2

// 3-5 are non-axial planes snapped to the nearest
#define PLANE_ANYX   3
#define PLANE_ANYY   4
#define PLANE_ANYZ   5
//++++++++++++++++++++++++++++++++++++++++++
typedef struct
{
float normal[3];
float dist;
int   type; // PLANE_X - PLANE_ANYZ ?remove? trivial to regenerate
} dplane_t;
//++++++++++++++++++++++++++++++++++++++++++


#define CONTENTS_EMPTY  -1
#define CONTENTS_SOLID  -2
#define CONTENTS_WATER  -3
#define CONTENTS_SLIME  -4
#define CONTENTS_LAVA   -5
#define CONTENTS_SKY    -6
#define CONTENTS_ORIGIN -7 // removed at csg time
#define CONTENTS_CLIP   -8 // changed to contents_solid

#define CONTENTS_CURRENT_0    -9
#define CONTENTS_CURRENT_90   -10
#define CONTENTS_CURRENT_180  -11
#define CONTENTS_CURRENT_270  -12
#define CONTENTS_CURRENT_UP   -13
#define CONTENTS_CURRENT_DOWN -14

#define CONTENTS_TRANSLUCENT  -15

//++++++++++++++++++++++++++++++++++++++++++
// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct
{
   int      planenum; 
   short    children[2];   // negative numbers are -(leafs+1), not nodes
   short    mins[3];       // for sphere culling
   short    maxs[3];
   unsigned short firstface;
   unsigned short numfaces;   // counting both sides
} dnode_t;
//++++++++++++++++++++++++++++++++++++++++++
typedef struct
{
   int      planenum;
   short    children[2];   // negative numbers are contents
} dclipnode_t;
//++++++++++++++++++++++++++++++++++++++++++
typedef struct texinfo_s
{
   float    vecs[2][4];    // [s/t][xyz offset]
   int      miptex;
   int      flags;
} texinfo_t;
//++++++++++++++++++++++++++++++++++++++++++
#define TEX_SPECIAL  1     // sky or slime, no lightmap or 256 subdivision
//++++++++++++++++++++++++++++++++++++++++++
// note that edge 0 is never used, because negative edge nums are used for
// counterclockwise use of the edge in a face
typedef struct
{
   unsigned short v[2];    // vertex numbers
} dedge_t;
//++++++++++++++++++++++++++++++++++++++++++
#define MAXLIGHTMAPS 4
//++++++++++++++++++++++++++++++++++++++++++
typedef struct
{
   short    planenum;
   short    side;

   int      firstedge;     // we must support > 64k edges
   short    numedges;
   short    texinfo;

// lighting info
   byte     styles[MAXLIGHTMAPS];
   int      lightofs;      // start of [numstyles*surfsize] samples
} dface_t;
//---------------------------------------------

#define AMBIENT_WATER   0
#define AMBIENT_SKY     1
#define AMBIENT_SLIME   2
#define AMBIENT_LAVA    3

#define NUM_AMBIENTS    4     // automatic ambient sounds

//++++++++++++++++++++++++++++++++++++++++++
// leaf 0 is the generic CONTENTS_SOLID leaf, used for all solid areas
// all other leafs need visibility info
typedef struct
{
   int      contents;
   int      visofs;        // -1 = no visibility info

   short    mins[3];       // for frustum culling
   short    maxs[3];

   unsigned short    firstmarksurface;
   unsigned short    nummarksurfaces;

   byte     ambient_level[NUM_AMBIENTS];
} dleaf_t;


//============================================================================
#define ANGLE_UP     -1
#define ANGLE_DOWN   -2

//++++++++++++++++++++++++++++++++++++++++++++++++++++
int FastChecksum(char *buffer, int bytes);

void DecompressVis (byte *in, byte *decompressed);
int CompressVis (byte *vis, byte *dest);
int CopyLump (int lump, void **dest, int size);

//++++++++++++++++++++++++++++++++++
struct epair_t
{
   epair_t *next;
   char  *key;
   char  *value;
	epair_t(): key(0), value(0), next(0) {}
	~epair_t() { delete[] key; delete[] value; }
};
//++++++++++++++++++++++++++++++++++
struct entity_t
{
   vec3_t      origin;
   int         firstbrush;
   int         numbrushes;
   epair_t     *epairs;
	entity_t(): epairs(0) { }
	~entity_t() 
		{	clear();	}
	void clear()
		{
      epair_t * pEpair_a;
		while( pEpair_a= epairs)
			{
         epairs= pEpair_a->next;
         delete pEpair_a;
			}
		}
};

//+++++++++++++++++++++++++++++++++++++++
class HLBspFile_ct
{
public:
static enum{ KB_LD_BspCore= 1, KB_LD_BspEntities= 2
	, KB_LD_BspTextures=4, KB_LD_BspVis= 8, KB_LD_BspLighting= 16};//loading flags

int      nummodels;
dmodel_t *dmodels;

int      visdatasize;
byte    *dvisdata;

int      lightdatasize;
byte    *dlightdata;

int      texdatasize;
byte    *dtexdata;

int      entdatasize;
char    *dentdata;

int        numtexinfo;
texinfo_t *texinfo;

 int        numleafs;
 dleaf_t   *dleafs;

 int        numplanes;
 dplane_t  *dplanes;

 int        numvertexes;
 dvertex_t *dvertexes;

 int        numnodes;
 dnode_t   *dnodes;

 int        numfaces;
 dface_t   *dfaces;

 int        numclipnodes;
 dclipnode_t *dclipnodes;

 int        numedges;
 dedge_t   *dedges;

 int        nummarksurfaces;
 unsigned short *dmarksurfaces;

 int        numsurfedges;
 int       *dsurfedges;

int        num_entities;
entity_t   entities[MAX_MAP_ENTITIES];

	HLBspFile_ct() { init(); }
	virtual ~HLBspFile_ct() {  clear(); }

	void loadFile (char *filename, unsigned LoadFlags_i);
	bool loadPakBSPFile  (char *filename, unsigned LoadFlags_i);
	void writeFile (char *filename);
	void printBSPFileSizes (void);
	void initForBuild( const HLBspFile_ct & BspSource_i);

	int          compressVis (byte *vis, byte *dest);
	void         decompressVis (byte *in, byte *decompressed);

	int          findLeaf( const vec3_t Coords_i);
	int          getPointContents( const vec3_t Coord_i);

	texinfo_t *  getFaceTexInfo( dface_t * pFace_i);
	texinfo_t *  getFaceTexInfo( int FaceIz_i);
	int          getTexIndex( dface_t * pFace_i);
	unsigned int getFaceFlags( int face_index_i);
	
	int          getFirstVtxId( dface_t * pFace_i, int Index_i) const;//, dedge_t* & pEdge_o);
//	int			 getSafeNormal( dface_t * pFace_i, Vector & Normal_o);

	void outText_entities(void);
	void outText_edges(void);
	void outText_faces(void);
	void outText_miptex(void);
	void outText_texinfo(void);
	void outText_nodes(void);
	void outText_leafs(void);
	void outText_planes(void);
	void outText_vertices(void);
	void outText_models(void);
	void outText_stats(FILE *);
	void outText_leafsPlanes2(void);	

	void outText_entity( std::ostream & OS_i, int i);
	void outText_entities( std::ostream & OS_i);
protected:
	virtual void init();
	virtual void clear();
	void swapBSPFile(int todisk);
	void swapBSPFileModels(int todisk);
	void swapBSPFileTexinfo(int todisk);
	void swapBSPFileMiptex (int todisk);
	void loadFile0 ( unsigned LoadFlags_i);
};

void SetKeyValue (entity_t *ent, char *key, char *value);
void RemoveKey(entity_t *ent, char *key);
char *ValueForKey (entity_t *ent, char *key);
vec_t FloatForKey (entity_t *ent, char *key);
void GetVectorForKey (entity_t *ent, char *key, vecd3_t vec);

epair_t *ParseEpair (void);
void parseEntities (HLBspFile_ct *);
void loadEntVars(HLBspFile_ct *);
int  findEntityByClassname( HLBspFile_ct *, int index, const char *classname);
int  findEntityByWildcard( HLBspFile_ct *,int index, const char *classname, int length);
bool loadTextures(HLBspFile_ct *);
void freeTextures();
void animateTextures();
int  initSpawnPoint( HLBspFile_ct * );

enum { KB_FD_Walkable= 1<< 0, KB_FD_Sky=1<<1, KB_FD_Last= KB_FD_Sky};
//
inline int HLBspFile_ct:: getFirstVtxId( dface_t * pFace_i, int Index_i) const
	{
	int SurfEdge_a= this-> dsurfedges[pFace_i->firstedge + Index_i];
	return SurfEdge_a< 0? this-> dedges[-SurfEdge_a]. v[ 1]: this-> dedges[SurfEdge_a]. v[ 0];
	}
//------------------------------------------------------
inline texinfo_t * HLBspFile_ct:: getFaceTexInfo( dface_t * pFace_i)
	{
	return & this-> texinfo[ pFace_i-> texinfo];
	//return texinfo + pFace_i-> texinfo;
	}
//------------------------------------------------------
inline texinfo_t * HLBspFile_ct:: getFaceTexInfo( int FaceIz_i)
	{
	return getFaceTexInfo( & this-> dfaces[FaceIz_i]);
	}
//------------------------------------------------------
inline int HLBspFile_ct:: getTexIndex( dface_t * pFace_i)
	{
	return getFaceTexInfo(pFace_i)-> miptex;
	}

//++++++++++++++++++++++++++++++++++++++++++++++++++
#endif  // BSPFILE_H
