// bsptrace.cpp 60
// Copyright (C) 2004 - Samplay
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "appgene.h"
#include "math.h"

#include "bspfile2.h"
//#include "mathlib.h"
#include "mathlib_tpl.h"

#include "bsptrace.h"

#ifdef M_DEBUGC
#define DBGASSERT(a) assert(a)
#define M_DBGLINE
#else
#define DBGASSERT(a) 
#define M_DBGLINE /##/
#endif

extern HLBspFile_ct HLBspFile;

char * getTextureName( HLBspFile_ct * pHLBspFile_i, dface_t * pFace_i);
/*
#define MF_FLOAT_AS_BITS(a) (*(int*) &(a))
#define MF_F_ABS(a)  (*(float*)(MF_FLOAT_AS_BITS(a) & 0x7fffffff))
#define MF_F_LT_ZERO(a) (MF_FLOAT_AS_BITS(a) & 0x80000000)
#define MF_F_GE_ZERO(a) (!MF_F_LT_ZERO(a))
#define MF_F_NEQ_ZERO(a) (MF_FLOAT_AS_BITS(a) & 0x7fffffff)
#define MF_F_EQ_ONE(a)  (MF_FLOAT_AS_BITS(a) & 0x3F800000)
#define MF_F_NEQ_ONE(a)  (MF_FLOAT_AS_BITS(a) ^ 0x3F800000)
*/
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

#define IMPL_CHECKTRACE
#define IMPL_TRACETOFILE

#ifndef M_DEBUGC
#undef IMPL_CHECKTRACE
#undef IMPL_TRACETOFILE
#endif

//++++++++++++++++++++++++++++++++++
struct TraceParam_st
{
//int    NodeNum;
float  StartFrac;
float  EndFrac;
vec3_t StartPoint;
vec3_t EndPoint;
int    StartPlane;
int    EndPlane;
};

const float	K_EpsilonTrace= 0.01f;

#ifdef IMPL_TRACETOFILE
char dbgTraceDataPFE[]= "_TraceData.txt";
static FILE * pFile_s;
static int DbgTraceSave_s= 0;
#endif

//============================================================
//-----------------------------------------------------------
void traceLineBSP(const vec3_t start, const vec3_t end, TraceData_st & Tr_o)
{
Trace_ct(&HLBspFile). traceMove( start, end, Tr_o);
}
//-----------------------------------------------------------
void Trace_ct:: traceMove( const vec3_t Start_i, const vec3_t End_i, TraceData_st & Tr_o)
{
#ifdef IMPL_TRACETOFILE
remove(dbgTraceDataPFE);
#endif

VectorCopy( Start_i, Start);
VectorCopy( End_i, End);
/*
Start[0]=;
Start[1]=;
Start[2]=;
End[0]=;
End[1]=;
End[2]=;
*/

dleaf_t  * startleaf;
int StartLeafNodeNumber_a= pHLBspFile-> findLeaf(Start);
startleaf= & pHLBspFile-> dleafs[ StartLeafNodeNumber_a];

memset(&Tr_o, 0x00, sizeof(TraceData_st));
//Tr_o. pPlane= 0;
Tr_o. planenum= -1;
Tr_o. FaceNum= -1;
//Tr_o. TextureName= NULL;
VectorCopy( End, Tr_o. endpos);
Tr_o. StartLeafNum= StartLeafNodeNumber_a;
Tr_o. fraction= 1.0f;
Tr_o. StartContents= startleaf-> contents;
if( startleaf-> contents== CONTENTS_SOLID)
   Tr_o.startsolid= TRUE;

TmpTrace. planenum= -1;
TmpTrace. fraction= 1.0f;
TmpTrace. FurthestNonBlockingLeafNum= -1;
int NodeNum_a= 0;// start at the root of the BSP tree
   	
TraceParam_st TraceParam_a;
//TraceParam_a. NodeNum= NodeNum_a;
TraceParam_a. StartFrac= 0.0f;
TraceParam_a. EndFrac= 1.0f;
VectorCopy(Start, TraceParam_a. StartPoint);
VectorCopy(End, TraceParam_a. EndPoint);
TraceParam_a. StartPlane= -1;
TraceParam_a. EndPlane= -1;
	
//	Type= K_Point;  traceMovePointR( NodeNum_a, TraceParam_a, Tr_o);
traceMoveR( NodeNum_a, TraceParam_a, Tr_o);

if( Tr_o.fraction!= 1.0f && Tr_o.fraction!= 0.0f)
	{
	Tr_o. pPlane=  & pHLBspFile-> dplanes[ Tr_o. planenum];
	memcpy( & Tr_o. plane, &pHLBspFile-> dplanes[ Tr_o. planenum], sizeof(dplane_t));
	VectorInterp( Start, End, Tr_o. fraction, Tr_o. endpos);

	vec3_t Delta_a;
	VectorSubtract(End, Start, Delta_a);
	float Length_a= VectorLength(Delta_a);
	Tr_o. HitDist= Length_a* Tr_o. fraction;

	if( Type== K_Point)
		{
M_DBGLINE	Flags|= (KB_Face|KB_Texture);
		if( Flags & (KB_Face|KB_Texture))
			{
			TraceData_st TraceF_a;
			memcpy( &TraceF_a, & Tr_o, sizeof(TraceData_st));
/*
			memcpy( &TraceF_a, & Tr_o, sizeof(TraceData_st));
		//off plane, normal to plane && closer to start
//			Offset_a= -Offset_a;
//			vec3_t V_a;
//			VectorScale( pPlane_a-> normal, Offset_a, V_a);
			VectorSubtract( TraceF_a. endpos, V_a, TraceF_a. endpos);
			Tr_o. FaceNum= traceMoveFindFace( Start, TraceF_a);
*/
			Tr_o. FaceNum= traceMoveFindFace( Start, TraceF_a);

			if( Tr_o. FaceNum== -2) 
				{		
				char Msg_a[256];
				sprintf(Msg_a,"TraceBUG: No Face on Plane; Start= %10.4f %10.4f %10.4f End=%10.4f %10.4f %10.4f\n"
					, Start[0], Start[1], Start[2], 	End[0], End[1], End[2] );
				affDbgMsg(1,Msg_a);
				}
			else if( Tr_o. FaceNum== -1) 
				{		
				char Msg_a[256];
				sprintf(Msg_a,"TraceBUG: Endpos not on Face; Start= %10.4f %10.4f %10.4f End=%10.4f %10.4f %10.4f\n"
					, Start[0], Start[1], Start[2], 	End[0], End[1], End[2] );
				affDbgMsg(1,Msg_a);
				}

			if( Flags & KB_Texture && Tr_o. FaceNum>= 0)
				Tr_o. TextureName= getTextureName( pHLBspFile, & pHLBspFile-> dfaces[Tr_o. FaceNum]);
			}
			
		// endpos moved slightly off plane, normal to plane && in blocking leaf; 
		// this ensure that finLeaf(endpos) returns the same leaf number as BlockingLeafNum
		TraceData_st Trace_a;
		memcpy( &Trace_a, & Tr_o, sizeof(TraceData_st));
		dplane_t * pPlane_a= & pHLBspFile-> dplanes[ Tr_o.planenum];
		float Offset_a= DotProduct( pPlane_a-> normal, Delta_a);
		if( Offset_a> 0)// change to "if( Offset_a< 0)" if endpos needs to be IN the NON blocking leaf
			Offset_a= K_EpsilonTrace;
		else
			Offset_a= -K_EpsilonTrace;
		vec3_t V_a;
		VectorScale( pPlane_a-> normal, Offset_a, V_a);
		VectorAdd( Trace_a. endpos, V_a, Trace_a. endpos);
		int LeafF0_a= pHLBspFile-> findLeaf( Trace_a. endpos);
		DBGASSERT( Trace_a.contents!= CONTENTS_SOLID || LeafF0_a==0);
/*
		if( Trace_a.contents== CONTENTS_SOLID && LeafF0_a!=0) 
			{	affDbgMsg(1,"TraceBUG: findLeaf discrepancy on endpos: Blocking Leaf should be solid");
			}
*/
		}
	else
		{
		Tr_o.FurthestNonBlockingLeafNum= pHLBspFile-> findLeaf(Tr_o. endpos );

/*
		dplane_t * pPlane_a= & pHLBspFile-> dplanes[ Tr_o.planenum];
		float Offset_a;
		if( Type== K_AABox)
			{
			Offset_a= fabs( HullHSize[0]* pPlane_a-> normal[0])
				+ fabs( HullHSize[1]* pPlane_a-> normal[1]) 
				+ fabs( HullHSize[2]* pPlane_a-> normal[2]);
			}
		else if ( Type== K_AACylinder)
			{
			Offset_a= fabs( HullHSize[2]* pPlane_a-> normal[CylAx])
				+	Radius_()* sqrt( 1.0f- pPlane_a-> normal[CylAx]*pPlane_a-> normal[CylAx]);
			}
		else if ( Type== K_Sphere)
			Offset_a= Radius_();
		float OppDistStart_a;
		float DistSE_a= -DotProduct( pPlane_a-> normal, Start);
		OppDistStart_a= DistSE_a+ pPlane_a-> dist;
		if( OppDistStart_a<= 0)
			Offset_a= -Offset_a;
		DistSE_a+= DotProduct( pPlane_a-> normal, End);
		float Frac_a= (OppDistStart_a+ Offset_a)/ DistSE_a;
		vec3_t Endpos_a;
		VectorInterp( Start, End, Frac_a, Endpos_a);
		VectorSubtract( Endpos_a, Tr_o.endpos, Delta_a);
		float Bug_a= VectorLength( Delta_a);	
//		DBGASSERT( Bug_a< K_EpsilonTrace);
*/
		}
	}
#ifdef IMPL_TRACETOFILE
if( DbgTraceSave_s)
	{
	pFile_s= fopen(dbgTraceDataPFE,"a");
	fprintf(pFile_s," ===TRACE  Tr=%8.6f  HitDist=%7.2f Trplane#=%4i BlockLeaf#%4i(contents %2i) LastLeaf#%4i\n"
		, Tr_o. fraction, Tr_o. HitDist, Tr_o. planenum
		, Tr_o.BlockingLeafNum, Tr_o.contents
		, Tr_o.FurthestNonBlockingLeafNum);
	fclose(pFile_s);
	}
#endif
}
//---------------------------------------------------
int Trace_ct:: traceToLeaf (int LeafNodeNum_i, TraceData_st & Tr_o)
	{
   dleaf_t *pLeaf_a= & pHLBspFile-> dleafs[ LeafNodeNum_i];
   
	if ((pLeaf_a->contents== CONTENTS_SOLID)|| (pLeaf_a->contents== CONTENTS_SKY))
		{
		Tr_o. contents= pLeaf_a-> contents;
		if((pLeaf_a-> contents== CONTENTS_SOLID)&& (Tr_o.startsolid))
			Tr_o. allsolid= TRUE;	
		return 1;
		}
	else
		{
		// TO DO : trace against ents (partially) in leaf? ( solid brush models + studio models)
		// TO DO before?: findLeaf( entvar, leaflist) returning the leaf(s) the ent crosses ( at least for
		// non moving ents)
		// and be able to use tha data the other way round( from a leaf to ents that are in)
		return 0;
		}
	}
//----------------------------------------------------------------
int Trace_ct:: traceMoveFindFace  ( const vec3_t start, TraceData_st & Tr_o)
	{
	if( Tr_o. fraction== 1.0f)
		return -1;
	vec3_t v_temp;
	float min_diff = 9999.9f;
	int FaceIdx_a= -1;
	int Pb_a= -2;
	dleaf_t *endleaf = & pHLBspFile-> dleafs[Tr_o.FurthestNonBlockingLeafNum];
	unsigned short *p = pHLBspFile->  dmarksurfaces + endleaf->firstmarksurface;
	int PlaneHitNum_a= Tr_o. planenum ;
	dplane_t *pPlaneHit_a= & pHLBspFile-> dplanes [Tr_o.planenum];
	for (int i= 0; i < endleaf-> nummarksurfaces; i++)
		{
		FaceIdx_a = *p++;
		dface_t *pFace_a= & pHLBspFile-> dfaces[FaceIdx_a];
		if( pFace_a-> planenum!= PlaneHitNum_a)
			continue;
		Pb_a= -1;
		//check if inters is on face ( ray crosses plane inside face polygon)
		int Sign_a= 3;
		int SignPrev_a= 3;
      vec_t * vertex0;
		vec_t * vertex1;
		vec3_t VToIsec_a;
		vec3_t VEdge_a;
		int VtxId_a= pHLBspFile-> getFirstVtxId( pFace_a, 0);
		vertex1= pHLBspFile-> dvertexes[VtxId_a].point;
	   int edge_index= 1;
		while( (Sign_a & SignPrev_a) && edge_index< pFace_a->numedges)
         {
			vertex0= vertex1;
         VtxId_a= pHLBspFile-> getFirstVtxId( pFace_a, edge_index);
			vertex1= pHLBspFile-> dvertexes[VtxId_a].point;
			VectorSubtract(vertex0, Tr_o.endpos, VToIsec_a);
         VectorSubtract(vertex1, vertex0, VEdge_a);
			CrossProduct( VEdge_a, VToIsec_a, v_temp);
			float Dot_a= (float)DotProduct( v_temp, pPlaneHit_a-> normal);
//			if( Dot_a== 0)
			if( fabs(Dot_a)< K_EpsilonTrace)
				return FaceIdx_a;
			SignPrev_a= Sign_a;
			Sign_a= Dot_a> 0? 1 : 2;
			++edge_index;
			}
		if( Sign_a & SignPrev_a)       
			return FaceIdx_a;
		}
	return Pb_a;
	}
//-------------------------------------------------------------------
int Trace_ct:: traceMovePointR ( int NodeNum_i, TraceParam_st & TP_i, TraceData_st & Tr_o)
	{
	// this works only for point trace.
	// when starting from a solid leaf, the hit, if any, is computed as if back-face solid faces do not
	// exist. In other words, given a direction, 
	// the returned values are always the same whatever the starting point.

	//_ the "if( Tr_o. fraction> ...)" checks and using the Side_a value 
	// result in looking for the nearest ( from Start) blocking leaf
	// 
	if( NodeNum_i< 0)// in a leaf
		{
		int Ret_a= traceToLeaf ( -1- NodeNum_i, Tr_o);
		if( Ret_a== 0)//not a blocking leaf; store data in case its neighbour (further from Start) blocks
			{
			TmpTrace. planenum= TP_i. EndPlane;
			TmpTrace. fraction= TP_i. EndFrac;
			TmpTrace. FurthestNonBlockingLeafNum= -1- NodeNum_i;//this one=last ( closer to Start) non-blocking neighbour leaf of possibly blocking neighbour leaf
			return 0;			
			}
		else
			{
			Tr_o. BlockingLeafNum= -1- NodeNum_i;
			Tr_o. FurthestNonBlockingLeafNum= TmpTrace. FurthestNonBlockingLeafNum;//last leaf before hitting this one
			Tr_o. fraction= TmpTrace. fraction;
			Tr_o. planenum= TmpTrace. planenum;
			return 1;
			}
		}
	
	if( Tr_o. fraction<= TP_i. StartFrac)
		return -1;

	dnode_t	 * node= &pHLBspFile->dnodes[ NodeNum_i];
	dplane_t	 * plane= &pHLBspFile->dplanes[node->planenum];

	int Side_a;
	float DistNear_a= DotProduct( plane-> normal, TP_i. StartPoint)- plane-> dist;
	float DistFar_a= DotProduct( plane-> normal, TP_i. EndPoint)- plane-> dist;

	if( DistNear_a>= 0 && DistFar_a>= 0)
		{
		traceMovePointR( node-> children[ 0], TP_i, Tr_o);
		if( DistNear_a== 0 && DistFar_a== 0)
			traceMovePointR( node-> children[ 1], TP_i, Tr_o);
		}
	else if( DistNear_a<= 0 && DistFar_a<= 0)
		traceMovePointR( node-> children[ 1], TP_i, Tr_o);
	else
		{
		TraceParam_st TP_a;
		memcpy( & TP_a, & TP_i, sizeof( TraceParam_st));

		Side_a= DistNear_a< 0;//front/on is zero
		vec3_t Interp_a;
		float	Frac_a= -DistNear_a/(DistFar_a- DistNear_a);//intersection with plane
		float kInterp_a= TP_i. StartFrac+ (TP_i. EndFrac- TP_i. StartFrac)* Frac_a;

		DBGASSERT( TP_i. StartFrac<= kInterp_a && kInterp_a<= TP_i. EndFrac);//sometimes == due to float rounding
//		VectorInterp( Start, End, kInterp_a, Interp_a);
		VectorInterp( TP_i. StartPoint, TP_i. EndPoint, Frac_a, Interp_a);
		TP_a. EndFrac= kInterp_a;
		VectorCopy( Interp_a, TP_a. EndPoint); 
		TP_a. EndPlane= node-> planenum;
		traceMovePointR( node-> children[ Side_a], TP_a, Tr_o);
			
		if( Tr_o. fraction> kInterp_a)
			{
			memcpy( & TP_a, & TP_i, sizeof( TraceParam_st));
			TP_a. StartFrac= kInterp_a;
			VectorCopy(Interp_a, TP_a.StartPoint); 
			TP_a. StartPlane= node-> planenum;
			traceMovePointR( node-> children[ Side_a^1], TP_a, Tr_o);
			}
		}
	return -1;
	}
//-------------------------------------------------------------------
int Trace_ct:: traceMoveR ( int NodeNum_i, TraceParam_st & TP_i , TraceData_st & Tr_o)
	{
	// 1.As opposed to traceMovePoint, traces starting in a solid leaf "stop dead"
	
	// 2.For traces other than point trace ( Type!=K_Point),
	// this does not work in all cases.
	// in all cases ( I think), fraction will be less then 1 if something is hit on the way,
	// but the fraction value, endpos, ... may be meaningless when the ACTUAL hit point is not
	// a vertex ( a full edge/face span hit include >=1 vertex and then is OK)
	// For the sphere, the true hit point must be on the [Start, End] line
	// For the cylinder, the true hit point must be on one of the circle bases.

	// In other words, hits in the middle of an edge or a face result in inaccurate fraction, endpos....
	// the worst case is probably moving the hull towards an acute wegde pointing at the hull

	// It will then work 100% in particular
	//-for traces strictly within a convex closed volume.  
	//-and for AABox traces (Type== K_AABox) against axis aligned objects in the world.


	//_ the "if( Tr_o. fraction> ...)" checks and using the Side_a value 
	// result in looking for the nearest ( from Start) blocking leaf
	// 
	if( NodeNum_i< 0)// in a leaf
		{
		int Ret_a= traceToLeaf ( -1-NodeNum_i, Tr_o);
		if( Ret_a== 0)//not a blocking leaf
			{
			TmpTrace. FurthestNonBlockingLeafNum= -1-NodeNum_i;
#ifdef IMPL_TRACETOFILE
			if( DbgTraceSave_s)
				{
				pFile_s= fopen(dbgTraceDataPFE,"a");
				fprintf(pFile_s,"-TTf=%8.6f TTp#%4i TTL#%4i"
					" Tf=%8.6f  Tp#=%4i TL#%4i"
					" n#=%6i  near= %8.6f far= %8.6f\n"
					, TmpTrace. fraction, TmpTrace. planenum, TmpTrace. FurthestNonBlockingLeafNum
					, Tr_o. fraction, Tr_o. planenum, Tr_o. FurthestNonBlockingLeafNum
					, NodeNum_i, TP_i. StartFrac, TP_i. EndFrac);
				fclose(pFile_s);
				}
#endif
			return 0;			
			}
		else
			{
			if( TP_i. StartFrac<= Tr_o. fraction)
				{
				Tr_o. BlockingLeafNum= -1-NodeNum_i;
				Tr_o. FurthestNonBlockingLeafNum= TmpTrace. FurthestNonBlockingLeafNum;
				Tr_o. fraction= TP_i. StartFrac;
				Tr_o. planenum= TP_i. StartPlane;
				}
#ifdef IMPL_TRACETOFILE
			if( DbgTraceSave_s)
				{
				pFile_s= fopen(dbgTraceDataPFE,"a");
				fprintf(pFile_s,"+TTf=%8.6f TTp#%4i TTL#%4i"
					" Tf=%8.6f  Tp#=%4i TL#%4i"
					" n#=%6i  near= %8.6f far= %8.6f\n"
					, TmpTrace. fraction, TmpTrace. planenum, TmpTrace. FurthestNonBlockingLeafNum
					, Tr_o. fraction, Tr_o. planenum, Tr_o. FurthestNonBlockingLeafNum
					, NodeNum_i, TP_i. StartFrac, TP_i. EndFrac);
				fclose(pFile_s);
				}
#endif	
			return 1;
			}
		}

	float		   DistNear_a;
	float       DistFar_a;
	dnode_t	 * node= &pHLBspFile->dnodes[NodeNum_i];
	dplane_t	 * plane= &pHLBspFile->dplanes[node->planenum];


	float Offset_a=0.0f;
	if( Type== K_AABox)
		{
		Offset_a= fabs( HullHSize[0]* plane->normal[0])
			+ fabs( HullHSize[1]* plane->normal[1]) 
			+ fabs( HullHSize[2]* plane->normal[2]);
		}
	else if ( Type== K_AACylinder)
		{
		Offset_a= fabs( HullHSize[2]* plane->normal[CylAx])
			+ Radius_()* sqrt( plane->normal[CylB1]*plane->normal[CylB1]+plane->normal[CylB2]*plane->normal[CylB2]);
		}
	else if ( Type== K_Sphere)
		Offset_a= Radius_();

	int Side_a;
	DistNear_a= DotProduct( plane-> normal, TP_i. StartPoint)- plane-> dist;
	DistFar_a= DotProduct( plane-> normal, TP_i. EndPoint)- plane-> dist;

#ifdef IMPL_TRACETOFILE	
	if( DbgTraceSave_s)
		{
		pFile_s= fopen(dbgTraceDataPFE,"a");
		fprintf(pFile_s," TTf=%8.6f TTp#%4i TTL#%4i"
			" Tf=%8.6f  Tp#=%4i TL#%4i"
			" n#=%6i  p#%4i near= %8.6f far= %8.6f"
			" n0#=%6i n1#=%6i\n"
				, TmpTrace. fraction, TmpTrace. planenum, TmpTrace. FurthestNonBlockingLeafNum
				, Tr_o. fraction, Tr_o. planenum, Tr_o. FurthestNonBlockingLeafNum
				, NodeNum_i, node->planenum, TP_i. StartFrac, TP_i. EndFrac
				, node-> children[ 0], node-> children[ 1]);
		fclose(pFile_s);
		}
#endif

	if( Tr_o. fraction< TP_i. StartFrac)
		return -1;

	if( DistNear_a>= Offset_a && DistFar_a>= Offset_a)
		{
		traceMoveR( node-> children[ 0], TP_i, Tr_o);
		if( Type== K_Point && DistNear_a== 0 && DistFar_a== 0)
			traceMoveR( node-> children[ 1], TP_i, Tr_o);
		}
	else if( DistNear_a<= -Offset_a && DistFar_a<= -Offset_a)
		{
		traceMoveR( node-> children[ 1], TP_i, Tr_o);
		}
	else
		{
		Side_a= DistNear_a< DistFar_a;
		if( DistNear_a> DistFar_a)
			Offset_a= -Offset_a;
	
		TraceParam_st TP_a;
		memcpy( & TP_a, & TP_i, sizeof(TraceParam_st));
	
		float kInterpN_a;
		float FracN_a= (-DistNear_a+ Offset_a)/(DistFar_a- DistNear_a);//near: ends after plane
		if( FracN_a>= 1.0f)
			{
			kInterpN_a= TP_i. EndFrac;
			}
		else
			{
			kInterpN_a= TP_i. StartFrac+ (TP_i. EndFrac- TP_i. StartFrac)* FracN_a;
			DBGASSERT( TP_i. StartFrac<= kInterpN_a && kInterpN_a<= TP_i. EndFrac);
//			VectorInterp( Start, End, kInterp_a, Interp_a);
			VectorInterp( TP_i. StartPoint, TP_i. EndPoint, FracN_a, TP_a. EndPoint);
			}
		TP_a. EndFrac= kInterpN_a;

		int RetN_a= traceMoveR( node-> children[ Side_a], TP_a, Tr_o);

		float FracF_a= (-DistNear_a-Offset_a)/(DistFar_a- DistNear_a);//far: starts before plane
		float kInterpF_a;
		if( FracF_a<= 0.0f)
			kInterpF_a= TP_i. StartFrac;
		else
			{
			kInterpF_a= TP_i. StartFrac+ (TP_i. EndFrac- TP_i. StartFrac)* FracF_a;
			DBGASSERT( TP_i. StartFrac<= kInterpF_a && kInterpF_a<= TP_i. EndFrac);
			}
		if( Tr_o. fraction> kInterpF_a)
			{
			memcpy( & TP_a, & TP_i, sizeof(TraceParam_st));
			TP_a. StartFrac= kInterpF_a;
			TP_a. StartPlane= node-> planenum;
			if( FracF_a> 0.0f)
				VectorInterp( TP_i. StartPoint, TP_i. EndPoint, FracF_a, TP_a.StartPoint);
			traceMoveR( node-> children[ Side_a^1], TP_a, Tr_o);
			}
		}
	return -1;
	}

//
#if 0
UNTRIED
//-----------------------------------------------------------
void Trace_ct:: traceExpand( TraceData_st & Tr_o)
	{
	int StartLeafNodeNumber_a= pHLBspFile-> findLeaf(Start);
	dleaf_t  * startleaf= & pHLBspFile-> dleafs[ StartLeafNodeNumber_a];

	memset( &Tr_o, 0x00, sizeof(TraceData_st));
	Tr_o. planenum= -1;
	Tr_o. FaceNum= -1;
	Tr_o. TextureName= NULL;
	VectorCopy( End, Tr_o.endpos);
	Tr_o. StartLeafNum= StartLeafNodeNumber_a;
	Tr_o. HitDist= 99999.f;//MAGIC#
	if( startleaf-> contents== CONTENTS_SOLID)
		Tr_o.startsolid= TRUE;

	TmpTrace. planenum= -1;
	TmpTrace. fraction= 1.0f;
	TmpTrace. FurthestNonBlockingLeafNum= -1;
	int NodeNum_a= 0;// start at the root of the BSP tree
	traceExpandR( NodeNum_a, Tr_o);

	if( Tr_o.fraction!= 1.0f)//KO
		{
		dplane_t * pPlane_a= & pHLBspFile-> dplanes[ Tr_o.planenum];
		memcpy( & Tr_o.plane, pPlane_a, sizeof(dplane_t));
		VectorInterp( Start, pPlane_a-> normal, Tr_o. HitDist, Tr_o. endpos);

		vec3_t Delta_a;
		VectorSubtract(End, Start, Delta_a);
		float Length_a= VectorLength(Delta_a);

		if( Flags & (KB_Face|KB_Texture))
			{
			TraceData_st TraceF_a;
			memcpy( &TraceF_a, & Tr_o, sizeof(TraceData_st));
/*
			memcpy( &TraceF_a, & Tr_o, sizeof(TraceData_st));
			//off plane, normal to plane && closer to start
//			Offset_a= -Offset_a;
//			vec3_t V_a;
//			VectorScale( pPlane_a-> normal, Offset_a, V_a);
			VectorSubtract( TraceF_a. endpos, V_a, TraceF_a. endpos);
			Tr_o. FaceNum= traceMoveFindFace( Start, TraceF_a);
*/
			Tr_o. FaceNum= traceMoveFindFace( Start, TraceF_a);
			if( Tr_o. FaceNum== -2) 
				{		
				affDbgMsg(1,"TraceB: No Face on Plane");
				}
			else if( Tr_o. FaceNum== -1) 
				{		
				affDbgMsg(1,"TraceB: Endpos not in Face");
				}
			if( Flags & KB_Texture && Tr_o. FaceNum>= 0)
				Tr_o. TextureName= getTextureName( pHLBspFile, & pHLBspFile-> dfaces[Tr_o. FaceNum]);
			}
	
		//off plane, normal to plane && in blocking leaf
		TraceData_st Trace_a;
		memcpy( &Trace_a, & Tr_o, sizeof(TraceData_st));
		float Offset_a= DotProduct( pPlane_a-> normal, Delta_a);
		if( Offset_a> 0)// "if( Offset_a< 0)" if endpos needs to be in a the non blocking leaf
			Offset_a= K_EpsilonTrace;
		else
			Offset_a= -K_EpsilonTrace;
		vec3_t V_a;
		VectorScale( pPlane_a-> normal, Offset_a, V_a);
		VectorAdd( Trace_a. endpos, V_a, Trace_a. endpos);
		int LeafF0_a= pHLBspFile-> findLeaf( Trace_a. endpos);
		if( Trace_a.contents== CONTENTS_SOLID && LeafF0_a!=0) 
			{		affDbgMsg(1,"TraceB: BUG: Blockg Leaf should be solid");
			}
//DBGASSERT( Trace_a.contents!= CONTENTS_SOLID || LeafF0_a==0);
		}

#ifdef IMPL_TRACETOFILE
	if( DbgTraceSave_s)
		{
		pFile_s= fopen(dbgTraceDataPFE,"a");
		fprintf(pFile_s," ===TRACE  Tr=%8.6f  HitDist=%7.2f Trplane#=%4i BlockLeaf#%4i(contents %2i) LastLeaf#%4i\n"
			, Tr_o. fraction, Tr_o. HitDist, Tr_o. planenum
			, Tr_o.BlockingLeafNum, Tr_o.contents
			, Tr_o.FurthestNonBlockingLeafNum);
		fclose(pFile_s);
		}
#endif
	}
//-------------------------------------------------------------------
int Trace_ct:: traceExpandR ( int NodeNum_i, TraceData_st & Tr_o)
	{
	// implementation is consistent with Bsp tree building process of inserting faces, ie it is
	// adapted to assign a segment to leafs.

	//_ the "if( Tr_o. fraction> ...)" checks and using the Side_a value 
	// result in looking for the nearest ( from Start) blocking leaf
	// 
	if( NodeNum_i< 0)// in a leaf
		{
		dleaf_t *pLeaf_a= & pHLBspFile-> dleafs[ -1- NodeNum_i];
   
		if ((pLeaf_a-> contents== CONTENTS_SOLID) ||  (pLeaf_a->contents== CONTENTS_SKY))
			{
			Tr_o. contents= pLeaf_a-> contents;
			if(( pLeaf_a-> contents== CONTENTS_SOLID)&& (Tr_o.startsolid))
				Tr_o. allsolid= TRUE;	

			Tr_o. BlockingLeafNum= -1-NodeNum_i;
			Tr_o. FurthestNonBlockingLeafNum= TmpTrace. FurthestNonBlockingLeafNum;//last leaf before hitting this one
			Tr_o. HitDist= TmpTrace. HitDist;
			Tr_o. fraction= 0.0f;//something else? based on hull max halfsize/radius...
			Tr_o. planenum= TmpTrace. planenum;
			Tr_o. pPlane=  & pHLBspFile-> dplanes[ TmpTrace. planenum];
			return 1;
			}
		else
			{
			TmpTrace. FurthestNonBlockingLeafNum= -1-NodeNum_i;//last leaf before hitting this one
			return 0;
			}
		}

	dnode_t	 * node= &pHLBspFile->dnodes[NodeNum_i];
	dplane_t	 * plane= &pHLBspFile->dplanes[node->planenum];

	float Dist_a= DotProduct( plane-> normal, Start)- plane-> dist;
	int Side_a= Dist_a< 0;//front/on is zero
	if( Dist_a< 0 ) 
		Dist_a= -Dist_a;
	if( Tr_o. HitDist> Dist_a)
		{
		TmpTrace. planenum= node->planenum;
		TmpTrace. HitDist= Dist_a;
		float Offset_a=0;
		if( Type== K_AABox)
			{
			Offset_a= fabs( HullHSize[0]* plane->normal[0])
				+	fabs( HullHSize[1]* plane->normal[1]) 
				+	fabs( HullHSize[2]* plane->normal[2]);
			}
		else if ( Type== K_AACylinder)
			{
			Offset_a= fabs( HalfHeight()* plane->normal[CylAx])
				+	Radius_()* sqrt( 1.0f - plane->normal[CylAx]*plane->normal[CylAx]);
			}
		else if ( Type== K_Sphere)
			Offset_a= Radius_();
		
		int Ret_a= traceExpandR( node-> children[ Side_a], Tr_o);//closest to center first
		if( Dist_a< Offset_a && Tr_o. HitDist> Dist_a)//the plane cuts the volume
			{
			Ret_a= traceExpandR( node-> children[ Side_a^1], Tr_o);
			}
		}
	return -1;
	}
#endif
/*

   // solid objects
for (ent_index = 0; ent_index < num_entvars; ent_index++)
   {
   if (entvars[ent_index].brush_model_index> 0)
     {
     model_index = entvars[ent_index].brush_model_index;
     if( entvars[ent_index].renderamt== 255)
         {
         dmodel_t* m = &this-> dmodels[model_index];
         for( int c = 0; c < m->numfaces; c++)
             {
             Face_ct *face = faces[m->firstface + c];
             if (face == NULL)
              continue;
				 ....
				 }
          }
      }
  else if( entvars[ent_index].studio_model)
      {
		faces in world?
		}
   }
}

to get studio_model faces in world?
see StudioModel::DrawModel?
*/
