#ifdef _WIN32
#pragma once
#endif
#ifndef BSPTRACE_H_
#define BSPTRACE_H_

class HLBspFile_ct;
struct TraceParam_st;
//+++++++++++++++++++++++++++++++++++++
struct TraceData_st
{
   int      allsolid;   // if true, plane is not valid
   int		startsolid; // if true, the initial point is in a solid leaf
 	int		inopen,  inwater;
   float    fraction;   // 1.0 = didn't hit anything
   vec3_t   endpos;     // final position ( in blocking leaf)
	dplane_t plane; 
//	edict_t * ent; 
//	int		hitgroup;

   int      contents;   // contents of endpos
	float    HitDist;
	int      FaceNum;//hit face if fraction<1.0f
	dplane_t * pPlane;
   int      StartContents; // contents of start pos
	int		StartLeafNum;  //start leaf
	int      BlockingLeafNum; //last non-blocking leaf
	int      FurthestNonBlockingLeafNum;//blocking leaf
	char   * TextureName;//texture of hit face if fraction<1.0f
	int planenum;

#ifdef M_DEBUGC
   vec3_t   hitpos;     /* surface hit position (in solid) */
#endif
};
/*HL 
struct trace_t
{
	qboolean	allsolid;	// if true, plane is not valid
	qboolean	startsolid;	// if true, the initial point was in a solid area
	qboolean	inopen, inwater;
	float	fraction;		// time completed, 1.0 = didn't hit anything
	vec3_t	endpos;			// final position
	plane_t	plane;			// surface normal at impact
	edict_t	*ent;			// entity the surface is on
	int		hitgroup;		// 0 == generic, non zero is specific body part
} trace_t;
*/
//+++++++++++++++++++++++++++++++++++++
class Trace_ct
{
public:
static enum { K_Point=0, K_AABox= 1, K_Sphere=2, K_AACylinder= 3, K_Traces};//Type
static enum{ KB_Face= 1, KB_Texture=2};//Flags for data request
vec3_t Start;
vec3_t End;
unsigned Type;
unsigned Flags;
HLBspFile_ct * pHLBspFile;

	Trace_ct( HLBspFile_ct * pHLBspFile_i=0): Type( K_Point), Flags(0), pHLBspFile(pHLBspFile_i) { }
//	void reset();
	void setAABoxMinMax( vec3_t mins_i, vec3_t maxs_i) 
		{ 
		VectorInterp( mins_i, maxs_i, 0.5f, Start);
		VectorSubtract( maxs_i, mins_i, HullHSize);
		VectorScale( HullHSize, (float)0.5f, HullHSize);
		}
	void setAABoxCenterSize(  vec3_t Center_i, vec3_t AABoxHalfSize_i) 
		{ 
		VectorCopy( Center_i, Start);
		VectorCopy( AABoxHalfSize_i, HullHSize);
		}
	void setAACylinder( vec3_t Center_i, float  Radius_i, float HalfLength_i, int Axis_i)
		{
		VectorCopy( Center_i, Start);
		HullHSize[0]= Radius_i;
		HullHSize[1]= HalfLength_i;
		CylAx= Axis_i;
		CylB1= (1<< CylAx)& 3;
		CylB2= (1<< CylB1)& 3;
		}
	void setSphere( vec3_t Center_i, float  Radius_i)
		{
		VectorCopy( Center_i, Start);
		HullHSize[0]= Radius_i;
		}

	float XSize_() { return 2*HullHSize[0]; }
	float YSize_() { return 2*HullHSize[1]; }
	float HalfHeight_() { return HullHSize[2]; }
	float Height_() { return 2*HullHSize[2]; }
	float Radius_() { return HullHSize[0]; }
	int   traceToLeaf ( int LeafNodeNum_i, TraceData_st & Tr_o);
	void  traceMove( const vec3_t Start_i, const vec3_t End_i, TraceData_st & Tr_o);
	int   traceMoveFindFace (const vec3_t start, TraceData_st & Tr_o);
	void  traceExpand( TraceData_st & Tr_o);
	int	traceToLeaf2 (int LeafNodeNum_i, TraceData_st & Tr_o);

protected:
int CylAx;
int CylB1;
int CylB2;
TraceData_st TmpTrace;

vec3_t HullHSize;//positive values for HalfSizes if AABox; HullHSize[0]= Radius if sphere
	int   traceExpandR ( int NodeNum_i, TraceData_st & Tr_o);

	int  traceMovePointR ( int NodeNum_i, TraceParam_st &, TraceData_st & Tr_o);
	int  traceMoveR ( int NodeNum_i, TraceParam_st &, TraceData_st & Tr_o);
};
#endif //eof

