//
// BSP_tool - botman's Half-Life BSP utilities
//
// Copyright (C) 2001 - Jeffrey "botman" Broome
//
// modified by SamPlay
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifdef WIN32
#include <windows.h>
#endif

#include <memory.h>
#include <GL/gl.h>

#include "appgene.h"
#include "cmdlib.h"
#include "camera.h"
#include "World_ct.h"

extern WorldExtd_ct World;
//=============================================
int bBoundCamRange= 1;
//==========================================
//-----------------------------------------
void Camera_ct:: move_delta(float xd, float yd, float zd)
{
int Check_a= 1;
if( bBoundCamRange)
	{
	Check_a= 
 ((m_eye[0] + xd > M_WHullXMin) && (m_eye[0] + xd < M_WHullXMax) &&
  (m_eye[1] + yd > M_WHullYMin) && (m_eye[1] + yd < M_WHullYMax) &&
  (m_eye[2] + zd > M_WHullZMin) && (m_eye[2] + zd < M_WHullZMax));
	}
if( Check_a)
   {
	m_eye[0]+=xd, m_eye[1]+=yd, m_eye[2]+=zd;
	}
else
  Error("camera position out of range!\n");
}
//-----------------------------------------------
void Camera_ct:: move_delta(const vec3_t v)
{
int Check_a= 1;
if( bBoundCamRange)
	{
   Check_a=  ((m_eye[0] + v[ 0]> M_WHullXMin) && (m_eye[0] + v[ 0] < M_WHullXMax) &&
       (m_eye[1] + v[ 1] > M_WHullYMin) && (m_eye[1] + v[ 1] < M_WHullYMax) &&
       (m_eye[2] + v[ 2] > M_WHullZMin) && (m_eye[2] + v[ 2] < M_WHullZMax));
	}
if( Check_a)
   {
     m_eye[0]+=v[0], m_eye[1]+=v[1], m_eye[2]+=v[2];
   }
else
     Error("camera position out of range!\n");
}
//----------------------------------------------
void Camera_ct:: move(float x, float y, float z)
{
int Check_a= 1;
if( bBoundCamRange)
	{
   Check_a= ((x > M_WHullXMin) && (x < M_WHullXMax) &&
       (y > M_WHullYMin) && (y < M_WHullYMax) &&
       (z > M_WHullZMin) && (z < M_WHullZMax));

	}
if( Check_a)
     {
		m_eye[0]=x;
		m_eye[1]=y;
		m_eye[2]=z;
		}
else
   Error("camera position out of range!\n");
}
//-------------------------------------------------
void Camera_ct:: move(const vec3_t v)
{
int Check_a= 1;
if( bBoundCamRange)
	{
   Check_a= ((v[0] > M_WHullXMin) && (v[0] < M_WHullXMax) 
	&&  (v[1] > M_WHullYMin) && (v[1] < M_WHullYMax) 
	&&  (v[2] > M_WHullZMin) && (v[2] < M_WHullZMax));
	}
if( Check_a)
	{	
	m_eye[0]=v[0];
	m_eye[1]=v[1];
	m_eye[2]=v[2];
	}
else
   Error("camera position out of range!\n");
}
//-----------------------------------------------
void Camera_ct:: rotate_delta(float yaw_delta, float pitch_delta, float roll_delta)
{
m_pitch+= pitch_delta;
m_roll+= roll_delta;
m_yaw+= yaw_delta;
pitch_bounds();
}
//-------------------------------------------
void Camera_ct:: rotate(float yaw, float pitch, float roll)
{
m_pitch= pitch;
m_roll= roll;
m_yaw= yaw;
pitch_bounds();
}
//--------------------------------------------
void Camera_ct:: rotate_delta(const vec3_t v)
{
m_yaw-= v[0];
m_pitch-= v[1];
m_roll+= v[2];
pitch_bounds();
}
//-----------------------------------------------
void Camera_ct::rotate(const vec3_t v)
{
m_pitch = v[0];
m_roll = v[1];
m_yaw = v[2];
pitch_bounds();
}
//---------------------------------------
void Camera_ct:: pitch_bounds()
{
if (m_pitch > 0.0f)
   m_pitch= 0.0f;
else if (m_pitch < -180.f)
   m_pitch= -180.0f;
}
//--------------------------------------------
void Camera_ct:: loadMatrix()
{
glLoadMatrixf(CamMatrix);
}
//------------------------------------------
void Camera_ct:: transform()
{
CamMatrix. load_identity();
matrix_t r;
r.xrot(m_pitch);
CamMatrix*=r;

r.load_identity();
r.yrot(m_roll);
CamMatrix*=r;
r.load_identity();
r.zrot(m_yaw);
CamMatrix*=r;
CamMatrix. translate(-m_eye[0], -m_eye[1], -m_eye[2]);
}
//-----------------------------------------
void Camera_ct:: right_vector(vec3_t  out)
{
out[0] = CamMatrix[ 0];
out[1] = CamMatrix[ 4];
out[2] = CamMatrix[ 8];
}
//-----------------------------------------
void Camera_ct:: up_vector( vec3_t  out)
{
out[0] = CamMatrix[ 1];
out[1] = CamMatrix[ 5];
out[2] = CamMatrix[ 9];
}
//------------------------------------------
void Camera_ct:: fwd_vec( vec3_t  out)
{
out[0]= CamMatrix[ 2];
out[1]= CamMatrix[ 6];
out[2]= CamMatrix[ 10];
}
