// collision.cpp
//
// Copyright (C) 2001 - Jeffrey "botman" Broome
//
// modified by SamPlay
/* 
 * HL rendering engine
 * Copyright (c) 2000,2001 Bart Sekura
 *
 * Permission to use, copy, modify and distribute this software
 * is hereby granted, provided that both the copyright notice and 
 * this permission notice appear in all copies of the software, 
 * derivative works or modified versions.
 *
 * THE AUTHOR ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION AND DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES 
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * OpenGL window
 */

//
// Parts of this code is from:
//
//                  POLY ENGINE version 0.79
//                Written by Alexey Goloshubin
//

//
// Parts of this code is from:
//
// WinQuake, GLQuake, QuakeWorld, and GLQuakeWorld
// Copyright (C) 1996-1997 Id Software, Inc.

#include <GL/glut.h>
#include "appgene.h"
#include "mathlib_tpl.h"
#include "ChunkedVector_ct.h"

#include "WorldRender_ct.h"
#include "RenderFace_ct.h"

extern WorldRender_ct  WorldRndr;
extern bool_t bEnable_noclip;

int CollisionCutoff= 0;
//+++++++++++++++++++++++++++++++++++++++
class CollisionData_ct 
{
public:
    CollisionData_ct() : found(false), nearest(-1.0f), dir_len(0.0f) {}

    vec3_t src;
    vec3_t dir;
    vec3_t ndir;
    float  dir_len;

    bool found;
    float nearest;
    vec3_t nearest_poly;
};

float CollRadius= 15.0f;
//========================================================
//------------------------------------------------------
bool point_in_poly0(const vec3_t& Point_i, const RenderFace_ct* f)
{//to do more efficient ( crossproducts)
    int i;
    vec3_t v[32];

for (i = 0; i < f->count; i++)
    {
    int c= f->first + i;
    VectorSubtract( Point_i, ::WorldRndr. vt_array[c], v[i]);
    VectorNormalize( v[i]);
    }

float total = 0.0f;

for (i = 0; i < f->count-1; i++)
   total += (float) acos(DotProduct(v[i], v[i+1]));

total += (float) acos(DotProduct(v[f->count-1], v[0]));
if (fabsf(total-6.28f)<0.01f)
    return true;
return false;
}
//--------------------------------------------------------
void closest_on_segment(const vec3_t& a, const vec3_t& b, const vec3_t& p, vec3_t out)
{
vec3_t c;
vec3_t v;

VectorSubtract(p, a, c);
VectorSubtract(b, a, v);
float d = VectorLength(v);

if (d)
    VectorScale(v, (1.0f/d), v);

float t = DotProduct(v, c);

if (t < 0.0f)
    {
    VectorCopy(a, out);
    return;
    }

if (t > d)
    {
    VectorCopy(b, out);
    return;
    }

vec3_t vec_scale;
VectorScale(v, t, vec_scale);
VectorAdd(a, vec_scale, out);
}
//-------------------------------------------------------
void closest_on_poly(const vec3_t& p, const RenderFace_ct* f, vec3_t out)
{
int i;
vec3_t v[32];
float d[32];
vec3_t temp1;
vec3_t temp2;

for (i = 0; i < f->count-1; i++)
    {
    int c = f->first + i;
    VectorCopy(::WorldRndr. vt_array[c], temp1);
    VectorCopy(::WorldRndr. vt_array[c+1], temp2);

    closest_on_segment(temp1, temp2, p, v[i]);
    vec3_t t;
    VectorSubtract(p, v[i], t);
    d[i] = VectorLength(t);
    }
i = f->count-1;
VectorCopy(::WorldRndr. vt_array[f->first + i], temp1);
VectorCopy(::WorldRndr. vt_array[f->first], temp2);
closest_on_segment(temp1, temp2, p, v[i]);

vec3_t t;
VectorSubtract(p, v[i], t);
d[i] = VectorLength(t);

float min = d[0];
vec3_t r;
VectorCopy(v[0], r);
for (i = 1; i < f->count; i++)
    {
    if(d[i] < min)
        {
        min = d[i];
        VectorCopy(v[i], r);
        }
    }   
VectorCopy(r, out);
}
//--------------------------------------------------------
float intersect(const vec3_t r0,
          const vec3_t rn,
          const vec3_t p0,
          const vec3_t pn)
{
float d = -DotProduct(pn, p0);

float numer = DotProduct(pn, r0)+ d;
float denom = DotProduct(pn, rn);
  
if(denom <= -0.001f || denom >= 0.001f)
    return -(numer/denom);
return -1.0f;
}
//----------------------------------------------------------------
float intersect_sphere(const vec3_t r, 
                 const vec3_t rv, 
                 const vec3_t s, 
                 float sr)
{//to do simpler?
vec3_t q;
VectorSubtract(s, r, q);//q=s-r
float c = VectorLength(q);
float v = DotProduct(q, rv);
float d = sr* sr- (c* c- v* v);
if( d < 0.0f) 
	return -1.0f;
return v- sqrt( d);
}
//---------------------------------------
int RenderPlane_ct:: checkIntersect(const vec3_t start, const vec3_t dir, vec3_t ret_o) const
	{
   float denom= DotProduct(normal, dir);
   if( denom<= -EPSILON || denom >= EPSILON) 
		{
      float numer= -DotProduct(normal, start)+ Dist;
      float u= numer/denom;
      if(u > 0 && u <= 1.0f) 
			{
         vec3_t vec_scale;
         VectorScale( dir, u, vec_scale);
         VectorAdd( start, vec_scale, ret_o);
         return TRUE;
			}
		}
	return FALSE;
	}
//------------------------------------------------------------------
void check_collision(const RenderFace_ct* f, CollisionData_ct& coldat)
{
vec3_t FaceNormal_a;
RenderPlane_ct FacePlane= f-> Plane;
//if(f-> side)	
//	FacePlane.changeOrient();
VectorCopy( FacePlane. normal_(), FaceNormal_a);

vec3_t vec_scale;
VectorScale( FaceNormal_a, CollRadius, vec_scale);

vec3_t s;
VectorSubtract( coldat.src, vec_scale, s);

float t= FacePlane. distOf(s);

if(t > coldat.dir_len)
    return;

if( t< -2*CollRadius)
    return;

vec3_t PlaneIsec_a;

if( t< 0.0f)
    {
    vec3_t vec_scale;
    VectorScale(FaceNormal_a, CollRadius, vec_scale);
    if( !FacePlane. checkIntersect( s, vec_scale, PlaneIsec_a))
        return;
    }
else
    {
    if( !FacePlane. checkIntersect( s, coldat. dir, PlaneIsec_a))
        return;
    }

if( !point_in_poly0(PlaneIsec_a, f))
    closest_on_poly(PlaneIsec_a, f, PlaneIsec_a);

vec3_t opposite;
VectorCopy(coldat.ndir, opposite);
VectorChangeSign(opposite);

t= intersect_sphere(PlaneIsec_a, opposite, coldat.src, CollRadius);
if(t >= 0.0f && t <= coldat.dir_len)
    {
    if(!coldat.found || t< coldat.nearest)
        {
        coldat.found = true;
        coldat.nearest = t;
        VectorCopy(PlaneIsec_a, coldat.nearest_poly);
        }
    }
}
//------------------------------------------------------
void check_collisions(const vec3_t Src_i, const vec3_t  Dir_i, vec3_t  out)
{//recursive
if (++CollisionCutoff>5)
    {
    VectorCopy(Src_i, out);
    return;
    }

CollisionData_ct Coldat_a;
VectorCopy( Src_i, Coldat_a.src);
VectorCopy( Dir_i, Coldat_a.dir);

Coldat_a. dir_len= VectorLength(Dir_i);

if (Coldat_a.dir_len < 0.001f)
    {
    VectorCopy(Src_i, out);
    return;
    }

VectorScale( Coldat_a. dir, (1.0f/ Coldat_a. dir_len), Coldat_a. ndir);

if( !bEnable_noclip)
    {
    for( int i = 0; i < WorldRndr. max_visible_faces; i++)
        {
        RenderFace_ct* pFace_a = WorldRndr. VisibleSameTextureFaceLists[ i];
        while( pFace_a)
            {
            check_collision( pFace_a, Coldat_a);
            pFace_a= pFace_a-> next;
            }
        }
    }

if( Coldat_a.found)
    {
    vec3_t Src_a;
    vec3_t vec_scale_a;
    VectorCopy( Src_i, Src_a);
    if( Coldat_a. nearest >= 0.001f)
        {
        VectorScale( Coldat_a. ndir, (Coldat_a. nearest- 0.001f), vec_scale_a);
        VectorAdd( Src_a, vec_scale_a, Src_a);
        }

     vec3_t dst;
     vec3_t n;

     VectorAdd(Src_i, Dir_i, dst);
     VectorSubtract( Src_a, Coldat_a. nearest_poly, n);
     VectorNormalize( n);

    float t= intersect( dst, n, Coldat_a. nearest_poly, n);

     vec3_t newdst;
     vec3_t newdir;

     VectorScale( n, t, vec_scale_a);
     VectorAdd( dst, vec_scale_a, newdst);
     VectorSubtract( newdst, Coldat_a. nearest_poly, newdir);
     check_collisions( Src_a, newdir, out);
     return;
    }
vec3_t vec_scale2_a;
VectorScale( Coldat_a.ndir, (Coldat_a.dir_len-0.001f), vec_scale2_a);
VectorAdd( Src_i, vec_scale2_a, out);
}
