//
// BSP_tool - botman's Half-Life BSP utilities
//
// (http://planethalflife.com/botman/)
//
// gl_win.cpp
//

/* 
 * HL rendering engine
 * Copyright (c) 2000,2001 Bart Sekura
 *
 * Permission to use, copy, modify and distribute this software
 * is hereby granted, provided that both the copyright notice and 
 * this permission notice appear in all copies of the software, 
 * derivative works or modified versions.
 *
 * THE AUTHOR ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION AND DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES 
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * OpenGL window
 */

//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "cmdlib.h"
#include "glwin.h"

#include "BspViewConfig_ct.h"
#include "UserInput_ct.h"

extern bool m_center_window;
extern bool bFullScreen;
extern int ScreenWidth;
extern int ScreenHeight;
extern int Bitsppix;
int x_pos;
int y_pos;

int WndScreenWidth;
int WndScreenHeight;

extern BspViewConfig_ct config;

/////////////////////////////////////////////////////////////////

void
GLWindow::GLInfo::init()
{
    const char* vendor = (const char*) glGetString(GL_VENDOR);
    const char* version = (const char*) glGetString(GL_VERSION);
    const char* renderer = (const char*) glGetString(GL_RENDERER);

    const char* error = "error";
    m_vendor = vendor ? vendor : error;
    m_version = version ? version : error;
    m_renderer = renderer ? renderer : error;

    const char* ext = (const char*) glGetString(GL_EXTENSIONS);
    if(ext) {
        // parse extensions
        char* s = new char[strlen(ext)+1];
        if (!s)
            Error("Error allocating GLInfo memory!\n");
        strcpy(s, ext);
        char seps[] = " ";
        char* token = strtok(s, seps);
        while(token) {
            m_extensions.push_back(token);
            token = strtok(0, seps);
        }
        delete s;
    }
}
///////////////////////////////////////////////////////////////////

const char* GLWindow::className = "GL Engine";
GLWindow* GLWindow::m_instance = 0;
GLWindow::GLInfo GLWindow::gl_info;
vector<DEVMODE> GLWindow::m_modes;

void GLWindow::create(int width_i, int height_i, int bpp_i, int hz_i, bool fullscreen_i)
{
    int x_pos_a, y_pos_a;

    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_HREDRAW | CS_VREDRAW | CS_OWNDC, 
                      WindowProc, 0L, 0L,
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      className, NULL };

    RegisterClassEx(&wc);
    m_hinst = wc.hInstance;

    // if window exists, kill it
    if(m_instance) {
        m_instance->destroy();
        m_instance = 0;
    }

    // go fullscreen in requested
    if(fullscreen_i) {
        DEVMODE dm;
        memset(&dm, 0, sizeof(dm));
        dm.dmSize=sizeof(dm);
        dm.dmPelsWidth    = width_i;
        dm.dmPelsHeight    = height_i;
        dm.dmBitsPerPel    = bpp_i;
        dm.dmFields        = DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;

        if(hz_i) {
            dm.dmDisplayFrequency = hz_i;
            dm.dmFields |= DM_DISPLAYFREQUENCY;
        }

        if(ChangeDisplaySettings(&dm, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL) {
            Error("Error starting in Fullscreen mode!\n");
        }
    }

    // Create a window
    DWORD style;
    DWORD exStyle;

    if(fullscreen_i) 
			{
        exStyle = WS_EX_TOPMOST|WS_EX_APPWINDOW;
        style = WS_POPUP;
			}
    else 
			{
        exStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
        style = WS_OVERLAPPEDWINDOW;
			}

    x_pos_a= 0;
    y_pos_a= 0;

    if ((/*config.*/::x_pos < 0) && (/*config.*/::y_pos < 0))
       m_center_window = TRUE;
    else
		{
       m_center_window = FALSE;
       x_pos_a= /*config.*/::x_pos;
       y_pos_a= /*config.*/::y_pos;
		}
	
   
	RECT Rect_a;
//	::ScreenWidth= width_i;//$$
// ::ScreenHeight= height_i;//$$
	if(!bFullScreen)//$$
	{
 /* adjust the window rectangle to get the expected client area rect */
	Rect_a.left= x_pos_a; 
   Rect_a.top= y_pos_a;
   Rect_a.right= x_pos_a+ width_i; 
   Rect_a.bottom= y_pos_a+ height_i;
	AdjustWindowRectEx(& Rect_a,style,0,exStyle);
	width_i= Rect_a.right - Rect_a.left;	/* adjusted width */
	height_i= Rect_a.bottom - Rect_a.top;	/* adjusted height */   
	if(Rect_a.left < 0) 
		x_pos_a= 0;
	if(Rect_a.top < 0) 
		y_pos_a= 0;
	}

	m_hwnd = CreateWindowEx(exStyle,
                            className,
                            "BSP view",
                            style,
                            x_pos_a, y_pos_a,
                            width_i, height_i,
                            0,
                            0,
                            wc.hInstance,
                            0);
    if (!m_hwnd) 
	 {
        Error("Error creating Window!\n");
    }
	GetWindowRect(m_hwnd, & Rect_a);
   ::WndScreenWidth= Rect_a.right- Rect_a.left;//$$
   ::WndScreenHeight= Rect_a.bottom- Rect_a.top;//$$
	GetClientRect(m_hwnd, & Rect_a);
	::ScreenWidth= Rect_a. right;//$$
   ::ScreenHeight= Rect_a. bottom;//$$

    m_hdc = GetDC(m_hwnd);
    if(!m_hdc) {
        Error("Error Getting Device Context!\n");
    }

    static PIXELFORMATDESCRIPTOR pfd =
    {
        sizeof(PIXELFORMATDESCRIPTOR),
        1,                                    // Version Number
        PFD_DRAW_TO_WINDOW |                // Format Must Support Window
        PFD_SUPPORT_OPENGL |                // Format Must Support OpenGL
        PFD_DOUBLEBUFFER,                    // Must Support Double Buffering
        PFD_TYPE_RGBA,                        // Request An RGBA Format
        bpp_i,                                // Select Our Color Depth
        0, 0, 0, 0, 0, 0,                    // Color Bits Ignored
        0,                                    // No Alpha Buffer
        0,                                    // Shift Bit Ignored
        0,                                    // No Accumulation Buffer
        0, 0, 0, 0,                            // Accumulation Bits Ignored
        24,                                    // Z-Buffer (Depth Buffer) bits
        0,                                    // No Stencil Buffer
        0,                                    // No Auxiliary Buffer
        PFD_MAIN_PLANE,                        // Main Drawing Layer
        0,                                    // Reserved
        0, 0, 0                                // Layer Masks Ignored
    };

    GLuint pixFmt = 0;
    if(!(pixFmt = ChoosePixelFormat(m_hdc, &pfd))) {
        Error("Error in ChoosePixelFormat!\n");
    }

    if(!SetPixelFormat(m_hdc, pixFmt, &pfd)) {
        Error("Error in SetPixelFormat!\n");
    }

    if(!(m_hrc = wglCreateContext(m_hdc))) {
        Error("Error in wglCreateContext!\n");
    }

    if(!wglMakeCurrent(m_hdc, m_hrc)) {
        Error("Error in wglMakeCurrent!\n");
    }

    ShowWindow(m_hwnd, SW_SHOW);
    UpdateWindow(m_hwnd);
    SetForegroundWindow(m_hwnd);
    SetFocus(m_hwnd);

    m_instance= this;
//  m_bpp = bpp_i;
//	 m_fullscreen = fullscreen_i;

    enum_display();
    gl_info.init();

    resize();

    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
    SwapBuffers(m_hdc);

    init();

    m_active = TRUE;
}
//------------------------------------------------
void GLWindow::destroy()
{
    if( ::bFullScreen) 
		{
        ChangeDisplaySettings(0, 0);
        ShowCursor(TRUE);
		}

    if(m_hrc) 
	 {
        if(!wglMakeCurrent(0, 0)) {    }

        if(!wglDeleteContext(m_hrc)) {   }

        m_hrc = 0;
    }

    if(m_hdc) 
	 {
        if(!ReleaseDC(m_hwnd, m_hdc))  {   }
    }

    if(m_hwnd) 
	 {
        if(!DestroyWindow(m_hwnd))  {   }
    }

    UnregisterClass(className, m_hinst);
    m_hinst = 0;
//  ::ScreenWidth = 0;
//  ::ScreenHeight = 0;
//    m_bpp = 0;
//    m_fullscreen = FALSE;
    m_active = FALSE;
}
//-----------------------------------------------------
LRESULT GLWindow::window_proc(UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch(msg) 
		{
        case WM_ACTIVATE:
            if(wparam == WA_INACTIVE) 
					{
                m_active = FALSE;
					}
            else if ((wparam == WA_ACTIVE) ||
                     (wparam == WA_CLICKACTIVE)) 
					{
                m_active = TRUE;
					}

            return 0;

        case WM_CLOSE:
            PostQuitMessage(0);
            return 0;

        case WM_SIZE:
            if(wparam == SIZE_MINIMIZED) 
					{
                m_active = FALSE;
                return 0;
					}
            else 
            if ((wparam == SIZE_RESTORED) ||
                (wparam == SIZE_MAXIMIZED)) 
					 {
                m_active = TRUE;
					}
            //::ScreenWidth = LOWORD(lparam);
            //::ScreenHeight = HIWORD(lparam);
				//$$ untried
				RECT Rect_a;
				GetWindowRect(m_hwnd, & Rect_a);
				::WndScreenWidth= Rect_a.right- Rect_a.left;//$$
				::WndScreenHeight= Rect_a.bottom- Rect_a.top;//$$
				GetClientRect(m_hwnd, & Rect_a);
				::ScreenWidth= Rect_a.right;//$$
				::ScreenHeight= Rect_a.bottom;//$$
            
				resize();
            return 0;

        case WM_KEYDOWN:
            UserInput_ct:: keys[wparam] = TRUE;
            if(wparam == VK_ESCAPE) 
					{
                PostQuitMessage(0);
					}
            return 0;

        case WM_KEYUP:
            UserInput_ct:: keys[wparam]= FALSE;
            return 0;

        case WM_MOUSEMOVE:
            {
            POINT p;
            GetCursorPos(&p);
  //$$     if(!m_fullscreen)
  //$$        ScreenToClient(m_hwnd,&p);
            UserInput_ct:: MousePosX = (float) p.x;
            UserInput_ct:: MousePosY = (float) p.y;
            }
            return 0;

        case WM_LBUTTONDOWN:
            UserInput_ct:: bLeft_button= TRUE;
            return 0;

        case WM_LBUTTONUP:
            UserInput_ct:: bLeft_button= FALSE;
            return 0;
    }

    return DefWindowProc(m_hwnd, msg, wparam, lparam);
}

void
GLWindow::enum_display()
{
    if(m_modes.size()) 
	 {
        return;
    }

    int c  = 0;
    DEVMODE mode;
    while(EnumDisplaySettings(0, c++, &mode)) 
	 {
    if(mode.dmBitsPerPel >= 16 &&   // only 16bpp or more
            mode.dmPelsWidth >= 640 &&  // at least 640x...
            mode.dmDisplayFlags == 0)   // color and noninterlaced
        {
            m_modes.push_back(mode);
        }
    }
}

