/***
*
*	Copyright (c) 1996-2001, Valve LLC. All rights reserved.
*	
*	This product contains software technology licensed from Id 
*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*	All Rights Reserved.
*
****/

#ifndef MATHLIB_TPL_H_
#define MATHLIB_TPL_H_

#include <math.h>
/*
#ifdef __cplusplus
extern "C" {
#endif
*/


#ifdef DOUBLEVEC_T
typedef double vec_t;
#else
//#endif
//#ifndef EXTDLL_H
typedef float vec_t;
#endif

typedef vec_t vec3_t[3];	// x,y,z
typedef vec_t vec4_t[4];	// x,y,z,w

typedef double vecd_t;
typedef vecd_t vecd3_t[3];	// x,y,z
typedef vecd_t vecd4_t[4];	// x,y,z,w

typedef float vecf_t;
typedef vecf_t vecf3_t[3];	// x,y,z
//typedef vecf_t vecf4_t[4];	// x,y,z,w

#define	SIDE_FRONT		0
#define	SIDE_ON			2
#define	SIDE_BACK		1
#define	SIDE_CROSS		-2

#define	Q_PI	3.14159265358979323846

extern vecf3_t vec3_origin;
extern vecd3_t vecd3_origin;

// Use this definition globally
#define	ON_EPSILON		0.01
#define	EQUAL_EPSILON	0.001

void R_ConcatTransforms (const float in1[3][4], const float in2[3][4], float out[3][4]);


template <typename T> int VectorCompare(const T v1[3], const T v2[3] )
{
	int		i;
	for (i=0 ; i<3 ; i++)
		if (fabs(v1[i]-v2[i]) > EQUAL_EPSILON)
			return false;
	return true;
}
/*
#define VectorAvg(a) ( ( (a)[0] + (a)[1] + (a)[2] ) / 3 )
*/
#define VectorSubtract _VectorSubtract
#define VectorAdd _VectorAdd
#define VectorCopy  M_VectorCopy
#define VectorScale _VectorScale
#define DotProduct _DotProduct
#define VectorFill M_VectorFill
#define VectorInverse VectorChangeSign

#define M_VectorFill(a,b) { (a)[0]=(b); (a)[1]=(b); (a)[2]=(b);}
#define M_VectorCopy(a,b) { b[0]= a[0]; b[1]= a[1]; b[2]= a[2]; }

template <typename T> T Q_rint( T in)
	{	return floor (in + 0.5);
	}

template <typename T1, typename T2> inline double _DotProduct( const T1 v1[3], const T2 v2[3])
	{	return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
	}
template <typename T> inline void _VectorSubtract (const T va[3], const T vb[3], T out[3])
	{
	out[0] = va[0]-vb[0];
	out[1] = va[1]-vb[1];
	out[2] = va[2]-vb[2];
	}
template <typename T> inline void _VectorAdd (const T va[3], const T vb[3], T out[3])
	{
	out[0] = va[0]+vb[0];
	out[1] = va[1]+vb[1];
	out[2] = va[2]+vb[2];
	}
template <typename T> inline void _VectorCopy (const T in[3], T out[3])
	{
	out[0] = in[0];
	out[1] = in[1];
	out[2] = in[2];
	}
template <typename T1,typename T2> inline void _VectorScale (const T1 v[3], float scale, T2 out[3])
	{
	out[0] = v[0] * scale;
	out[1] = v[1] * scale;
	out[2] = v[2] * scale;
	}

template <typename T>  double Vector2DLength(const T v[3])
{
   int   i;
   double length;

   length = 0;
   for (i=0 ; i< 2 ; i++)  // ignore the Z axis
      length += v[i]*v[i];
   length = (float)sqrt(length); // FIXME

   return length;
}

template <typename T>  double VectorLength(const T v[3])
{
	int		i;
	double	length;
	
	length = 0;
	for (i=0 ; i< 3 ; i++)
		length += v[i]*v[i];
	length = sqrt (length);		// FIXME

	return length;
}

template <typename T>  inline void VectorMA (const T va[3], float scale, const T vb[3], T vc[3])
{
	vc[0] = va[0] + scale*vb[0];
	vc[1] = va[1] + scale*vb[1];
	vc[2] = va[2] + scale*vb[2];
}

template <typename T>  inline void VectorInterp (const T vstart[3], const T vend[3], float scale, T vc[3])
{
	vc[0] = vstart[0] + scale*(vend[0]-vstart[0]);
	vc[1] = vstart[1] + scale*(vend[1]-vstart[1]);
	vc[2] = vstart[2] + scale*(vend[2]-vstart[2]);
}

template <typename T>  void CrossProduct (const T v1[3], const T v2[3], T cross[3])
{
	cross[0] = v1[1]*v2[2] - v1[2]*v2[1];
	cross[1] = v1[2]*v2[0] - v1[0]*v2[2];
	cross[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

template <typename T>  T VectorNormalize (T v[3])
{
	int		i;
	double	length;

	if ( fabs(v[1] - 0.000215956) < 0.0001)
		i=1;

	length = 0;
	for (i=0 ; i< 3 ; i++)
		length += v[i]*v[i];
	length = sqrt (length);
	if (length == 0)
		return 0;
		
	for (i=0 ; i< 3 ; i++)
		v[i]= (T)(v[i]/length);	

	return (T)length;
}

template <typename T>  inline void VectorChangeSign (T v[3])
{
	v[0] = -v[0];
	v[1] = -v[1];
	v[2] = -v[2];
}

template <typename T>  inline void ClearBounds (T mins[3], T maxs[3])
{
	mins[0] = mins[1] = mins[2] = 99999;
	maxs[0] = maxs[1] = maxs[2] = -99999;
}

template <typename T>  void AddPointToBounds (T v[3], T mins[3], T maxs[3])
{
	int		i;
	T	val;

	for (i=0 ; i<3 ; i++)
	{
		val = v[i];
		if (val < mins[i])
			mins[i] = val;
		if (val > maxs[i])
			maxs[i] = val;
	}
}


template <typename T>  void AngleMatrix (const T angles[3], float matrix[3][4] )
{
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	
	angle = angles[2] * (Q_PI*2 / 360);
	sy = sin(angle);
	cy = cos(angle);
	angle = angles[1] * (Q_PI*2 / 360);
	sp = sin(angle);
	cp = cos(angle);
	angle = angles[0] * (Q_PI*2 / 360);
	sr = sin(angle);
	cr = cos(angle);

	// matrix = (Z * Y) * X
	matrix[0][0] = cp*cy;
	matrix[1][0] = cp*sy;
	matrix[2][0] = -sp;
	matrix[0][1] = sr*sp*cy+cr*-sy;
	matrix[1][1] = sr*sp*sy+cr*cy;
	matrix[2][1] = sr*cp;
	matrix[0][2] = (cr*sp*cy+-sr*-sy);
	matrix[1][2] = (cr*sp*sy+-sr*cy);
	matrix[2][2] = cr*cp;
	matrix[0][3] = 0.0;
	matrix[1][3] = 0.0;
	matrix[2][3] = 0.0;
}

template <typename T>  void AngleIMatrix (const T angles[3], float matrix[3][4] )
{
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	
	angle = angles[2] * (Q_PI*2 / 360);
	sy = sin(angle);
	cy = cos(angle);
	angle = angles[1] * (Q_PI*2 / 360);
	sp = sin(angle);
	cp = cos(angle);
	angle = angles[0] * (Q_PI*2 / 360);
	sr = sin(angle);
	cr = cos(angle);

	// matrix = (Z * Y) * X
	matrix[0][0] = cp*cy;
	matrix[0][1] = cp*sy;
	matrix[0][2] = -sp;
	matrix[1][0] = sr*sp*cy+cr*-sy;
	matrix[1][1] = sr*sp*sy+cr*cy;
	matrix[1][2] = sr*cp;
	matrix[2][0] = (cr*sp*cy+-sr*-sy);
	matrix[2][1] = (cr*sp*sy+-sr*cy);
	matrix[2][2] = cr*cp;
	matrix[0][3] = 0.0;
	matrix[1][3] = 0.0;
	matrix[2][3] = 0.0;
}

template <typename T>  void VectorRotate (const T in1[3], const float in2[3][4], T out[3])
{
	out[0] = DotProduct(in1, in2[0]);
	out[1] = DotProduct(in1, in2[1]);
	out[2] = DotProduct(in1, in2[2]);
}

template <typename T>  void VectorIRotate (const T in1[3], const float in2[3][4], T out[3])
{
	out[0] = in1[0]*in2[0][0] + in1[1]*in2[1][0] + in1[2]*in2[2][0];
	out[1] = in1[0]*in2[0][1] + in1[1]*in2[1][1] + in1[2]*in2[2][1];
	out[2] = in1[0]*in2[0][2] + in1[1]*in2[1][2] + in1[2]*in2[2][2];
}

template <typename T>  void VectorTransform (const T in1[3], const float in2[3][4], T out[3])
{
	out[0] = DotProduct(in1, in2[0]) + in2[0][3];
	out[1] = DotProduct(in1, in2[1]) +	in2[1][3];
	out[2] = DotProduct(in1, in2[2]) +	in2[2][3];
}

template <typename T>  float DistanceToIntersection(const T origin[3], const T vector[3],
                             const T plane_origin[3], const T plane_normal[3])
{
   float d= -((float)DotProduct(plane_normal, plane_origin));
 
   float numerator = (float)DotProduct(plane_normal, origin) + d;
   float denominator = (float)DotProduct(plane_normal, vector);
 
   if (fabs(denominator) < 0.00001)
      return (-1.0f);  // normal is orthogonal to vector, no intersection

   return -(numerator/denominator);
}

template <typename T>  bool VectorIntersectPlane(const T origin[3], const T vector[3],
                          const T plane_origin[3], const T plane_normal[3],
                          T intersect_point[3])
{
   float dist;
   T v_temp[3];

   dist = DistanceToIntersection(origin, vector, plane_origin, plane_normal);

   if (dist < 0)
      return FALSE;

   VectorScale(vector, dist, v_temp);
   VectorAdd(origin, v_temp, intersect_point);

   return TRUE;
}

template <typename T>  void AngleQuaternion( const T angles[3], T quaternion[4] )
{
	float		angle;
	float		sr, sp, sy, cr, cp, cy;

	// FIXME: rescale the inputs to 1/2 angle
	angle = angles[2] * 0.5;
	sy = sin(angle);
	cy = cos(angle);
	angle = angles[1] * 0.5;
	sp = sin(angle);
	cp = cos(angle);
	angle = angles[0] * 0.5;
	sr = sin(angle);
	cr = cos(angle);

	quaternion[0] = sr*cp*cy-cr*sp*sy; // X
	quaternion[1] = cr*sp*cy+sr*cp*sy; // Y
	quaternion[2] = cr*cp*sy-sr*sp*cy; // Z
	quaternion[3] = cr*cp*cy+sr*sp*sy; // W
}

template <typename T>  void QuaternionMatrix( const T quaternion[4], float (*matrix)[4] )
{

	matrix[0][0] = 1.0 - 2.0 * quaternion[1] * quaternion[1] - 2.0 * quaternion[2] * quaternion[2];
	matrix[1][0] = 2.0 * quaternion[0] * quaternion[1] + 2.0 * quaternion[3] * quaternion[2];
	matrix[2][0] = 2.0 * quaternion[0] * quaternion[2] - 2.0 * quaternion[3] * quaternion[1];

	matrix[0][1] = 2.0 * quaternion[0] * quaternion[1] - 2.0 * quaternion[3] * quaternion[2];
	matrix[1][1] = 1.0 - 2.0 * quaternion[0] * quaternion[0] - 2.0 * quaternion[2] * quaternion[2];
	matrix[2][1] = 2.0 * quaternion[1] * quaternion[2] + 2.0 * quaternion[3] * quaternion[0];

	matrix[0][2] = 2.0 * quaternion[0] * quaternion[2] + 2.0 * quaternion[3] * quaternion[1];
	matrix[1][2] = 2.0 * quaternion[1] * quaternion[2] - 2.0 * quaternion[3] * quaternion[0];
	matrix[2][2] = 1.0 - 2.0 * quaternion[0] * quaternion[0] - 2.0 * quaternion[1] * quaternion[1];
}

template <typename T>  void QuaternionSlerp( const T p[4], T q[4], float t, T qt[4] )
{
	int i;
	float omega, cosom, sinom, sclp, sclq;

	// decide if one of the quaternions is backwards
	float a = 0;
	float b = 0;
	for (i = 0; i < 4; i++) {
		a += (p[i]-q[i])*(p[i]-q[i]);
		b += (p[i]+q[i])*(p[i]+q[i]);
	}
	if (a > b) {
		for (i = 0; i < 4; i++) {
			q[i] = -q[i];
		}
	}

	cosom = p[0]*q[0] + p[1]*q[1] + p[2]*q[2] + p[3]*q[3];

	if ((1.0 + cosom) > 0.00000001) {
		if ((1.0 - cosom) > 0.00000001) {
			omega = acos( cosom );
			sinom = sin( omega );
			sclp = sin( (1.0 - t)*omega) / sinom;
			sclq = sin( t*omega ) / sinom;
		}
		else {
			sclp = 1.0 - t;
			sclq = t;
		}
		for (i = 0; i < 4; i++) {
			qt[i] = sclp * p[i] + sclq * q[i];
		}
	}
	else {
		qt[0] = -p[1];
		qt[1] = p[0];
		qt[2] = -p[3];
		qt[3] = p[2];
		sclp = sin( (1.0 - t) * 0.5 * Q_PI);
		sclq = sin( t * 0.5 * Q_PI);
		for (i = 0; i < 3; i++) {
			qt[i] = sclp * p[i] + sclq * qt[i];
		}
	}
}


/*
#ifdef __cplusplus
}
#endif
*/
#endif

