//
// BSP_tool - botman's Half-Life BSP utilities
//
// (http://planethalflife.com/botman/)
//
// paklib.cpp
//
// Copyright (C) 2001 - Jeffrey "botman" Broome
//
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "cmdlib.h"
#include "bspfile2.h"
#include "scriplib.h"
#include "paklib.h"
#include "wadlib.h"
#include "HalfLifeMods_ct.h"

extern HalfLifeMods_ct  HalfLifeMods;
extern dheader_t *header;
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
//===============================================
//-----------------------------------------------
FILE *P_OpenPak (char *filename)
{
FILE *pakhandle;
pakhandle= SafeOpenRead(filename);
if (pakhandle == NULL)
   {
   Error ("Can't open PAK file %s\n", filename);
   return NULL;
   }
return pakhandle;
}
//---------------------------------------------------
int P_ReadPakHeader(FILE *pakhandle, pakheader_t *pakheader)
{
SafeRead(pakhandle, pakheader, sizeof(pakheader_t));
if (strncmp(pakheader->identification, "PACK", 4))
	return 0;
pakheader-> dir_offset = LittleLong(pakheader->dir_offset);
pakheader-> dir_length = LittleLong(pakheader->dir_length);
return 1;
}
//---------------------------------------------------
void P_ReadPakInfo(FILE *pakhandle, long offset, int num_entries, pakinfo_t *pakinfo)
{
int index;
fseek(pakhandle, offset, SEEK_SET);
for (index=0; index < num_entries; index++)
   {
   SafeRead(pakhandle, &pakinfo[index], sizeof(pakinfo_t));
   pakinfo[index].file_pos = LittleLong(pakinfo[index].file_pos);
   pakinfo[index].file_length = LittleLong(pakinfo[index].file_length);
   }
}
//---------------------------------------------------
void P_ReadPakItem(FILE *pakhandle, pakinfo_t *pakinfo, void *buffer)
{
fseek(pakhandle, pakinfo->file_pos, SEEK_SET);
SafeRead(pakhandle, buffer, pakinfo->file_length);
}
//---------------------------------------------------
	//_ Search PAK config info for filename
bool SearchPakFilename(char *filename, pakconfig_t **pakconfig, pakinfo_t **pakinfo)
{
int mod_index;
  //_ is the MOD directory known from the BSP file?
if ( HalfLifeMods.bsp_mod_index >= 0)
   {
   pakconfig_t *pPakConfig =HalfLifeMods.mods[HalfLifeMods.bsp_mod_index].pPak;
   int index;
   while (pPakConfig)
      {
      for (index=0; index < pPakConfig->num_entries; index++)
         {
         if (stricmp(pPakConfig->pakinfo[index].filename, filename) == 0)
            {
            *pakconfig = pPakConfig;
            *pakinfo = &pPakConfig->pakinfo[index];
            return TRUE;
            }
         }
      pPakConfig = pPakConfig->next_config;
      }
   }
   // use the valve directory by default...
pakconfig_t *pPakConfig = HalfLifeMods.mods[HalfLifeMods.valve_mod_index].pPak;
int index;
while (pPakConfig)
   {
   for (index=0; index < pPakConfig->num_entries; index++)
      {
      if (stricmp(pPakConfig->pakinfo[index].filename, filename) == 0)
         {
         *pakconfig = pPakConfig;
         *pakinfo = &pPakConfig->pakinfo[index];
         return TRUE;
         }
      }
   pPakConfig = pPakConfig->next_config;
   }

   // if a MOD directory is known and PAK file not found so far, return error
if( HalfLifeMods.bsp_mod_index >= 0)
	return FALSE;

for (mod_index=0; mod_index < HalfLifeMods.num_mods; mod_index++)
   {
   pakconfig_t *pPakConfig = HalfLifeMods.mods[mod_index].pPak;
   int index;
   while (pPakConfig)
      {
      for (index=0; index < pPakConfig->num_entries; index++)
         {
         if (stricmp(pPakConfig->pakinfo[index].filename, filename) == 0)
            {
            *pakconfig = pPakConfig;
            *pakinfo = &pPakConfig->pakinfo[index];
            return TRUE;
            }
         }
      pPakConfig = pPakConfig->next_config;
      }
   }
return FALSE;
}
//--------------------------------------------------------
//Load BSP file from PAK config info
bool HLBspFile_ct:: loadPakBSPFile  (char *filename, unsigned LoadFlags_i)
{
char BspFileName_a[256];
pakconfig_t *pakconfig = NULL;
pakinfo_t *pakinfo = NULL;
int i;

strcpy(BspFileName_a, filename);
if (!SearchPakFilename(BspFileName_a, &pakconfig, &pakinfo))
   {
   strcpy(BspFileName_a, "maps/");
   strcat(BspFileName_a, filename);
   SearchPakFilename(BspFileName_a, &pakconfig , &pakinfo);
   }
if (pakinfo)
   {
   for (i = 0; i < HalfLifeMods.num_mods; i++)
      {
      if (pakconfig == HalfLifeMods.mods[i].pPak)
         {
         HalfLifeMods.bsp_mod_index = i;
         break;
         }
      }
   byte * pHeaderBytes= 0;
	pHeaderBytes= new byte[pakinfo->file_length];
   P_ReadPakItem(pakconfig->pakhandle, pakinfo, pHeaderBytes);
	header= (dheader_t *)pHeaderBytes;

   // swap the header
   for (i=0 ; i< sizeof(dheader_t)/4 ; i++)
      ((int *)header)[i]= LittleLong ( ((int *)header)[i]);
   if (header->version!= BSPVERSION)
       Error ("%s is version %i, not %i", filename, header->version, BSPVERSION);

   loadFile0( LoadFlags_i);

   delete [] pHeaderBytes; 
   return TRUE;
   }
return FALSE;
}

void ListPak(char* pakname)
{
        FILE* f = SafeOpenRead(pakname);
        pakheader_t head;
        pakfile_t* pdir;
        long i=0,imax=0;
        long totlen=0;

        SafeRead(f,&head,sizeof(pakheader_t));
        imax=head.dir_length/sizeof(pakfile_t);

        pdir = new pakfile_t[imax];
        fseek(f,head.dir_offset,SEEK_SET);
        SafeRead(f,pdir,head.dir_length);
       
        fseek(f,0,SEEK_END);
        totlen=ftell(f);
        fclose(f);
        for(i;i<imax;i++)
				{
            printf ("%64s : %7i\n", pdir[i].name,pdir[i].filelen);
				}
        printf("Total Length: %li bytes\n",totlen);
		  delete []pdir;
}
