//
// BSP_tool - botman's Half-Life BSP utilities
//
// (http://planethalflife.com/botman/)
//
// print.cpp
//
// Copyright (C) 2001 - Jeffrey "botman" Broome
//
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "appgene.h"
#include "cmdlib.h"
#include "bspfile2.h"
#include "bsp_ext.h"
#include "RenderFace_ct.h"
#include "u_RzbVector_ctpl.h"

#include <fstream>
using namespace std;

char DataOutDirS[]= "dataout"M_PATHSEP;

extern HLBspFile_ct HLBspFile;
extern char MapName[];
int getLeafFacePlanes( HLBspFile_ct * pHLBspFile_i, const dleaf_t & Leaf_a, int * & PlaneIds_o);

int haveSameNormal( int planenum1_i, int planenum2_i);
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
char DataPFE[]= "_BspData.txt";
char DataLeafs[ 512]="";

static FILE * fp_s;

#define MF_outText printf

//================================================
/*
unsigned short* p = dmarksurfaces + leaf.firstmarksurface;
for( int x = 0; x < leaf. nummarksurfaces; x++)
	{
   int face_idx = *p++;
   RenderFace_ct* f = faces[face_idx];
	}
*/
/*
//-------------------------------
void outBSPInfoToTxtFile()
{
FILE * fp= fopen(DataPFE,"w");
fclose(fp);
outText_leafsPlanes2( );
}
*/
//------------------------------------------------
void HLBspFile_ct:: outText_stats(FILE * pFile_i)
{
fprintf(pFile_i,
	"nummodels= %d\n"
	"numvertexes= %d\n"
	"numplanes= %d\n"
	"numleafs= %d\n"
	"numnodes= %d\n"
	"numtexinfo= %d\n"
	"numfaces= %d\n"
	"numedges= %d\n"
	"num_entities= %d\n"
	,
	nummodels,
	numvertexes,
	numplanes,
	numleafs,
	numnodes,
	numtexinfo,	
	numfaces,
	numedges,
	num_entities
	);
}
//------------------------------------------------
void HLBspFile_ct:: outText_models(void)
{
int i;

MF_outText("Printing %d Models...\n\n", nummodels);

for (i=0; i < nummodels; i++)
   {
      MF_outText("Model index           : %d\n", i);
      MF_outText("Model boundbox min (X): %f\n", dmodels[i].mins[0]);
      MF_outText("                   (Y): %f\n", dmodels[i].mins[1]);
      MF_outText("                   (Z): %f\n", dmodels[i].mins[2]);
      MF_outText("Model boundbox max (X): %f\n", dmodels[i].maxs[0]);
      MF_outText("                   (Y): %f\n", dmodels[i].maxs[1]);
      MF_outText("                   (Z): %f\n", dmodels[i].maxs[2]);
      MF_outText("Model origin       (X): %f\n", dmodels[i].origin[0]);
      MF_outText("                   (Y): %f\n", dmodels[i].origin[1]);
      MF_outText("                   (Z): %f\n", dmodels[i].origin[2]);
      MF_outText("Model headnode[0]     : %d\n", dmodels[i].headnode[0]);
      MF_outText("      headnode[1]     : %d\n", dmodels[i].headnode[1]);
      MF_outText("      headnode[2]     : %d\n", dmodels[i].headnode[2]);
      MF_outText("      headnode[3]     : %d\n", dmodels[i].headnode[3]);
      MF_outText("Model Visleafs        : %d\n", dmodels[i].visleafs);
      MF_outText("Model first poly ID   : %d\n", dmodels[i].firstface);
      MF_outText("Model num faces/polys : %d\n\n", dmodels[i].numfaces);
   }
}
//-------------------------------------------------------------
void HLBspFile_ct:: outText_vertices(void)
{
int i;

   MF_outText("Printing %d Vertexes...\n\n", numvertexes);

   for (i=0; i < numvertexes; i++)
   {
      MF_outText("Vertex index : %d\n", i);
      MF_outText("Vertex   (X) : %f\n",   dvertexes[i].point[0]);
      MF_outText("         (Y) : %f\n",   dvertexes[i].point[1]);
      MF_outText("         (Z) : %f\n\n", dvertexes[i].point[2]);
   }
}
//---------------------------------------------------------
void HLBspFile_ct:: outText_planes(void)
{
   int i;

   MF_outText("Printing %d Planes...\n\n", numplanes);

   for (i=0; i < numplanes; i++)
   {
      MF_outText("Plane index            : %d\n", i);
      MF_outText("Plane normal  (X)      : %f\n", dplanes[i].normal[0]);
      MF_outText("Plane normal  (Y)      : %f\n", dplanes[i].normal[1]);
      MF_outText("Plane normal  (Z)      : %f\n", dplanes[i].normal[2]);
      MF_outText("Plane dist from (0,0,0): %f\n", dplanes[i].dist);
      MF_outText("Plane type             : %d\n\n", dplanes[i].type);
   }
}
//----------------------------------------------
void HLBspFile_ct:: outText_leafs(void)
{
   int i;

   MF_outText("Printing %d Leafs...\n\n", numleafs);

   for (i=0; i < numleafs; i++)
   {
      MF_outText("Leaf index            : %d\n", i);
      MF_outText("Leaf type             : %d\n", dleafs[i].contents);
      MF_outText("Leaf vislist          : %d\n", dleafs[i].visofs);
      MF_outText("Leaf boundbox min (X) : %d\n", dleafs[i].mins[0]);
      MF_outText("                  (Y) : %d\n", dleafs[i].mins[1]);
      MF_outText("                  (Z) : %d\n", dleafs[i].mins[2]);
      MF_outText("Leaf boundbox max (X) : %d\n", dleafs[i].maxs[0]);
      MF_outText("                  (Y) : %d\n", dleafs[i].maxs[1]);
      MF_outText("                  (Z) : %d\n", dleafs[i].maxs[2]);
      MF_outText("Leaf idx of facelist  : %d\n", dleafs[i].firstmarksurface);
      MF_outText("Leaf number of faces  : %d\n", dleafs[i].nummarksurfaces);
      MF_outText("Ambient sounds: water : %d\n", dleafs[i].ambient_level[0]);
      MF_outText("                sky   : %d\n", dleafs[i].ambient_level[1]);
      MF_outText("                slime : %d\n", dleafs[i].ambient_level[2]);
      MF_outText("                lava  : %d\n\n", dleafs[i].ambient_level[3]);
  }
}
//---------------------------------------------------
void HLBspFile_ct:: outText_nodes(void)
{
   int i;

   MF_outText("Printing %d Nodes...\n\n", numnodes);

   for (i=0; i < numnodes; i++)
   {
      MF_outText("Node index           : %d\n", i);
      MF_outText("Plane ID             : %d\n", dnodes[i].planenum);

      if (dnodes[i].children[0]>= 0)
         MF_outText("Front child node     : %d\n", dnodes[i].children[0]);
      else
         MF_outText("Front child leaf     : %d\n", -1-dnodes[i].children[0]);

      if(dnodes[i].children[1]>= 0)
         MF_outText("Back child node      : %d\n", dnodes[i].children[1]);
      else
         MF_outText("Back child leaf      : %d\n", -1-dnodes[i].children[0]);

      MF_outText("Box minimum (X)      : %d\n", dnodes[i].mins[0]);
      MF_outText("            (Y)      : %d\n", dnodes[i].mins[1]);
      MF_outText("            (Z)      : %d\n", dnodes[i].mins[2]);
      MF_outText("Box maximum (X)      : %d\n", dnodes[i].maxs[0]);
      MF_outText("            (Y)      : %d\n", dnodes[i].maxs[1]);
      MF_outText("            (Z)      : %d\n", dnodes[i].maxs[2]);
      MF_outText("First polygon ID     : %d\n", dnodes[i].firstface);
      MF_outText("Number of faces/polys: %d\n\n", dnodes[i].numfaces);
   }
}
//---------------------------------------------------------------------
void HLBspFile_ct:: outText_texinfo(void)
{
   int i;

   MF_outText("Printing %d Texinfos...\n\n", numtexinfo);

   for (i=0; i < numtexinfo; i++)
   {
      MF_outText("Texinfo index : %d\n", i);
      MF_outText("   vecs[0][0] : %5.2f\n", texinfo[i].vecs[0][0]);
      MF_outText("   vecs[0][1] : %5.2f\n", texinfo[i].vecs[0][1]);
      MF_outText("   vecs[0][2] : %5.2f\n", texinfo[i].vecs[0][2]);
      MF_outText("   vecs[0][3] : %5.2f\n", texinfo[i].vecs[0][3]);
      MF_outText("   vecs[1][0] : %5.2f\n", texinfo[i].vecs[1][0]);
      MF_outText("   vecs[1][1] : %5.2f\n", texinfo[i].vecs[1][1]);
      MF_outText("   vecs[1][2] : %5.2f\n", texinfo[i].vecs[1][2]);
      MF_outText("   vecs[1][3] : %5.2f\n", texinfo[i].vecs[1][3]);
      MF_outText("       miptex : %d\n", texinfo[i].miptex);
      MF_outText("        flags : %d\n\n", texinfo[i].flags);
   }
}
//---------------------------------------------------------
void HLBspFile_ct:: outText_miptex(void)
{
   int i;
   dmiptexlump_t  *mtl;
   miptex_t* MipTex;
  
mtl = (dmiptexlump_t *)dtexdata;

MF_outText("Printing %d Miptexes...\n\n", mtl->nummiptex);

for (i=0; i < mtl->nummiptex; i++)
   {
   MF_outText("Mip index       : %d\n", i);
   if (mtl->dataofs[i] != -1)
      {
         MF_outText("Mip offset      : %ld\n", mtl->dataofs[i]);

         MipTex = (miptex_t*)(dtexdata + mtl->dataofs[i]);

         MF_outText("Mip name        : %s\n", MipTex->name);
         MF_outText("Mip width       : %d\n", MipTex->width);
         MF_outText("Mip height      : %d\n", MipTex->height);
         MF_outText("Mip offset0     : %ld\n", MipTex->offsets[0]);
         MF_outText("Mip offset1     : %ld\n", MipTex->offsets[1]);
         MF_outText("Mip offset2     : %ld\n", MipTex->offsets[2]);
         MF_outText("Mip offset3     : %ld\n\n", MipTex->offsets[3]);
      }
      else
         MF_outText("Mip offset      : %ld (mip is unused)\n\n", mtl->dataofs[i]);
   }
}
//-------------------------------------------------------
void HLBspFile_ct:: outText_faces(void)
{
   int i;

   MF_outText("Printing %d Faces...\n\n", numfaces);

   for (i=0; i < numfaces; i++)
   {
      MF_outText("Face index         : %d\n", i);
      MF_outText("Face plane Id      : %d\n", dfaces[i].planenum);
      MF_outText("Face plane side    : %d\n", dfaces[i].side);
      MF_outText("Face First edge    : %d\n", dfaces[i].firstedge);
      MF_outText("Face nr. of edges  : %d\n", dfaces[i].numedges);
      MF_outText("Face Texinfo Id    : %d\n", dfaces[i].texinfo);
      MF_outText("Face type of light : %d\n", dfaces[i].styles[0]);
      MF_outText("Face baselight     : %d\n", dfaces[i].styles[1]);
      MF_outText("Face light [1]     : %d\n", dfaces[i].styles[2]);
      MF_outText("Face light [2]     : %d\n", dfaces[i].styles[3]);
      MF_outText("Face lightmap      : %d\n\n", dfaces[i].lightofs);
   }
}
//--------------------------------------------------------
void HLBspFile_ct:: outText_edges(void)
{
   int i;

   MF_outText("Printing %d Edges...\n\n", numedges);

   for (i=0; i < numedges; i++)
   {
      MF_outText("Edge index            : %d\n", i);
      MF_outText("Edge vertex index 0   : %d\n", dedges[i].v[0]);
      MF_outText("Edge vertex index 1   : %d\n\n", dedges[i].v[1]);
   }
}
//-----------------------------------------------
void HLBspFile_ct:: outText_entity( std::ostream & OS_i, int i)
	{
   epair_t *pEpair;
	char Txt_a[ 1024];
    sprintf(Txt_a, " Entity: index=%d xyz=(%f %f %f) First brush=%d Number of brushes %d\n"
		, i
		, entities[i].origin[0], entities[i].origin[1], entities[i].origin[2]
		, entities[i].firstbrush, entities[i].numbrushes);
    OS_i<<Txt_a;
	 pEpair = entities[i].epairs;
    while (pEpair)
	   {
      OS_i<<"  KEY= "<< pEpair->key<<"  VALUE= "<<  pEpair->value<<'\n';
      pEpair= pEpair->next;
		}
   OS_i<<'\n';
	}
//--------------------------------------------------------
void HLBspFile_ct:: outText_entities(std::ostream & OS_i)
	{
	int i;
   OS_i<<"Entities count="<< num_entities<<'\n';
   for (i=0; i < num_entities; i++)
	   {
      outText_entity(OS_i, i);
		}
	OS_i<<'\n';
	}
//----------------------------------------------
void outText_leafsPlanes2(HLBspFile_ct * pHLBspFile_i)
{
char Txt_a [512];
FILE * fp= fopen("_BspData.txt","a");

fprintf(fp,"Printing %d LeafFacePlanes...\n\n", pHLBspFile_i-> numleafs);

int i;
int * PlaneIds_a=0;
for (i=0; i < pHLBspFile_i-> numleafs; i++)
   {
   const dleaf_t &Leaf_a = pHLBspFile_i-> dleafs[i];
	int Size_a= getLeafFacePlanes( pHLBspFile_i, Leaf_a, PlaneIds_a);
	if( Size_a>0)
		{
		fprintf(fp,"Leaf index= %d  contents= %d\n", i, Leaf_a.contents);
		int j=0;
		while( j< Size_a)
			{
			sprintf(Txt_a, "  FacePlaneNum= %d\n", PlaneIds_a[ j]);
			fprintf(fp, Txt_a);
			++j;
			}
		}
	}
delete[] PlaneIds_a;

fclose(fp);
}
//------------------------------------------
void saveEntites( char * PFE_i)
{
std::ofstream OFS_a;
char PFE_a[1024];
sprintf( PFE_a, "%s%s", DataOutDirS, PFE_a);
OFS_a.open(PFE_i);
HLBspFile.outText_entities(OFS_a);
OFS_a.close(); 
}