//
// BSP_tool - botman's Half-Life BSP utilities
//
// (http://planethalflife.com/botman/)
//
// render.cpp
//
// Copyright (C) 2001 - Jeffrey "botman" Broome
//
// modified by SamPlay
/* 
 * HL rendering engine
 * Copyright (c) 2000,2001 Bart Sekura
 *
 * Permission to use, copy, modify and distribute this software
 * is hereby granted, provided that both the copyright notice and 
 * this permission notice appear in all copies of the software, 
 * derivative works or modified versions.
 *
 * THE AUTHOR ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION AND DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES 
 * WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * OpenGL window
 */

//
// Parts of this code is from:
//
//                  POLY ENGINE version 0.79
//                Written by Alexey Goloshubin
//

//
// Parts of this code is from:
//
// WinQuake, GLQuake, QuakeWorld, and GLQuakeWorld
// Copyright (C) 1996-1997 Id Software, Inc.
//
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
// See the GNU General Public License for more details at:
// http://www.gnu.org/copyleft/gpl.html
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "appgene.h"
#include "entity.h"

#include "render.h"
#include "matrix.h"
//#include "mathlib.h"
#include "mathlib_tpl.h"

#include "bspfile2.h"
#include "camera.h"
#include "sys_time.h"
#include "world_ct.h"
#include "bsptrace.h"
#include "WorldRender_ct.h"
#include "fastmath.h"
#include "UserInput_ct.h"

#include <cassert>

extern int  ScreenWidth;
extern int  ScreenHeight;
extern bool bFullScreen;

extern bool_t bDrawTraceline;
extern HLBspFile_ct HLBspFile;
extern int bGetMoveVol;

#ifdef MI_HLBSPBUILD
#include "DisplayOption_ct.h"
extern DisplayOption_ct DispOpt;
extern HLBspFile_ct BspMove;
#endif

void Cam_init();
void Cam_update();

void Hud_init(/*int bOrthoIsSet_i*/);
void Hud_update();

void drawCompl();
void drawSurfs();
//
void draw2DHud( int );
void draw2DCrosshair(void);
void Simul_update();
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
WorldRender_ct         WorldRndr;
//BspMoveRender_ct       BspMoveRndr;
extern Camera_ct       Camera;

int   ScreenCenterX;
int   ScreenCenterY;

bool m_center_window;

double fovx= 70;
int frustum_checkCull(const short* mins, const short* maxs);

//iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii

#define Z_DEPTH (8000.0f)
//++++++++++++++++++++++++++++++++++
struct frustum_plane_t 
{
    vec3_t normal;
    float dist;
};
//+++++++++++++++++++++++++++++++++++++++++++++
struct frustum_t 
{
double left;
double right;
double bottom;
double top;
double znear;
double zfar;
};
	//_ frustum clipping planes
static frustum_plane_t frustum_planes[5];
	//_ frustum params
static double projection;
static double angle_horiz;
static double angle_vert;
static double sin_horiz;
static double cos_horiz;
static double sin_vert;
static double cos_vert;

static void projection_setup_s(int width, int height);
static void frustum_setup_s(const Camera_ct& cam);

static double znear_s = 1.0;

static frustum_t frustum;
//===============================================
//-------------------------------------------------------------
void projection_setup_s(int width, int height)
	{
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);

	double aspect = (double) height/ (double) width;
	double rprojz = tan(deg2rad(fovx)*0.5);

	glLoadIdentity();
	glFrustum(-rprojz, +rprojz, -aspect*rprojz, +aspect*rprojz, znear_s, Z_DEPTH);

	frustum.left = -rprojz;
	frustum.right = +rprojz;
	frustum.bottom = -aspect*rprojz;
	frustum.top = +aspect*rprojz;
	frustum.znear = 1.0;
	frustum.zfar = Z_DEPTH;

	double s = sin(deg2rad(fovx)*0.5);
	double c = cos(deg2rad(fovx)*0.5);

    // precompute stuff for frustum culling
	projection = (c*(double)(width>>1))/s;
	angle_horiz = atan2((double)(width>>1),projection);
	angle_vert = atan2((double)(height>>1),projection);
	sin_horiz = sin(angle_horiz);
	sin_vert = sin(angle_vert);
	cos_horiz = cos(angle_horiz);
	cos_vert = cos(angle_vert);

    // back to model view
	glMatrixMode(GL_MODELVIEW);
	}
//--------------------------------------------------
void frustum_setup_s( const Camera_ct& Cam_i)
	{
	vec3_t v;
	matrix_t m= Cam_i. matrix();

    // znear plane
	v[0]= 0;  
	v[1]= 0;  
	v[2]= 1.0f;
	m. multiply( v, frustum_planes[ 0]. normal);
	frustum_planes[ 0].dist= DotProduct( Cam_i.eye(), frustum_planes[ 0].normal);

    // left plane
	v[0]= cos_horiz;  
	v[1]= 0;  
	v[2]= sin_horiz;
   m. multiply( v, frustum_planes[1]. normal);
   frustum_planes[ 1].dist = DotProduct(Cam_i.eye(), frustum_planes[ 1].normal);
  
     // right plane
   v[0]= -cos_horiz;  
	v[1]= 0;  
	v[2]= sin_horiz;
   m. multiply( v, frustum_planes[ 2].normal);
   frustum_planes[ 2].dist = DotProduct(Cam_i.eye(), frustum_planes[ 2].normal);

    // top plane
    v[0]= 0;
	 v[1]= cos_vert;
	 v[2]= sin_vert;
    m. multiply( v, frustum_planes[ 3].normal);
    frustum_planes[ 3].dist = DotProduct(Cam_i.eye(), frustum_planes[ 3].normal);

    // bottom plane
    v[0]= 0;
	 v[1]= -cos_vert;
	 v[2]= sin_vert;
    m. multiply( v, frustum_planes[ 4].normal);
    frustum_planes[ 4].dist = DotProduct(Cam_i.eye(), frustum_planes[ 4].normal);
	}
//---------------------------------------------------------
int frustum_checkCull(const short* mins, const short* maxs)
{
for( int i= 0; i< 5; i++)
	{
   const short * min = mins;
	const short * max = maxs;
   const float * n= frustum_planes[ i]. normal;
   float d= -frustum_planes[ i].dist;

	d+= (*n>0 ? (*n*(float)*min) : (*n*(float)*max)); min++, max++; n++;
	d+= (*n>0 ? (*n*(float)*min) : (*n*(float)*max)); min++, max++; n++;
	d+= (*n>0 ? (*n*(float)*min) : (*n*(float)*max)); min++, max++; n++;
	if( d> 0)
       return 1;//ie cull
    }
return 0;
}
//-----------------------------------------------------------
Render::Render()
{
}
//------------------------------------------------------
Render::~Render()
{
}
//----------------------------------------------------
void Render::resize()
{
::projection_setup_s(::ScreenWidth, ::ScreenHeight);
::ScreenCenterX = ::ScreenWidth/ 2;
::ScreenCenterY = ::ScreenHeight/ 2;
}
//------------------------------------------------------
void Render::init()
{
glClearColor(0.0f, 0.0f, 0.0f, 0.0f);    
glClearDepth(1.0f);

    // misc
glDisable(GL_POLYGON_SMOOTH);
glDisable(GL_POLYGON_STIPPLE);
glDisable(GL_DITHER);
glDisable(GL_NORMALIZE);
glDisable(GL_LOGIC_OP);

    // texture settings
glEnable(GL_TEXTURE_2D);
glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    // depth buffer
glEnable(GL_DEPTH_TEST);
glDepthFunc(GL_LESS);
glDepthMask(GL_TRUE);

    // nice perspective
glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    // culling
glCullFace(GL_FRONT);
glEnable(GL_CULL_FACE);


#ifdef WIN32
PROC swp = wglGetProcAddress("wglSwapIntervalEXT");
if(swp)
    {
    typedef BOOL (WINAPI *swp_t)(int);
    swp_t s = (swp_t) swp;
    if (!bFullScreen)
       s(0);
    }
#endif
}
//--------------------------------------------------
void Render::shutdown()
{
}
//-----------------------------------------------------
void Render:: InitRender()
{
glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
#ifdef WIN32
flip();
#else
glutSwapBuffers();
#endif

UserInput_ct:: init();

::WorldRndr. pHLBspFile= & HLBspFile;
::WorldRndr. init();

#ifdef MI_HLBSPBUILD
if( bGetMoveVol)
	{

	::BspMoveRndr. pHLBspFile= & BspMove;
	::BspMoveRndr. init();
	}
#endif //MI_HLBSPBUILD

::Cam_init();
::Hud_init();

#ifdef WIN32
ShowCursor(FALSE);
#endif
}
//---------------------------------------------------
void begin_orto()
{
    // save projection matrix and set
glMatrixMode(GL_PROJECTION);
glPushMatrix();
glLoadIdentity();
glOrtho(0, ::ScreenWidth, 0, ::ScreenHeight, -1, 1);
 
    // save modelview
glMatrixMode(GL_MODELVIEW);
glPushMatrix();
glLoadIdentity();

glDisable(GL_CULL_FACE);
glDisable(GL_DEPTH_FUNC);
}
//------------------------------------------------------------
void end_orto()
{
glEnable(GL_DEPTH_FUNC);
glEnable(GL_CULL_FACE);

    // restore projection and modelview
glMatrixMode(GL_PROJECTION);
glPopMatrix();   
glMatrixMode(GL_MODELVIEW);
glPopMatrix();
}
//
//-----------------------------------------------
void resetMouseWin32()
{
	// auto recentering of the mouse
/*
POINT p;
p.x= ScreenCenterX, p.y= ScreenCenterY;
if (!m_fullscreen)
  ClientToScreen(m_hwnd,&p);
SetCursorPos(p.x, p.y);
*/
SetCursorPos(ScreenCenterX, ScreenCenterY);
}
//--------------------------------------------------
void Render:: frame()
{
static bool was_inactive = FALSE;

SysTime_Update();

#ifndef __linux__
if (!m_active)  // are we no longer the active window?
    {
    was_inactive = TRUE;
    Sleep(100);  // sleep 100 milliseconds to reduce CPU cycles
    return;
    }
#endif

if (was_inactive)  // are we just waking up?
	{
	was_inactive = FALSE;
    // do we need to reset the mouse or do anything after waking up here?
	}

UserInput_ct:: update();
    
frustum_setup_s(Camera);// setup for frustum culling first

WorldRndr. update();
//if( BspMoveRndr. pHLBspFile)
//	BspMoveRndr. update();

Simul_update();
Cam_update();
Hud_update();

Camera. loadMatrix();

glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

WorldRndr. draw();

#ifdef MI_HLBSPBUILD
DispOpt. drawSurfs();

if( BspMoveRndr. pHLBspFile)
	BspMoveRndr. draw();
#endif


::begin_orto();
::draw2DCrosshair();
::draw2DHud(1);
::end_orto();

#ifdef WIN32
flip();
#else
//glutSwapBuffers();
#endif
}
